#line 2 "libemilua.so.0.11.5.p/bc2_future.cpp"
#include <cstring>

namespace emilua::gperf::detail { using std::size_t; using std::strcmp; }
#line 6 "libemilua.so.0.11.5.p/bc2_future.cpp"


#line 9 "libemilua.so.0.11.5.p/bc2_future.cpp"

#include <boost/container/small_vector.hpp>
#include <boost/scope_exit.hpp>

#include <emilua/detail/core.hpp>
#include <emilua/async_base.hpp>
#include <emilua/future.hpp>

#line 18 "libemilua.so.0.11.5.p/bc2_future.cpp"
namespace emilua {


static char future_shared_state_mt_key;
static char promise_mt_key;
static char future_mt_key;
static char future_get_key;

struct future_shared_state
{
    boost::container::small_vector<lua_State*, 1> waiters;
    enum : char
    {
        empty,
        broken,
        value_ready,
        error_ready
    } state = empty;
    int value_ref = LUA_NOREF;
};

} // namespace emilua

#line 42 "libemilua.so.0.11.5.p/bc2_future.cpp"
namespace emilua {


static int promise_set_value(lua_State* L)
{
    lua_settop(L, 2);

    if (!lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &promise_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_getfenv(L, 1);
    lua_rawgeti(L, -1, 1);
    auto& state = *static_cast<future_shared_state*>(lua_touserdata(L, -1));

    if (state.state != future_shared_state::empty) {
        push(L, errc::promise_already_satisfied);
        return lua_error(L);
    }

    lua_pushvalue(L, 2);
    state.value_ref = luaL_ref(L, LUA_REGISTRYINDEX);
    state.state = future_shared_state::value_ready;

    auto vm_ctx = get_vm_context(L).shared_from_this();
    for (auto& fiber : state.waiters) {
        vm_ctx->strand().post([vm_ctx,fiber,ref=state.value_ref]() {
            auto push_value = [&ref](lua_State* L) {
                lua_rawgeti(L, LUA_REGISTRYINDEX, ref);
            };

            vm_ctx->fiber_resume(
                fiber,
                hana::make_set(
                    hana::make_pair(
                        vm_context::options::arguments,
                        hana::make_tuple(std::nullopt, push_value))));
        }, std::allocator<void>{});
    }
    state.waiters.clear();
    return 0;
}

static int promise_set_error(lua_State* L)
{
    lua_settop(L, 2);

    if (!lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &promise_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (lua_isnil(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    lua_getfenv(L, 1);
    lua_rawgeti(L, -1, 1);
    auto& state = *static_cast<future_shared_state*>(lua_touserdata(L, -1));

    if (state.state != future_shared_state::empty) {
        push(L, errc::promise_already_satisfied);
        return lua_error(L);
    }

    lua_pushvalue(L, 2);
    state.value_ref = luaL_ref(L, LUA_REGISTRYINDEX);
    state.state = future_shared_state::error_ready;

    auto vm_ctx = get_vm_context(L).shared_from_this();
    for (auto& fiber : state.waiters) {
        vm_ctx->strand().post([vm_ctx,fiber,ref=state.value_ref]() {
            auto push_value = [&ref](lua_State* L) {
                lua_rawgeti(L, LUA_REGISTRYINDEX, ref);
            };

            vm_ctx->fiber_resume(
                fiber,
                hana::make_set(
                    hana::make_pair(
                        vm_context::options::arguments,
                        hana::make_tuple(push_value))));
        }, std::allocator<void>{});
    }
    state.waiters.clear();
    return 0;
}

} // namespace emilua

#line 145 "libemilua.so.0.11.5.p/bc2_future.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_1  */
/* Computed positions: -k'5' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_1 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_1
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_1 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_1::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 1, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 0, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[4])];
}

const struct word_type_1 *
Perfect_Hash_1::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 9,
      MAX_WORD_LENGTH = 9,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_1 wordlist[] =
    {
      {"set_value", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, promise_set_value);
                return 1;
            }},
      {"set_error", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, promise_set_error);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 270 "libemilua.so.0.11.5.p/bc2_future.cpp"
namespace emilua {


static int future_get(lua_State* L)
{
    if (!lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &future_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto& vm_ctx = get_vm_context(L);
    auto current_fiber = vm_ctx.current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(vm_ctx, L);

    lua_getfenv(L, 1);
    lua_rawgeti(L, -1, 1);
    auto& state = *static_cast<future_shared_state*>(lua_touserdata(L, -1));

    switch (state.state) {
    case future_shared_state::broken:
        push(L, errc::broken_promise);
        return lua_error(L);
    case future_shared_state::value_ready:
        lua_pushnil(L);
        lua_rawgeti(L, LUA_REGISTRYINDEX, state.value_ref);
        return 2;
    case future_shared_state::error_ready:
        lua_rawgeti(L, LUA_REGISTRYINDEX, state.value_ref);
        return lua_error(L);
    case future_shared_state::empty:
        break;
    default:
        assert(false);
    }

    lua_pushvalue(L, -1);
    lua_pushlightuserdata(L, current_fiber);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto vm_ctx = get_vm_context(L).shared_from_this();
            auto& state = *static_cast<future_shared_state*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            auto current_fiber = static_cast<lua_State*>(
                lua_touserdata(L, lua_upvalueindex(2)));

            auto it = std::find(state.waiters.begin(), state.waiters.end(),
                                current_fiber);
            if (it == state.waiters.end()) {
                // set_value() was called before interrupt()
                return 0;
            }

            state.waiters.erase(it);
            vm_ctx->strand().post([vm_ctx,current_fiber]() {
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        hana::make_pair(
                            opt_args, hana::make_tuple(errc::fiber_canceled))));
            }, std::allocator<void>{});
            return 0;
        },
        2);
    set_interrupter(L, vm_ctx);

    state.waiters.emplace_back(current_fiber);
    return lua_yield(L, 0);
}

} // namespace emilua

#line 349 "libemilua.so.0.11.5.p/bc2_future.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_2  */
/* Computed positions: -k'' */

struct word_type_2 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 1, duplicates = 0 */

class Perfect_Hash_2
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_2 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_2::hash (const char *str, size_t len)
{
  return 0;
}

const struct word_type_2 *
Perfect_Hash_2::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 1,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 3,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 0
    };

  static const struct word_type_2 wordlist[] =
    {
      {"get", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &future_get_key);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail
#line 1 "../src/future.cpp"
// Copyright (c) 2023 Vinícius dos Santos Oliveira
// SPDX-License-Identifier: MIT OR BSL-1.0

#line 11 "../src/future.cpp"


namespace emilua {

char future_key;

#line 36 "../src/future.cpp"


static int future_shared_state_mt_gc(lua_State* L)
{
    auto& state = *static_cast<future_shared_state*>(lua_touserdata(L, 1));
    BOOST_SCOPE_EXIT_ALL(&) { state.~future_shared_state(); };

    switch (state.state) {
    case future_shared_state::empty:
    case future_shared_state::broken:
        return 0;
    case future_shared_state::value_ready:
    case future_shared_state::error_ready:
        break;
    }

    luaL_unref(L, LUA_REGISTRYINDEX, state.value_ref);
    return 0;
}

#line 154 "../src/future.cpp"


static int promise_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 451 "libemilua.so.0.11.5.p/bc2_future.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_1::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 177 "../src/future.cpp"
(L);
}

static int promise_mt_gc(lua_State* L)
{
    rawgetp(L, LUA_REGISTRYINDEX, &detail::context_key);
    auto vm_ctx = static_cast<vm_context*>(lua_touserdata(L, -1));
    lua_pop(L, 1);
    if (!vm_ctx) {
        // VM is closing and context_key in LUA_REGISTRYINDEX was collected
        // already
        return 0;
    }

    lua_getfenv(L, 1);
    if (lua_isnil(L, -1)) {
        // VM is closing and userdata's env was collected already
        return 0;
    }

    lua_rawgeti(L, -1, 1);
    auto state = static_cast<future_shared_state*>(lua_touserdata(L, -1));
    if (!state) {
        // VM is closing and userdata's env contents were collected already
        return 0;
    }

    state->state = future_shared_state::broken;

    std::shared_ptr<vm_context> vm_ctx2;
    try {
        vm_ctx2 = vm_ctx->shared_from_this();
    } catch (const std::bad_weak_ptr&) {
        // reached here from VM shutdown, so just ignore waiters
        return 0;
    }
    for (auto& fiber : state->waiters) {
        vm_ctx->strand().post([vm_ctx2,fiber]() {
            vm_ctx2->fiber_resume(
                fiber,
                hana::make_set(
                    hana::make_pair(
                        vm_context::options::arguments,
                        hana::make_tuple(errc::broken_promise))));
        }, std::allocator<void>{});
    }
    state->waiters.clear();
    return 0;
}

#line 301 "../src/future.cpp"


static int future_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 514 "libemilua.so.0.11.5.p/bc2_future.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_2::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 318 "../src/future.cpp"
(L);
}

static int future_mt_gc(lua_State* L)
{
    rawgetp(L, LUA_REGISTRYINDEX, &detail::context_key);
    auto vm_ctx = static_cast<vm_context*>(lua_touserdata(L, -1));
    lua_pop(L, 1);
    if (!vm_ctx) {
        // VM is closing and context_key in LUA_REGISTRYINDEX was collected
        // already
        return 0;
    }

    lua_getfenv(L, 1);
    if (lua_isnil(L, -1)) {
        // VM is closing and userdata's env was collected already
        return 0;
    }

    lua_rawgeti(L, -1, 1);
    auto state = static_cast<future_shared_state*>(lua_touserdata(L, -1));
    if (!state) {
        // VM is closing and userdata's env contents were collected already
        return 0;
    }

    constexpr auto spec{FMT_STRING(
        "Underlying promise for future {} is broken"
    )};
    if (state->waiters.size() != 0)
        vm_ctx->notify_deadlock(fmt::format(spec, lua_touserdata(L, 1)));
    return 0;
}

static int future_new(lua_State* L)
{
    lua_settop(L, 0);

    auto state = static_cast<future_shared_state*>(
        lua_newuserdata(L, sizeof(future_shared_state))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &future_shared_state_mt_key);
    setmetatable(L, -2);
    new (state) future_shared_state{};

    lua_createtable(L, /*narr=*/1, /*nrec=*/0);
    lua_pushvalue(L, 1);
    lua_rawseti(L, -2, 1);

    lua_newuserdata(L, /*size=*/1);
    lua_pushvalue(L, 2);
    lua_setfenv(L, -2);
    rawgetp(L, LUA_REGISTRYINDEX, &promise_mt_key);
    setmetatable(L, -2);

    lua_newuserdata(L, /*size=*/1);
    lua_pushvalue(L, 2);
    lua_setfenv(L, -2);
    rawgetp(L, LUA_REGISTRYINDEX, &future_mt_key);
    setmetatable(L, -2);

    return 2;
}

void init_future(lua_State* L)
{
    lua_pushlightuserdata(L, &future_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/1);

        lua_pushliteral(L, "new");
        lua_pushcfunction(L, future_new);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &future_shared_state_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/1);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, future_shared_state_mt_gc);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &promise_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "promise");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, promise_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, promise_mt_gc);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &future_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "future");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, future_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, future_mt_gc);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &future_get_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, future_get);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
}

} // namespace emilua
