#line 2 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"
#include <cstring>

namespace emilua::gperf::detail { using std::size_t; using std::strcmp; }
#line 6 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"


#line 9 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"

#include <limits>
#include <deque>

#include <fmt/format.h>

#include <emilua/recursive_mutex.hpp>

#line 18 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"
namespace emilua {


struct recursive_mutex_handle
{
    std::deque<lua_State*> pending;
    std::size_t nlocked = 0;
    lua_State* owner = nullptr;
};

static char recursive_mutex_mt_key;

} // namespace emilua

#line 33 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"
namespace emilua {


static int recursive_mutex_lock(lua_State* L)
{
    auto handle = static_cast<recursive_mutex_handle*>(lua_touserdata(L, 1));
    if (!handle || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &recursive_mutex_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto& vm_ctx = get_vm_context(L);
    auto current_fiber = vm_ctx.current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED_ASSUMING_INTERRUPTION_DISABLED(vm_ctx, L);

    if (handle->owner) {
        if (handle->owner == current_fiber) {
            if (
                handle->nlocked ==
                std::numeric_limits<decltype(handle->nlocked)>::max()
            ) {
                push(L, std::errc::value_too_large);
                return lua_error(L);
            }

            ++handle->nlocked;
            return 0;
        } else {
            handle->pending.emplace_back(current_fiber);
            return lua_yield(L, 0);
        }
    }

    handle->owner = current_fiber;
    assert(handle->nlocked == 0);
    handle->nlocked = 1;
    return 0;
}

static int recursive_mutex_try_lock(lua_State* L)
{
    auto handle = static_cast<recursive_mutex_handle*>(lua_touserdata(L, 1));
    if (!handle || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &recursive_mutex_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto current_fiber = get_vm_context(L).current_fiber();

    if (handle->owner) {
        if (handle->owner == current_fiber) {
            if (
                handle->nlocked ==
                std::numeric_limits<decltype(handle->nlocked)>::max()
            ) {
                push(L, std::errc::value_too_large);
                return lua_error(L);
            }

            ++handle->nlocked;
            lua_pushboolean(L, 1);
            return 1;
        } else {
            lua_pushboolean(L, 0);
            return 1;
        }
    }

    handle->owner = current_fiber;
    assert(handle->nlocked == 0);
    handle->nlocked = 1;
    lua_pushboolean(L, 1);
    return 1;
}

static int recursive_mutex_unlock(lua_State* L)
{
    auto handle = static_cast<recursive_mutex_handle*>(lua_touserdata(L, 1));
    if (!handle || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &recursive_mutex_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto& vm_ctx = get_vm_context(L);

    if (handle->owner != vm_ctx.current_fiber()) {
        push(L, std::errc::operation_not_permitted);
        return lua_error(L);
    }

    if (--handle->nlocked != 0)
        return 0;

    if (handle->pending.size() == 0) {
        handle->owner = nullptr;
        return 0;
    }

    auto next = handle->pending.front();
    handle->pending.pop_front();
    handle->owner = next;
    handle->nlocked = 1;
    vm_ctx.strand().post([vm_ctx=vm_ctx.shared_from_this(),next]() {
        vm_ctx->fiber_resume(
            next, hana::make_set(vm_context::options::skip_clear_interrupter));
    }, std::allocator<void>{});
    return 0;
}

} // namespace emilua

#line 160 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_1  */
/* Computed positions: -k'' */

struct word_type_1 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 5, duplicates = 0 */

class Perfect_Hash_1
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_1 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_1::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_1 *
Perfect_Hash_1::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 8,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 8
    };

  static const struct word_type_1 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"lock", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, recursive_mutex_lock);
                return 1;
            }},
      {"",{}},
      {"unlock", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, recursive_mutex_unlock);
                return 1;
            }},
      {"",{}},
      {"try_lock", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, recursive_mutex_try_lock);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail
#line 1 "../src/recursive_mutex.cpp"
// Copyright (c) 2020, 2023 Vinícius dos Santos Oliveira
// SPDX-License-Identifier: MIT OR BSL-1.0

#line 11 "../src/recursive_mutex.cpp"


namespace emilua {

char recursive_mutex_key;

#line 27 "../src/recursive_mutex.cpp"


#line 151 "../src/recursive_mutex.cpp"


static int recursive_mutex_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 257 "libemilua.so.0.11.5.p/bc2_recursive_mutex.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_1::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 180 "../src/recursive_mutex.cpp"
(L);
}

static int recursive_mutex_new(lua_State* L)
{
    auto buf = static_cast<recursive_mutex_handle*>(
        lua_newuserdata(L, sizeof(recursive_mutex_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &recursive_mutex_mt_key);
    setmetatable(L, -2);
    new (buf) recursive_mutex_handle{};
    return 1;
}

void init_recursive_mutex_module(lua_State* L)
{
    lua_pushlightuserdata(L, &recursive_mutex_key);
    lua_newtable(L);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "recursive_mutex");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(
            L,
            [](lua_State* L) -> int {
                auto key = tostringview(L, 2);
                if (key == "new") {
                    lua_pushcfunction(L, recursive_mutex_new);
                    return 1;
                } else {
                    push(L, errc::bad_index, "index", 2);
                    return lua_error(L);
                }
            });
        lua_rawset(L, -3);

        lua_pushliteral(L, "__newindex");
        lua_pushcfunction(
            L,
            [](lua_State* L) -> int {
                push(L, std::errc::operation_not_permitted);
                return lua_error(L);
            });
        lua_rawset(L, -3);
    }
    setmetatable(L, -2);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &recursive_mutex_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "recursive_mutex");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, recursive_mutex_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<recursive_mutex_handle>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);
}

} // namespace emilua
