#line 2 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
#include <cstring>

namespace emilua::gperf::detail { using std::size_t; using std::strcmp; }
#line 6 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"


#line 9 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"

#include <boost/scope_exit.hpp>

#include <emilua/file_descriptor.hpp>
#include <emilua/serial_port.hpp>
#include <emilua/async_base.hpp>
#include <emilua/byte_span.hpp>

#if EMILUA_CONFIG_USE_STANDALONE_ASIO
#include <asio/serial_port.hpp>
#else // EMILUA_CONFIG_USE_STANDALONE_ASIO
#include <boost/asio/serial_port.hpp>
#endif // EMILUA_CONFIG_USE_STANDALONE_ASIO

#line 24 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua {


static char serial_port_read_some_key;
static char serial_port_write_some_key;

static int serial_port_open(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

#if BOOST_OS_UNIX
    // Boost.Asio unfortunately ruins the state of whatever serial port you try
    // to open by setting "default serial port options" without ever asking the
    // user whether he wanted that. We skip all this crap by opening the device
    // ourselves.
    int fd = open(lua_tostring(L, 2), O_RDWR | O_NONBLOCK | O_NOCTTY);
    if (fd == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    asio_error_code ec;
    port->assign(fd, ec);
    assert(!ec); boost::ignore_unused(ec);
    return 0;
#else
    asio_error_code ec;
    port->open(static_cast<std::string>(tostringview(L, 2)), ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
#endif // BOOST_OS_UNIX
}

static int serial_port_close(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    port->close(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

static int serial_port_cancel(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    port->cancel(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

#if BOOST_OS_UNIX
static int serial_port_assign(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto handle = static_cast<file_descriptor_handle*>(lua_touserdata(L, 2));
    if (!handle || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    if (*handle == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::device_or_resource_busy);
        return lua_error(L);
    }

    lua_pushnil(L);
    setmetatable(L, 2);

    asio_error_code ec;
    port->assign(*handle, ec);
    assert(!ec); boost::ignore_unused(ec);

    return 0;
}

static int serial_port_release(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (port->native_handle() == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::bad_file_descriptor);
        return lua_error(L);
    }

    // https://github.com/chriskohlhoff/asio/issues/1182
    int newfd = dup(port->native_handle());
    BOOST_SCOPE_EXIT_ALL(&) {
        if (newfd != -1) {
            int res = close(newfd);
            boost::ignore_unused(res);
        }
    };
    if (newfd == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    asio_error_code ignored_ec;
    port->close(ignored_ec);

    auto fdhandle = static_cast<file_descriptor_handle*>(
        lua_newuserdata(L, sizeof(file_descriptor_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    setmetatable(L, -2);

    *fdhandle = newfd;
    newfd = -1;
    return 1;
}

static int serial_port_isatty(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (port->native_handle() == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::bad_file_descriptor);
        return lua_error(L);
    }

    lua_pushboolean(L, isatty(port->native_handle()));
    return 1;
}

static int serial_port_tcgetpgrp(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (port->native_handle() == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::bad_file_descriptor);
        return lua_error(L);
    }

    pid_t res = tcgetpgrp(port->native_handle());
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }
    lua_pushnumber(L, res);
    return 1;
}

static int serial_port_tcsetpgrp(lua_State* L)
{
    lua_settop(L, 2);

    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (port->native_handle() == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::bad_file_descriptor);
        return lua_error(L);
    }

    if (tcsetpgrp(port->native_handle(), luaL_checknumber(L, 2)) == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }
    return 0;
}
#endif // BOOST_OS_UNIX

static int serial_port_send_break(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    port->send_break(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

} // namespace emilua

#line 302 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua {


inline int serial_port_is_open(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    lua_pushboolean(L, port->is_open());
    return 1;
}

} // namespace emilua

#line 315 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_2  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_2 { const char* name; 
                        asio::serial_port_base::flow_control::type action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_2
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_2 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_2::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 1, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 0, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_2 *
Perfect_Hash_2::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 8,
      MAX_WORD_LENGTH = 8,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_2 wordlist[] =
    {
      {"software", 
                        asio::serial_port_base::flow_control::software},
      {"hardware", 
                        asio::serial_port_base::flow_control::hardware}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 435 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_3  */
/* Computed positions: -k'' */

struct word_type_3 { const char* name; 
                        asio::serial_port_base::parity::type action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_3
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_3 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_3::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_3 *
Perfect_Hash_3::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 4,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 4
    };

  static const struct word_type_3 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"odd",  asio::serial_port_base::parity::odd},
      {"even",  asio::serial_port_base::parity::even}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 498 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_4  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_4 { const char* name; 
                        asio::serial_port_base::stop_bits::type action; };
/* maximum key range = 13, duplicates = 0 */

class Perfect_Hash_4
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_4 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_4::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16,  1, 16, 16, 16, 16,  0, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16, 16, 16, 16, 16,
      16, 16, 16, 16, 16, 16
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_4 *
Perfect_Hash_4::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 15
    };

  static const struct word_type_4 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"two",  asio::serial_port_base::stop_bits::two},
      {"one",  asio::serial_port_base::stop_bits::one},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"one_point_five", 
                        asio::serial_port_base::stop_bits::onepointfive}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 620 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_1  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_1 { const char* name; int (*action)(lua_State*, asio::serial_port*); };
/* maximum key range = 9, duplicates = 0 */

class Perfect_Hash_1
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_1 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_1::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15,  2,  0,
      15, 15,  0, 15, 15, 15, 15, 15, 15, 15,
      15, 15,  0, 15, 15,  0, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15, 15, 15, 15, 15,
      15, 15, 15, 15, 15, 15
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_1 *
Perfect_Hash_1::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 6,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 6,
      MAX_HASH_VALUE = 14
    };

  static const struct word_type_1 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"parity", 
            [](lua_State* L, asio::serial_port* port) -> int {
                switch (lua_type(L, 3)) {
                case LUA_TNIL: {
                    asio::serial_port_base::parity o{
                        asio::serial_port_base::parity::none};
                    asio_error_code ec;
                    port->set_option(o, ec);
                    if (ec) {
                        push(L, static_cast<std::error_code>(ec));
                        return lua_error(L);
                    }
                    return 0;
                }
                case LUA_TSTRING:
                    break;
                default:
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }

                auto value = tostringview(L, 3);
                auto o =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_3::in_word_set((value).data(), (value).size()));
                if (!o) {
                    push(L, std::errc::not_supported);
                    return lua_error(L);
                }

                asio_error_code ec;
                port->set_option(asio::serial_port_base::parity{*o}, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}},
      {"stop_bits", 
            [](lua_State* L, asio::serial_port* port) -> int {
                luaL_checktype(L, 3, LUA_TSTRING);

                auto value = tostringview(L, 3);
                auto o =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_4::in_word_set((value).data(), (value).size()));
                if (!o) {
                    push(L, std::errc::not_supported);
                    return lua_error(L);
                }

                asio_error_code ec;
                port->set_option(asio::serial_port_base::stop_bits{*o}, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"baud_rate", 
            [](lua_State* L, asio::serial_port* port) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::serial_port_base::baud_rate o(lua_tointeger(L, 3));
                asio_error_code ec;
                port->set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"flow_control", 
            [](lua_State* L, asio::serial_port* port) -> int {
                switch (lua_type(L, 3)) {
                case LUA_TNIL: {
                    asio::serial_port_base::flow_control o{
                        asio::serial_port_base::flow_control::none};
                    asio_error_code ec;
                    port->set_option(o, ec);
                    if (ec) {
                        push(L, static_cast<std::error_code>(ec));
                        return lua_error(L);
                    }
                    return 0;
                }
                case LUA_TSTRING:
                    break;
                default:
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }

                auto value = tostringview(L, 3);
                auto o =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_2::in_word_set((value).data(), (value).size()));
                if (!o) {
                    push(L, std::errc::not_supported);
                    return lua_error(L);
                }

                asio_error_code ec;
                port->set_option(asio::serial_port_base::flow_control{*o}, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"character_size", 
            [](lua_State* L, asio::serial_port* port) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::serial_port_base::character_size o(lua_tointeger(L, 3));
                asio_error_code ec;
                port->set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 853 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_5  */
/* Computed positions: -k'3' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_5 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 31, duplicates = 0 */

class Perfect_Hash_5
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_5 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_5::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 10, 35,  5, 35, 35,
      35,  0, 35, 25, 35,  5, 35, 35,  0, 35,
       0,  0, 35, 35, 10, 15, 35, 20, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35, 35, 35, 35, 35,
      35, 35, 35, 35, 35, 35
    };
  return len + asso_values[static_cast<unsigned char>(str[2])];
}

const struct word_type_5 *
Perfect_Hash_5::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 17,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 34
    };

  static const struct word_type_5 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"open", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, serial_port_open);
                return 1;
            }},
      {"close", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, serial_port_close);
                return 1;
            }},
      {"cancel", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, serial_port_cancel);
                return 1;
            }},
      {"release", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, serial_port_release);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"",{}},
      {"stop_bits", 
            [](lua_State* L) -> int {
                auto port = static_cast<asio::serial_port*>(
                    lua_touserdata(L, 1));
                asio::serial_port_base::stop_bits o;
                asio_error_code ec;
                port->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                switch (o.value()) {
                case asio::serial_port_base::stop_bits::one:
                    lua_pushliteral(L, "one");
                    break;
                case asio::serial_port_base::stop_bits::onepointfive:
                    lua_pushliteral(L, "one_point_five");
                    break;
                case asio::serial_port_base::stop_bits::two:
                    lua_pushliteral(L, "two");
                }
                return 1;
            }},
      {"send_break", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, serial_port_send_break);
                return 1;
            }},
      {"isatty", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, serial_port_isatty);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"flow_control", 
            [](lua_State* L) -> int {
                auto port = static_cast<asio::serial_port*>(
                    lua_touserdata(L, 1));
                asio::serial_port_base::flow_control o;
                asio_error_code ec;
                port->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                switch (o.value()) {
                case asio::serial_port_base::flow_control::none:
                    lua_pushnil(L);
                    break;
                case asio::serial_port_base::flow_control::software:
                    lua_pushliteral(L, "software");
                    break;
                case asio::serial_port_base::flow_control::hardware:
                    lua_pushliteral(L, "hardware");
                }
                return 1;
            }},
      {"",{}},
      {"read_some", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &serial_port_read_some_key);
                return 1;
            }},
      {"write_some", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &serial_port_write_some_key);
                return 1;
            }},
      {"parity", 
            [](lua_State* L) -> int {
                auto port = static_cast<asio::serial_port*>(
                    lua_touserdata(L, 1));
                asio::serial_port_base::parity o;
                asio_error_code ec;
                port->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                switch (o.value()) {
                case asio::serial_port_base::parity::none:
                    lua_pushnil(L);
                    break;
                case asio::serial_port_base::parity::odd:
                    lua_pushliteral(L, "odd");
                    break;
                case asio::serial_port_base::parity::even:
                    lua_pushliteral(L, "even");
                }
                return 1;
            }},
      {"is_open",  serial_port_is_open},
      {"",{}},
      {"character_size", 
            [](lua_State* L) -> int {
                auto port = static_cast<asio::serial_port*>(
                    lua_touserdata(L, 1));
                asio::serial_port_base::character_size o;
                asio_error_code ec;
                port->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"",{}},
      {"assign", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, serial_port_assign);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"",{}}, {"",{}},
      {"tcsetpgrp", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, serial_port_tcsetpgrp);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"baud_rate", 
            [](lua_State* L) -> int {
                auto port = static_cast<asio::serial_port*>(
                    lua_touserdata(L, 1));
                asio::serial_port_base::baud_rate o;
                asio_error_code ec;
                port->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"tcgetpgrp", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, serial_port_tcgetpgrp);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail
#line 1 "../src/serial_port.cpp"
// Copyright (c) 2022 Vinícius dos Santos Oliveira
// SPDX-License-Identifier: MIT OR BSL-1.0

#line 17 "../src/serial_port.cpp"


namespace emilua {

char serial_port_key;
char serial_port_mt_key;

#line 297 "../src/serial_port.cpp"


static int serial_port_read_some(lua_State* L)
{
    lua_settop(L, 2);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    port->async_read_some(
        asio::buffer(bs->data.get(), bs->size),
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                boost::ignore_unused(buf);
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int serial_port_write_some(lua_State* L)
{
    lua_settop(L, 2);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    if (!port || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    port->async_write_some(
        asio::buffer(bs->data.get(), bs->size),
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                boost::ignore_unused(buf);
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

#line 417 "../src/serial_port.cpp"


static int serial_port_mt_newindex(lua_State* L)
{
    auto port = static_cast<asio::serial_port*>(lua_touserdata(L, 1));
    auto key = tostringview(L, 2);
    return 
#line 1280 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_1::in_word_set((key).data(), (key).size()), [](lua_State* L, asio::serial_port*) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 601 "../src/serial_port.cpp"
(L, port);
}

static int serial_port_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 1293 "libemilua.so.0.11.5.p/bc2_serial_port.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_5::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 802 "../src/serial_port.cpp"
(L);
}

static int serial_port_new(lua_State* L)
{
    int nargs = lua_gettop(L);
    auto& vm_ctx = get_vm_context(L);

    if (nargs == 0) {
        auto port = static_cast<asio::serial_port*>(
            lua_newuserdata(L, sizeof(asio::serial_port))
        );
        rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
        setmetatable(L, -2);
        new (port) asio::serial_port{vm_ctx.strand().context()};
        return 1;
    }

#if BOOST_OS_UNIX
    auto handle = static_cast<file_descriptor_handle*>(lua_touserdata(L, 1));
    if (!handle || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (*handle == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::device_or_resource_busy);
        return lua_error(L);
    }

    auto port = static_cast<asio::serial_port*>(
        lua_newuserdata(L, sizeof(asio::serial_port))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    setmetatable(L, -2);
    new (port) asio::serial_port{vm_ctx.strand().context()};

    lua_pushnil(L);
    setmetatable(L, 1);

    asio_error_code ec;
    port->assign(*handle, ec);
    assert(!ec); boost::ignore_unused(ec);

    return 1;
#else // BOOST_OS_UNIX
    push(L, std::errc::invalid_argument, "arg", 1);
    return lua_error(L);
#endif // BOOST_OS_UNIX
}

#if BOOST_OS_UNIX
static int serial_ptypair(lua_State* L)
{
    auto& vm_ctx = get_vm_context(L);

    int masterfd = posix_openpt(O_RDWR | O_NOCTTY);
    if (masterfd == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }
    BOOST_SCOPE_EXIT_ALL(&) { if (masterfd != -1) { close(masterfd); } };

    if (grantpt(masterfd) == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    if (unlockpt(masterfd) == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    char* slavepath = ptsname(masterfd);
    if (slavepath == NULL) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    int slavefd = open(slavepath, O_RDWR | O_NONBLOCK | O_NOCTTY);
    if (slavefd == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }
    BOOST_SCOPE_EXIT_ALL(&) { if (slavefd != -1) { close(slavefd); } };

    auto master = static_cast<asio::serial_port*>(
        lua_newuserdata(L, sizeof(asio::serial_port))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &serial_port_mt_key);
    setmetatable(L, -2);
    new (master) asio::serial_port{vm_ctx.strand().context()};

    asio_error_code ec;
    master->assign(masterfd, ec);
    assert(!ec); boost::ignore_unused(ec);
    masterfd = -1;

    auto slave = static_cast<file_descriptor_handle*>(
        lua_newuserdata(L, sizeof(file_descriptor_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    setmetatable(L, -2);
    *slave = slavefd;
    slavefd = -1;

    return 2;
}
#endif // BOOST_OS_UNIX

void init_serial_port(lua_State* L)
{
    lua_pushlightuserdata(L, &serial_port_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/2);

        lua_pushliteral(L, "new");
        lua_pushcfunction(L, serial_port_new);
        lua_rawset(L, -3);

#if BOOST_OS_UNIX
        lua_pushliteral(L, "ptypair");
        lua_pushcfunction(L, serial_ptypair);
        lua_rawset(L, -3);
#endif // BOOST_OS_UNIX
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &serial_port_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/4);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "serial_port");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__newindex");
        lua_pushcfunction(L, serial_port_mt_newindex);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, serial_port_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<asio::serial_port>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &serial_port_read_some_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, serial_port_read_some);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &serial_port_write_some_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, serial_port_write_some);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
}

} // namespace emilua
