#line 2 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
#include <cstring>

namespace emilua::gperf::detail { using std::size_t; using std::strcmp; }
#line 6 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"


#line 9 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"

#include <boost/predef/library/std.h>

#include <emilua/file_descriptor.hpp>
#include <emilua/filesystem.hpp>
#include <emilua/windows.hpp>
#include <emilua/system.hpp>
#include <emilua/actor.hpp>
#include <emilua/time.hpp>

#include <boost/scope_exit.hpp>

#if BOOST_OS_UNIX || BOOST_OS_MACOS
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

#if BOOST_OS_LINUX
#include <sys/capability.h>
#include <sys/sysmacros.h>
#endif // BOOST_OS_LINUX

#line 33 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


namespace fs = std::filesystem;

extern unsigned char mode_bytecode[];
extern std::size_t mode_bytecode_size;

char filesystem_key;
char filesystem_path_mt_key;
static char space_info_mt_key;
static char directory_iterator_mt_key;
static char path_ctors_key;
static char clock_ctors_key;
static char mode_key;

struct directory_iterator
{
    template<class... Args>
    directory_iterator(Args&&... args)
        : iterator{std::forward<Args>(args)...}
    {}

    fs::directory_iterator iterator;
    bool increment = false;

    static int next(lua_State* L);
    static int make(lua_State* L);
};

char file_clock_time_point_mt_key;
static char filesystem_path_iterator_mt_key;
static char file_status_mt_key;
static char directory_entry_mt_key;
static char recursive_directory_iterator_mt_key;

using lua_Seconds = std::chrono::duration<lua_Number>;

struct recursive_directory_iterator
{
    template<class... Args>
    recursive_directory_iterator(Args&&... args)
        : iterator{std::forward<Args>(args)...}
    {}

    fs::recursive_directory_iterator iterator;
    bool increment = false;

    static int pop(lua_State* L);
    static int disable_recursion_pending(lua_State* L);
    static int recursion_pending(lua_State* L);
    static int mt_index(lua_State* L);
    static int next(lua_State* L);
    static int make(lua_State* L);
};

} // namespace emilua

#line 92 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


static int path_to_generic(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    try {
        auto ret = path->generic_u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

static int path_iterator(lua_State* L)
{
    static constexpr auto iter = [](lua_State* L) {
        auto path = static_cast<fs::path*>(
            lua_touserdata(L, lua_upvalueindex(1)));
        auto iter = static_cast<fs::path::iterator*>(
            lua_touserdata(L, lua_upvalueindex(2)));

        if (*iter == path->end())
            return 0;

        try {
            auto ret = (*iter)->u8string();
            ++*iter;
            lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
            return 1;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    };

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushvalue(L, 1);
    {
        auto iter = static_cast<fs::path::iterator*>(
            lua_newuserdata(L, sizeof(fs::path::iterator)));
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_iterator_mt_key);
        setmetatable(L, -2);
        new (iter) fs::path::iterator{};
        try {
            *iter = path->begin();
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    }
    lua_pushcclosure(L, iter, 2);
    return 1;
}

static int path_make_preferred(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{*path};
    ret->make_preferred();

    return 1;
}

static int path_remove_filename(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{*path};
    ret->remove_filename();

    return 1;
}

static int path_replace_filename(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::path path2;

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto p = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!p || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *p;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{*path};
    ret->replace_filename(path2);

    return 1;
}

static int path_replace_extension(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::path path2;

    switch (lua_type(L, 2)) {
    case LUA_TNIL:
        break;
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto p = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!p || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *p;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{*path};
    ret->replace_extension(path2);

    return 1;
}

static int path_lexically_normal(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    try {
        *ret = path->lexically_normal();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    return 1;
}

static int path_lexically_relative(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::path path2;

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto p = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!p || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *p;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    try {
        *ret = path->lexically_relative(path2);
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    return 1;
}

static int path_lexically_proximate(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::path path2;

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto p = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!p || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *p;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    try {
        *ret = path->lexically_proximate(path2);
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    return 1;
}

inline int path_root_name(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    try {
        auto ret = path->root_name().u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_root_directory(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    try {
        auto ret = path->root_directory().u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_root_path(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    try {
        *ret = path->root_path();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    return 1;
}

inline int path_relative_path(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    try {
        *ret = path->relative_path();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    return 1;
}

inline int path_parent_path(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    try {
        *ret = path->parent_path();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    return 1;
}

inline int path_filename(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    try {
        auto ret = path->filename().u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_stem(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    try {
        auto ret = path->stem().u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_extension(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    try {
        auto ret = path->extension().u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_empty(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    lua_pushboolean(L, path->empty());
    return 1;
}

inline int path_has_root_path(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_root_path());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_root_name(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_root_name());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_root_directory(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_root_directory());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_relative_path(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_relative_path());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_parent_path(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_parent_path());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_filename(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_filename());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_stem(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_stem());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_has_extension(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->has_extension());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_is_absolute(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->is_absolute());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

inline int path_is_relative(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    try {
        lua_pushboolean(L, path->is_relative());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

} // namespace emilua

#line 816 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_1  */
/* Computed positions: -k'6,$' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_1 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 46, duplicates = 0 */

class Perfect_Hash_1
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_1 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_1::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50,  0, 50,  0,
      15,  0, 50, 50,  5, 15, 50, 50, 20,  0,
       0, 10,  0, 50, 30, 15,  5, 50, 50, 50,
      25, 20, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50, 50, 50, 50, 50,
      50, 50, 50, 50, 50, 50
    };
  unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[static_cast<unsigned char>(str[5])];
      /*FALLTHROUGH*/
      case 5:
      case 4:
        break;
    }
  return hval + asso_values[static_cast<unsigned char>(str[len - 1])];
}

const struct word_type_1 *
Perfect_Hash_1::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 28,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 19,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 49
    };

  static const struct word_type_1 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"stem",  path_stem},
      {"",{}}, {"",{}}, {"",{}},
      {"filename",  path_filename},
      {"root_name",  path_root_name},
      {"to_generic", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_to_generic);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"has_stem",  path_has_stem},
      {"root_path",  path_root_path},
      {"remove_filename", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_remove_filename);
                return 1;
            }},
      {"replace_filename", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_replace_filename);
                return 1;
            }},
      {"replace_extension", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_replace_extension);
                return 1;
            }},
      {"lexically_relative", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_lexically_relative);
                return 1;
            }},
      {"lexically_proximate", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_lexically_proximate);
                return 1;
            }},
      {"has_parent_path",  path_has_parent_path},
      {"parent_path",  path_parent_path},
      {"has_relative_path",  path_has_relative_path},
      {"has_root_name",  path_has_root_name},
      {"extension",  path_extension},
      {"empty",  path_empty},
      {"is_absolute",  path_is_absolute},
      {"has_filename",  path_has_filename},
      {"has_root_path",  path_has_root_path},
      {"make_preferred", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_make_preferred);
                return 1;
            }},
      {"",{}},
      {"is_relative",  path_is_relative},
      {"",{}},
      {"relative_path",  path_relative_path},
      {"",{}}, {"",{}},
      {"lexically_normal", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_lexically_normal);
                return 1;
            }},
      {"",{}},
      {"has_extension",  path_has_extension},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"iterator", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_iterator);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"has_root_directory",  path_has_root_directory},
      {"root_directory",  path_root_directory}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1014 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


static int path_new(lua_State* L)
{
    lua_settop(L, 1);

    auto path = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (path) fs::path{};

    switch (lua_type(L, 1)) {
    case LUA_TNIL:
        break;
    case LUA_TSTRING:
        try {
            *path = fs::path{tou8stringview(L, 1), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    return 1;
}

static int path_from_generic(lua_State* L)
{
    if (lua_type(L, 1) != LUA_TSTRING) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (path) fs::path{};

    try {
        *path = fs::path{tou8stringview(L, 1), fs::path::generic_format};
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

} // namespace emilua

#line 1075 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


static int file_clock_time_point_add(lua_State* L)
{
    lua_settop(L, 2);

    auto tpmem = lua_touserdata(L, 1);
    if (!tpmem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp = aligned_userdata<std::chrono::file_clock::time_point>(tpmem);

    lua_Number secs = luaL_checknumber(L, 2);
    if (std::isnan(secs) || std::isinf(secs)) {
        push(L, std::errc::argument_out_of_domain, "arg", 2);
        return lua_error(L);
    }

    lua_Seconds dur{secs};
    if (
        dur > std::chrono::file_clock::duration::max() ||
        dur < std::chrono::file_clock::duration::min()
    ) {
        push(L, std::errc::value_too_large);
        return lua_error(L);
    }

    try {
        *tp += std::chrono::round<std::chrono::file_clock::duration>(dur);
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
    return 0;
}

static int file_clock_time_point_sub(lua_State* L)
{
    lua_settop(L, 2);

    auto tpmem = lua_touserdata(L, 1);
    if (!tpmem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp = aligned_userdata<std::chrono::file_clock::time_point>(tpmem);

    lua_Number secs = luaL_checknumber(L, 2);
    if (std::isnan(secs) || std::isinf(secs)) {
        push(L, std::errc::argument_out_of_domain, "arg", 2);
        return lua_error(L);
    }

    lua_Seconds dur{secs};
    if (
        dur > std::chrono::file_clock::duration::max() ||
        dur < std::chrono::file_clock::duration::min()
    ) {
        push(L, std::errc::value_too_large);
        return lua_error(L);
    }

    try {
        *tp -= std::chrono::round<std::chrono::file_clock::duration>(dur);
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
    return 0;
}

static int file_clock_time_point_to_system(lua_State* L)
{
    auto tpmem = lua_touserdata(L, 1);
    if (!tpmem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp = aligned_userdata<std::chrono::file_clock::time_point>(tpmem);

    auto ret = static_cast<std::chrono::system_clock::time_point*>(
        lua_newuserdata(L, sizeof(std::chrono::system_clock::time_point))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &system_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (ret) std::chrono::system_clock::time_point{};
#if BOOST_LIB_STD_GNU || BOOST_LIB_STD_CXX
    // current libstdc++ hasn't got clock_cast yet
    //
    // libc++ (on FreeBSD) uses duration<__int128, ratio<1, 1000000000>> for
    // to_sys()'s return value which can't be implicitly converted to
    // duration<long long, std::ratio<1, 1000000>> so we need time_point_cast<>.
    *ret = std::chrono::time_point_cast<std::chrono::system_clock::duration>(
        std::chrono::file_clock::to_sys(*tp));
#else
    *ret = std::chrono::clock_cast<std::chrono::system_clock>(*tp);
#endif
    return 1;
}

inline int file_clock_time_point_seconds_since_epoch(lua_State* L)
{
    auto tp = static_cast<std::chrono::file_clock::time_point*>(
        lua_touserdata(L, 1));
    lua_pushnumber(L, lua_Seconds{tp->time_since_epoch()}.count());
    return 1;
}

inline int file_clock_time_point_seconds_since_unix_epoch(lua_State* L)
{
    auto tp = static_cast<std::chrono::file_clock::time_point*>(
        lua_touserdata(L, 1));
#if BOOST_LIB_STD_GNU || BOOST_LIB_STD_CXX
    // current libstdc++ hasn't got clock_cast yet
    auto tp2 = std::chrono::file_clock::to_sys(*tp);
#else
    std::chrono::system_clock::time_point tp2 =
        std::chrono::clock_cast<std::chrono::system_clock>(*tp);
#endif
    lua_pushnumber(L, lua_Seconds{tp2.time_since_epoch()}.count());
    return 1;
}

} // namespace emilua

#line 1224 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_2  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_2 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 22, duplicates = 0 */

class Perfect_Hash_2
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_2 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_2::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25,  5, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25,  0,  0, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_2 *
Perfect_Hash_2::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 24,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 24
    };

  static const struct word_type_2 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"sub", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_time_point_sub);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"add", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_time_point_add);
                return 1;
            }},
      {"to_system", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_time_point_to_system);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}},
      {"seconds_since_epoch",  file_clock_time_point_seconds_since_epoch},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"seconds_since_unix_epoch", 
            file_clock_time_point_seconds_since_unix_epoch}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1361 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


inline int space_info_capacity(lua_State* L)
{
    auto space = static_cast<fs::space_info*>(lua_touserdata(L, 1));
    lua_pushinteger(L, space->capacity);
    return 1;
}

inline int space_info_free(lua_State* L)
{
    auto space = static_cast<fs::space_info*>(lua_touserdata(L, 1));
    lua_pushinteger(L, space->free);
    return 1;
}

inline int space_info_available(lua_State* L)
{
    auto space = static_cast<fs::space_info*>(lua_touserdata(L, 1));
    lua_pushinteger(L, space->available);
    return 1;
}

} // namespace emilua

#line 1388 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_3  */
/* Computed positions: -k'' */

struct word_type_3 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 6, duplicates = 0 */

class Perfect_Hash_3
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_3 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_3::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_3 *
Perfect_Hash_3::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 9,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 9
    };

  static const struct word_type_3 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"free",  space_info_free},
      {"",{}}, {"",{}}, {"",{}},
      {"capacity",  space_info_capacity},
      {"available",  space_info_available}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1452 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


inline int file_status_type(lua_State* L)
{
    auto st = static_cast<fs::file_status*>(lua_touserdata(L, 1));
    switch (st->type()) {
    case fs::file_type::none:
        lua_pushnil(L);
        return 1;
    case fs::file_type::not_found:
        lua_pushliteral(L, "not_found");
        return 1;
    case fs::file_type::regular:
        lua_pushliteral(L, "regular");
        return 1;
    case fs::file_type::directory:
        lua_pushliteral(L, "directory");
        return 1;
    case fs::file_type::symlink:
        lua_pushliteral(L, "symlink");
        return 1;
    case fs::file_type::block:
        lua_pushliteral(L, "block");
        return 1;
    case fs::file_type::character:
        lua_pushliteral(L, "character");
        return 1;
    case fs::file_type::fifo:
        lua_pushliteral(L, "fifo");
        return 1;
    case fs::file_type::socket:
        lua_pushliteral(L, "socket");
        return 1;
#if BOOST_OS_WINDOWS
    case fs::file_type::junction:
        lua_pushliteral(L, "junction");
        return 1;
#endif // BOOST_OS_WINDOWS
    case fs::file_type::unknown:
        // by avoiding an explicit `default` case we get compiler warnings in
        // case we miss any
        break;
    }

    lua_pushliteral(L, "unknown");
    return 1;
}

inline int file_status_mode(lua_State* L)
{
    auto st = static_cast<fs::file_status*>(lua_touserdata(L, 1));
    if (st->permissions() == fs::perms::unknown) {
        lua_pushliteral(L, "unknown");
        return 1;
    }
    auto p = static_cast<std::underlying_type_t<fs::perms>>(st->permissions());
    lua_pushinteger(L, p);
    return 1;
}

} // namespace emilua

#line 1516 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_4  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_4 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_4
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_4 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_4::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 1,
      2, 2, 2, 2, 2, 2, 0, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_4 *
Perfect_Hash_4::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 4,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_4 wordlist[] =
    {
      {"type",  file_status_type},
      {"mode",  file_status_mode}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1633 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


static int directory_entry_refresh(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));
    if (!entry || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &directory_entry_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    entry->refresh(ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = entry->path();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 0;
}

inline int directory_entry_path(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));
    auto path = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (path) fs::path{};
    *path = entry->path();
    return 1;
}

inline int directory_entry_file_size(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));

    std::error_code ec;
    auto ret = entry->file_size(ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = entry->path();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }
    lua_pushinteger(L, ret);
    return 1;
}

inline int directory_entry_hardlink_count(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));

    std::error_code ec;
    auto ret = entry->hard_link_count(ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = entry->path();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }
    lua_pushinteger(L, ret);
    return 1;
}

inline int directory_entry_last_write_time(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));

    std::error_code ec;
    auto ret = entry->last_write_time(ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = entry->path();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }

    auto tp = aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (tp) std::chrono::file_clock::time_point{ret};
    return 1;
}

inline int directory_entry_status(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));

    std::error_code ec;
    auto ret = entry->status(ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = entry->path();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }

    auto st = static_cast<fs::file_status*>(
        lua_newuserdata(L, sizeof(fs::file_status))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_status_mt_key);
    setmetatable(L, -2);
    new (st) fs::file_status{ret};
    return 1;
}

inline int directory_entry_symlink_status(lua_State* L)
{
    auto entry = static_cast<fs::directory_entry*>(lua_touserdata(L, 1));

    std::error_code ec;
    auto ret = entry->symlink_status(ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = entry->path();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }

    auto st = static_cast<fs::file_status*>(
        lua_newuserdata(L, sizeof(fs::file_status))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_status_mt_key);
    setmetatable(L, -2);
    new (st) fs::file_status{ret};
    return 1;
}

} // namespace emilua

#line 1830 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_5  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_5 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 16, duplicates = 0 */

class Perfect_Hash_5
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_5 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_5::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20,  0, 20,  5, 20, 20, 20,  0, 20,
      20, 20,  0, 20,  0,  0, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20, 20, 20, 20, 20,
      20, 20, 20, 20, 20, 20
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_5 *
Perfect_Hash_5::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 7,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 15,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 19
    };

  static const struct word_type_5 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"path",  directory_entry_path},
      {"",{}},
      {"status",  directory_entry_status},
      {"refresh", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, directory_entry_refresh);
                return 1;
            }},
      {"",{}},
      {"file_size",  directory_entry_file_size},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"symlink_status",  directory_entry_symlink_status},
      {"last_write_time",  directory_entry_last_write_time},
      {"",{}}, {"",{}}, {"",{}},
      {"hardlink_count",  directory_entry_hardlink_count}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1960 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


int recursive_directory_iterator::pop(lua_State* L)
{
    auto self = static_cast<recursive_directory_iterator*>(
        lua_touserdata(L, 1));
    if (!self || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &recursive_directory_iterator_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    self->iterator.pop(ec);
    if (ec) {
        push(L, ec);
        return lua_error(L);
    }
    return 0;
}

int recursive_directory_iterator::disable_recursion_pending(lua_State* L)
{
    auto self = static_cast<recursive_directory_iterator*>(
        lua_touserdata(L, 1));
    if (!self || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &recursive_directory_iterator_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (self->iterator == fs::recursive_directory_iterator{}) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    try {
        self->iterator.disable_recursion_pending();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
    return 0;
}

int recursive_directory_iterator::recursion_pending(lua_State* L)
{
    auto self = static_cast<recursive_directory_iterator*>(
        lua_touserdata(L, 1));
    if (self->iterator == fs::recursive_directory_iterator{}) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, self->iterator.recursion_pending());
    return 1;
}

} // namespace emilua

#line 2033 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_6  */
/* Computed positions: -k'' */

struct word_type_6 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 23, duplicates = 0 */

class Perfect_Hash_6
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_6 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_6::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_6 *
Perfect_Hash_6::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 25,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 25
    };

  static const struct word_type_6 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"pop", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, recursive_directory_iterator::pop);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"recursion_pending", 
            recursive_directory_iterator::recursion_pending},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"disable_recursion_pending", 
            [](lua_State* L) -> int {
                lua_pushcfunction(
                    L, recursive_directory_iterator::disable_recursion_pending);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2111 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


static int file_clock_now(lua_State* L)
{
    auto tp = aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (tp) std::chrono::file_clock::time_point{};
    *tp = std::chrono::file_clock::now();
    return 1;
}

static int file_clock_epoch(lua_State* L)
{
    auto tp = aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (tp) std::chrono::file_clock::time_point{};
    return 1;
}

static int file_clock_unix_epoch(lua_State* L)
{
    auto tp = aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (tp) std::chrono::file_clock::time_point{};
#if BOOST_LIB_STD_GNU || BOOST_LIB_STD_CXX
    // current libstdc++ hasn't got clock_cast yet
    *tp = std::chrono::file_clock::from_sys(
        std::chrono::system_clock::time_point{});
#else
    *tp = std::chrono::clock_cast<std::chrono::file_clock>(
        std::chrono::system_clock::time_point{});
#endif
    return 1;
}

static int file_clock_from_system(lua_State* L)
{
    auto tp = static_cast<std::chrono::system_clock::time_point*>(
        lua_touserdata(L, 1));
    if (!tp || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &system_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (ret) std::chrono::file_clock::time_point{};
#if BOOST_LIB_STD_GNU || BOOST_LIB_STD_CXX
    // current libstdc++ hasn't got clock_cast yet
    *ret = std::chrono::file_clock::from_sys(*tp);
#else
    *ret = std::chrono::clock_cast<std::chrono::file_clock>(*tp);
#endif
    return 1;
}

} // namespace emilua

#line 2180 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
#if BOOST_OS_UNIX || BOOST_OS_MACOS
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_7  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_7 { const char* name; int action; };
/* maximum key range = 21, duplicates = 0 */

class Perfect_Hash_7
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_7 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_7::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25,  5, 25,  0,
      15, 10, 25, 25, 25, 25, 25, 25, 25, 25,
       0, 25,  0, 25,  5,  0,  0, 25, 25,  0,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_7 *
Perfect_Hash_7::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 11,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 17,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 24
    };

  static const struct word_type_7 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"path", 
#ifdef O_PATH
                O_PATH
#else
                0
#endif // O_PATH
            },
      {"",{}},
      {"create",  O_CREAT},
      {"",{}},
      {"truncate",  O_TRUNC},
      {"no_follow",  O_NOFOLLOW},
      {"write_only",  O_WRONLY},
      {"append",  O_APPEND},
      {"",{}}, {"",{}},
      {"read_only",  O_RDONLY},
      {"read_write",  O_RDWR},
      {"",{}},
      {"sync_all_on_write",  O_SYNC},
      {"",{}},
      {"exclusive",  O_EXCL},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"directory",  O_DIRECTORY}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

#line 2320 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua {


static int absolute(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    std::error_code ec;
    *ret = fs::absolute(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 1;
}

static int canonical(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    std::error_code ec;
    *ret = fs::canonical(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 1;
}

static int weakly_canonical(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    std::error_code ec;
    *ret = fs::weakly_canonical(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 1;
}

static int relative(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::path path2;
    switch (lua_type(L, 2)) {
    case LUA_TNIL: {
        std::error_code ec;
        path2 = fs::current_path(ec);
        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
        break;
    }
    case LUA_TUSERDATA: {
        auto p = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!p || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *p;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    std::error_code ec;
    *ret = fs::relative(*path, path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        *ret = path2;
        lua_pushvalue(L, -3);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 1;
}

static int proximate(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::path path2;
    switch (lua_type(L, 2)) {
    case LUA_TNIL: {
        std::error_code ec;
        path2 = fs::current_path(ec);
        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
        break;
    }
    case LUA_TUSERDATA: {
        auto p = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!p || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *p;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    std::error_code ec;
    *ret = fs::proximate(*path, path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        *ret = path2;
        lua_pushvalue(L, -3);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 1;
}

static int last_write_time(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    case LUA_TNIL: {
        auto tp =
            aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
        rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
        setmetatable(L, -2);
        new (tp) std::chrono::file_clock::time_point{};
        std::error_code ec;
        *tp = fs::last_write_time(*path, ec);
        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
        return 1;
    }
    case LUA_TUSERDATA: {
        auto tpmem = lua_touserdata(L, 2);
        if (!tpmem || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        auto tp = aligned_userdata<std::chrono::file_clock::time_point>(tpmem);

        std::error_code ec;
        fs::last_write_time(*path, *tp, ec);
        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
        return 0;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
}

static int copy(lua_State* L)
{
    lua_settop(L, 3);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    fs::copy_options options = fs::copy_options::none;
    switch (lua_type(L, 3)) {
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        lua_getfield(L, 3, "existing");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TSTRING: {
            auto v = tostringview(L);
            if (v == "skip") {
                options |= fs::copy_options::skip_existing;
            } else if (v == "overwrite") {
                options |= fs::copy_options::overwrite_existing;
            } else if (v == "update") {
                options |= fs::copy_options::update_existing;
            } else {
                push(L, std::errc::invalid_argument, "arg", "existing");
                return lua_error(L);
            }
            break;
        }
        default:
            push(L, std::errc::invalid_argument, "arg", "existing");
            return lua_error(L);
        }

        lua_getfield(L, 3, "recursive");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TBOOLEAN:
            if (lua_toboolean(L, -1))
                options |= fs::copy_options::recursive;
            break;
        default:
            push(L, std::errc::invalid_argument, "arg", "recursive");
            return lua_error(L);
        }

        lua_getfield(L, 3, "symlinks");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TSTRING: {
            auto v = tostringview(L);
            if (v == "copy") {
                options |= fs::copy_options::copy_symlinks;
            } else if (v == "skip") {
                options |= fs::copy_options::skip_symlinks;
            } else {
                push(L, std::errc::invalid_argument, "arg", "symlinks");
                return lua_error(L);
            }
            break;
        }
        default:
            push(L, std::errc::invalid_argument, "arg", "symlinks");
            return lua_error(L);
        }

        lua_getfield(L, 3, "copy");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TSTRING: {
            auto v = tostringview(L);
            if (v == "directories_only") {
                options |= fs::copy_options::directories_only;
            } else if (v == "create_symlinks") {
                options |= fs::copy_options::create_symlinks;
            } else if (v == "create_hardlinks") {
                options |= fs::copy_options::create_hard_links;
            } else {
                push(L, std::errc::invalid_argument, "arg", "copy");
                return lua_error(L);
            }
            break;
        }
        default:
            push(L, std::errc::invalid_argument, "arg", "copy");
            return lua_error(L);
        }
        break;
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    try {
        fs::copy(*path, *path2, options);
        return 0;
    } catch (const fs::filesystem_error& e) {
        push(L, e.code());

        lua_pushliteral(L, "path1");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = e.path1();
        }
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        {
            auto path = static_cast<fs::path*>(
                lua_newuserdata(L, sizeof(fs::path)));
            rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
            setmetatable(L, -2);
            new (path) fs::path{};
            *path = e.path2();
        }
        lua_rawset(L, -3);

        return lua_error(L);
    }
}

static int copy_file(lua_State* L)
{
    lua_settop(L, 3);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    fs::copy_options options = fs::copy_options::none;
    switch (lua_type(L, 3)) {
    case LUA_TNIL:
        break;
    case LUA_TSTRING: {
        auto v = tostringview(L, 3);
        if (v == "skip") {
            options |= fs::copy_options::skip_existing;
        } else if (v == "overwrite") {
            options |= fs::copy_options::overwrite_existing;
        } else if (v == "update") {
            options |= fs::copy_options::update_existing;
        } else {
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        }
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::copy_file(*path, *path2, options, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    lua_pushboolean(L, ret);
    return 1;
}

static int copy_symlink(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    std::error_code ec;
    fs::copy_symlink(*path, *path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 0;
}

static int create_directory(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    case LUA_TNIL: {
        std::error_code ec;
        bool ret = fs::create_directory(*path, ec);
        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
        lua_pushboolean(L, ret);
        return 1;
    }
    case LUA_TUSERDATA: {
        auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!path2 || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        std::error_code ec;
        bool ret = fs::create_directory(*path, *path2, ec);
        if (ec) {
            push(L, ec);

            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);

            lua_pushliteral(L, "path2");
            lua_pushvalue(L, 2);
            lua_rawset(L, -3);

            return lua_error(L);
        }
        lua_pushboolean(L, ret);
        return 1;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
}

static int create_directories(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::create_directories(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    lua_pushboolean(L, ret);
    return 1;
}

static int create_hardlink(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    std::error_code ec;
    fs::create_hard_link(*path, *path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 0;
}

static int create_symlink(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    std::error_code ec;
    fs::create_symlink(*path, *path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 0;
}

static int create_directory_symlink(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    std::error_code ec;
    fs::create_directory_symlink(*path, *path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 0;
}

#if BOOST_OS_UNIX || BOOST_OS_MACOS
static int emilua_open(lua_State* L)
{
    lua_settop(L, 3);
    luaL_checktype(L, 2, LUA_TTABLE);

    auto path = static_cast<std::filesystem::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    int flags = 0;
#ifdef O_NOCTTY
    flags = O_NOCTTY;
#endif // O_NOCTTY
    for (int i = 1 ;; ++i) {
        lua_rawgeti(L, 2, i);
        switch (lua_type(L, -1)) {
        default:
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        case LUA_TNIL:
            lua_pop(L, 1);
            goto end_for;
        case LUA_TSTRING:
            break;
        }

        auto s = tostringview(L);
        lua_pop(L, 1);
        auto f =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_7::in_word_set((s).data(), (s).size()));
        if (f) {
            flags |= *f;
        } else if (s == "temporary") {
#ifdef O_TMPFILE
            flags |= O_TMPFILE;
#else
            push(L, std::errc::not_supported, "arg", 2);
            return lua_error(L);
#endif // defined(O_TMPFILE)
        } else {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
    }
 end_for:

    bool has_mode =
        ((flags & O_CREAT) == O_CREAT) ||
#ifdef O_TMPFILE
        ((flags & O_TMPFILE) == O_TMPFILE) ||
#endif // defined(O_TMPFILE)
        false;

    mode_t mode;
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        if (has_mode) {
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        }
        break;
    case LUA_TNUMBER:
        if (!has_mode) {
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        }
        mode = lua_tointeger(L, 3);
        break;
    }

    int res;
    if (has_mode) {
        res = open(path->c_str(), flags, mode);
    } else {
        res = open(path->c_str(), flags);
    }
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    int rawfd = res;
    BOOST_SCOPE_EXIT_ALL(&) {
        if (rawfd != INVALID_FILE_DESCRIPTOR) {
            int res = close(rawfd);
            boost::ignore_unused(res);
        }
    };

    auto handle = static_cast<file_descriptor_handle*>(
        lua_newuserdata(L, sizeof(file_descriptor_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    setmetatable(L, -2);

    *handle = rawfd;
    rawfd = INVALID_FILE_DESCRIPTOR;
    return 1;
}

static int mkdir(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    mode_t mode = luaL_checkinteger(L, 2);

    if (::mkdir(path->c_str(), mode) == -1) {
        push(L, std::error_code{errno, std::system_category()});

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        return lua_error(L);
    }

    return 0;
}

static int mkfifo(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    mode_t mode = luaL_checkinteger(L, 2);

    if (::mkfifo(path->c_str(), mode) == -1) {
        push(L, std::error_code{errno, std::system_category()});

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        return lua_error(L);
    }

    return 0;
}

static int mknod(lua_State* L)
{
    lua_settop(L, 4);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    mode_t mode = luaL_checkinteger(L, 2);
    dev_t dev = luaL_checkinteger(L, 3);

    switch (lua_type(L, 4)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 4);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TSTRING: {
        auto type = tostringview(L, 4);
        if (type == "character") {
            mode |= S_IFCHR;
        } else if (type == "block") {
            mode |= S_IFBLK;
        } else {
            push(L, std::errc::invalid_argument, "arg", 4);
            return lua_error(L);
        }
        break;
    }
    }

    if (::mknod(path->c_str(), mode, dev) == -1) {
        push(L, std::error_code{errno, std::system_category()});

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        return lua_error(L);
    }

    return 0;
}

static int fs_makedev(lua_State* L)
{
    int major = luaL_checkinteger(L, 1);
    int minor = luaL_checkinteger(L, 2);

    lua_pushinteger(L, makedev(major, minor));
    return 1;
}

static int fs_dev_major(lua_State* L)
{
    int dev = luaL_checkinteger(L, 1);
    lua_pushinteger(L, major(dev));
    return 1;
}

static int fs_dev_minor(lua_State* L)
{
    int dev = luaL_checkinteger(L, 1);
    lua_pushinteger(L, minor(dev));
    return 1;
}
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

static int equivalent(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::equivalent(*path, *path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    lua_pushboolean(L, ret);
    return 1;
}

static int file_size(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    auto ret = fs::file_size(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    lua_pushinteger(L, ret);
    return 1;
}

static int hardlink_count(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    auto ret = fs::hard_link_count(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    lua_pushinteger(L, ret);
    return 1;
}

#if BOOST_OS_UNIX || BOOST_OS_MACOS
static int chown(lua_State* L)
{
    lua_settop(L, 3);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    int res = ::chown(
        path->string().data(), luaL_checkinteger(L, 2),
        luaL_checkinteger(L, 3));
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 0;
}

static int lchown(lua_State* L)
{
    lua_settop(L, 3);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    int res = ::lchown(
        path->string().data(), luaL_checkinteger(L, 2),
        luaL_checkinteger(L, 3));
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 0;
}
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

static int chmod(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    fs::permissions(*path, static_cast<fs::perms>(luaL_checkinteger(L, 2)), ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 0;
}

static int lchmod(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    fs::permissions(
        *path, static_cast<fs::perms>(luaL_checkinteger(L, 2)),
        fs::perm_options::replace | fs::perm_options::nofollow, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 0;
}

static int read_symlink(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (ret) fs::path{};

    std::error_code ec;
    *ret = fs::read_symlink(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 1;
}

static int remove(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::remove(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    lua_pushboolean(L, ret);
    return 1;
}

static int remove_all(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    auto ret = fs::remove_all(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    lua_pushinteger(L, ret);
    return 1;
}

static int rename(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    if (!path2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    std::error_code ec;
    fs::rename(*path, *path2, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "path2");
        lua_pushvalue(L, 2);
        lua_rawset(L, -3);

        return lua_error(L);
    }
    return 0;
}

static int resize_file(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    fs::resize_file(*path, luaL_checkinteger(L, 2), ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 0;
}

static int is_empty(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_empty(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    lua_pushboolean(L, ret);
    return 1;
}

static int current_working_directory(lua_State* L)
{
    lua_settop(L, 1);

    if (lua_isnil(L, 1)) {
        auto path = static_cast<fs::path*>(
            lua_newuserdata(L, sizeof(fs::path)));
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        setmetatable(L, -2);
        new (path) fs::path{};

        std::error_code ec;
        *path = fs::current_path(ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }
        return 1;
    }

    auto& vm_ctx = get_vm_context(L);
    if (!vm_ctx.is_master()) {
        // we intentionally leave "path1" out as the path has nothing to do with
        // the failure here
        push(L, std::errc::operation_not_permitted);
        return lua_error(L);
    }

    void* arg1 = lua_touserdata(L, 1);
    fs::path* path = nullptr;
#if BOOST_OS_UNIX || BOOST_OS_MACOS
    int dirfd = -1;
    BOOST_SCOPE_EXIT_ALL(&) { if (path && dirfd != -1) close(dirfd); };
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

    if (!arg1 || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        assert(!path);
#if BOOST_OS_UNIX || BOOST_OS_MACOS
        rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
        if (!lua_rawequal(L, -1, -3)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        dirfd = *static_cast<file_descriptor_handle*>(arg1);
        if (dirfd == INVALID_FILE_DESCRIPTOR) {
            push(L, std::errc::device_or_resource_busy, "arg", 1);
            return lua_error(L);
        }
#else
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
    } else {
        path = static_cast<fs::path*>(arg1);
    }
#if BOOST_OS_UNIX || BOOST_OS_MACOS
    assert(path || dirfd != -1);
#else
    assert(path);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

#if BOOST_OS_UNIX || BOOST_OS_MACOS
    if (vm_ctx.appctx.ipc_actor_service_sockfd != -1) {
        int channel[2] = { -1, -1 };
        BOOST_SCOPE_EXIT_ALL(&) {
            if (channel[0] != -1) close(channel[0]);
            if (channel[1] != -1) close(channel[1]);
        };

        if (pipe(channel) != 0) {
            push(L, std::error_code{errno, std::system_category()});
            return lua_error(L);
        }

        if (dirfd == -1) {
            dirfd = open(path->string().data(), O_RDONLY | O_DIRECTORY);

            if (dirfd == -1) {
                push(L, std::error_code{errno, std::system_category()});
                lua_pushliteral(L, "path1");
                lua_pushvalue(L, 1);
                lua_rawset(L, -3);
                return lua_error(L);
            }
        }

        if (fchdir(dirfd) == -1) {
            push(L, std::error_code{errno, std::system_category()});
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }

        ipc_actor_start_vm_request request;
        std::memset(&request, 0, sizeof(request));
        request.type = ipc_actor_start_vm_request::CHDIR;

        struct msghdr msg;
        std::memset(&msg, 0, sizeof(msg));

        struct iovec iov;
        iov.iov_base = &request;
        iov.iov_len = sizeof(request);
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;

        alignas(cmsghdr) char cmsgbuf[CMSG_SPACE(sizeof(int) * 2)];
        msg.msg_control = cmsgbuf;
        msg.msg_controllen = CMSG_SPACE(sizeof(int) * 2);

        struct cmsghdr* cmsg = CMSG_FIRSTHDR(&msg);
        cmsg->cmsg_level = SOL_SOCKET;
        cmsg->cmsg_type = SCM_RIGHTS;
        cmsg->cmsg_len = CMSG_LEN(sizeof(int) * 2);

        {
            char* begin = (char*)CMSG_DATA(cmsg);
            char* it = begin;

            std::memcpy(it, &channel[1], sizeof(int));
            it += sizeof(int);

            std::memcpy(it, &dirfd, sizeof(int));
        }

        sendmsg(vm_ctx.appctx.ipc_actor_service_sockfd, &msg, MSG_NOSIGNAL);
        close(channel[1]);
        channel[1] = -1;

        char buf[1];
        auto nread = read(channel[0], &buf, 1);
        if (nread == -1 || nread == 0) {
            // as described in <https://ewontfix.com/17/> the only safe answer
            // is to SIGKILL when we cannot guarantee atomicity of failure
            std::exit(1);
        }
    } else {
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

        std::error_code ec;
        fs::current_path(*path, ec);
        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }

#if BOOST_OS_UNIX || BOOST_OS_MACOS
    }
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

    return 0;
}

#if BOOST_OS_UNIX || BOOST_OS_MACOS
static int chroot(lua_State* L)
{
    lua_settop(L, 1);

    auto& vm_ctx = get_vm_context(L);
    if (!vm_ctx.is_master()) {
        // we intentionally leave "path1" out as the path has nothing to do with
        // the failure here
        push(L, std::errc::operation_not_permitted);
        return lua_error(L);
    }

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    int channel[2] = { -1, -1 };
    BOOST_SCOPE_EXIT_ALL(&) {
        if (channel[0] != -1) close(channel[0]);
        if (channel[1] != -1) close(channel[1]);
    };

    int mfd = -1;
    BOOST_SCOPE_EXIT_ALL(&) { if (mfd != -1) close(mfd); };

    std::string::size_type mfd_size;

    if (vm_ctx.appctx.ipc_actor_service_sockfd != -1) {
        int res = pipe(channel);
        if (res != 0) {
            push(L, std::error_code{errno, std::system_category()});
            return lua_error(L);
        }

        auto as_str = path->string();
        mfd_size = as_str.size() + 1; //< include nul terminator

#if BOOST_OS_MACOS
        {
            char name[] = "emilua.XXXXXX\0";
            mfd = mkstemp(name);
            if (mfd != -1) {
                unlink(name);
            }
        }
#else
        mfd = memfd_create("emilua/chroot", /*flags=*/0);
#endif // BOOST_OS_MACOS
        if (mfd == -1) {
            push(L, std::error_code{errno, std::system_category()});
            return lua_error(L);
        }

        if (ftruncate(mfd, mfd_size) == -1) {
            push(L, std::error_code{errno, std::system_category()});
            return lua_error(L);
        }

        write(mfd, as_str.data(), mfd_size);
    }

    int res = ::chroot(path->string().data());
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    if (vm_ctx.appctx.ipc_actor_service_sockfd != -1) {
        ipc_actor_start_vm_request request;
        std::memset(&request, 0, sizeof(request));
        request.type = ipc_actor_start_vm_request::CHROOT;
        request.chroot_mfd_size = mfd_size;

        struct msghdr msg;
        std::memset(&msg, 0, sizeof(msg));

        struct iovec iov;
        iov.iov_base = &request;
        iov.iov_len = sizeof(request);
        msg.msg_iov = &iov;
        msg.msg_iovlen = 1;

        alignas(cmsghdr) char cmsgbuf[CMSG_SPACE(sizeof(int) * 2)];
        msg.msg_control = cmsgbuf;
        msg.msg_controllen = CMSG_SPACE(sizeof(int) * 2);

        struct cmsghdr* cmsg = CMSG_FIRSTHDR(&msg);
        cmsg->cmsg_level = SOL_SOCKET;
        cmsg->cmsg_type = SCM_RIGHTS;
        cmsg->cmsg_len = CMSG_LEN(sizeof(int) * 2);

        {
            char* begin = (char*)CMSG_DATA(cmsg);
            char* it = begin;

            std::memcpy(it, &channel[1], sizeof(int));
            it += sizeof(int);

            std::memcpy(it, &mfd, sizeof(int));
        }

        sendmsg(vm_ctx.appctx.ipc_actor_service_sockfd, &msg, MSG_NOSIGNAL);
        close(channel[1]);
        channel[1] = -1;

        char buf[1];
        auto nread = read(channel[0], &buf, 1);
        if (nread == -1 || nread == 0) {
            // as described in <https://ewontfix.com/17/> the only safe answer
            // is to SIGKILL when we cannot guarantee atomicity of failure
            std::exit(1);
        }
    }

    return 0;
}
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

static int exists(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::exists(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_block_device(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_block_file(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_character_device(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_character_file(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_directory(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_directory(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_fifo(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_fifo(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_other(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_other(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_regular_file(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_regular_file(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_socket(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_socket(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int is_symlink(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    bool ret = fs::is_symlink(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }

    lua_pushboolean(L, ret ? 1 : 0);
    return 1;
}

static int space(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::error_code ec;
    auto ret = fs::space(*path, ec);
    if (ec) {
        push(L, ec);

        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        return lua_error(L);
    }

    auto space = static_cast<fs::space_info*>(
        lua_newuserdata(L, sizeof(fs::space_info))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &space_info_mt_key);
    setmetatable(L, -2);
    new (space) fs::space_info{ret};
    return 1;
}

static int status(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::file_status*>(
        lua_newuserdata(L, sizeof(fs::file_status))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_status_mt_key);
    setmetatable(L, -2);
    new (ret) fs::file_status{};

    std::error_code ec;
    *ret = fs::status(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 1;
}

static int symlink_status(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<fs::file_status*>(
        lua_newuserdata(L, sizeof(fs::file_status))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_status_mt_key);
    setmetatable(L, -2);
    new (ret) fs::file_status{};

    std::error_code ec;
    *ret = fs::symlink_status(*path, ec);
    if (ec) {
        push(L, ec);
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);
        return lua_error(L);
    }
    return 1;
}

static int temp_directory_path(lua_State* L)
{
    lua_settop(L, 0);

    auto path = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (path) fs::path{};

    try {
        *path = fs::temp_directory_path();
        return 1;
    } catch (const fs::filesystem_error& e) {
        push(L, e.code());

        *path = e.path1();
        lua_pushliteral(L, "path1");
        lua_pushvalue(L, 1);
        lua_rawset(L, -3);

        return lua_error(L);
    }
}

// Windows' _umask() is nonsense and useless. umask is an UNIX concept. The
// attempt made to create a version for Windows under the same interface was
// lame and useless. Windows execution model has no space for UNIX's umask.
#if BOOST_OS_UNIX || BOOST_OS_MACOS
static int filesystem_umask(lua_State* L)
{
    auto& vm_ctx = get_vm_context(L);
    if (!vm_ctx.is_master()) {
        push(L, std::errc::operation_not_permitted);
        return lua_error(L);
    }

    mode_t mask = luaL_checkinteger(L, 1);
    mode_t res = umask(mask);
    lua_pushinteger(L, res);
    return 1;
}
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS

#if BOOST_OS_LINUX
static int filesystem_cap_get_file(lua_State* L)
{
    auto path = static_cast<std::filesystem::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::string path2;
    try {
        path2 = path->string();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    cap_t caps = cap_get_file(path2.c_str());
    if (caps == NULL) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }
    BOOST_SCOPE_EXIT_ALL(&) {
        if (caps != NULL)
            cap_free(caps);
    };

    auto& caps2 = *static_cast<cap_t*>(lua_newuserdata(L, sizeof(cap_t)));
    rawgetp(L, LUA_REGISTRYINDEX, &linux_capabilities_mt_key);
    setmetatable(L, -2);
    caps2 = caps;
    caps = NULL;

    return 1;
}

static int filesystem_cap_set_file(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<std::filesystem::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    std::string path2;
    try {
        path2 = path->string();
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }

    auto caps = static_cast<cap_t*>(lua_touserdata(L, 2));
    if (!caps || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &linux_capabilities_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    if (cap_set_file(path2.c_str(), *caps) == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }
    return 0;
}
#endif // BOOST_OS_LINUX

} // namespace emilua

#line 4501 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_8  */
/* Computed positions: -k'' */

struct word_type_8 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 17, duplicates = 0 */

class Perfect_Hash_8
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_8 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_8::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_8 *
Perfect_Hash_8::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 19,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 19
    };

  static const struct word_type_8 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"new", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_new);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}},
      {"from_generic", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, path_from_generic);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"preferred_separator", 
            [](lua_State* L) -> int {
                fs::path::value_type sep = fs::path::preferred_separator;
                push(L, narrow_on_windows(&sep, 1));
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 4580 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_9  */
/* Computed positions: -k'' */

struct word_type_9 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 9, duplicates = 0 */

class Perfect_Hash_9
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_9 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_9::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_9 *
Perfect_Hash_9::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 3,
      MAX_WORD_LENGTH = 11,
      MIN_HASH_VALUE = 3,
      MAX_HASH_VALUE = 11
    };

  static const struct word_type_9 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}},
      {"now", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_now);
                return 1;
            }},
      {"",{}},
      {"epoch", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_epoch);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"unix_epoch", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_unix_epoch);
                return 1;
            }},
      {"from_system", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_clock_from_system);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 4662 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_10  */
/* Computed positions: -k'1,5-6' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_10 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 103, duplicates = 0 */

class Perfect_Hash_10
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_10 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_10::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107,  20, 107,  55, 107,   0,
       15,   0,  40,  50,   5,   0, 107,  46,   0,   5,
       30,  55,  30, 107,  20,   5,   0,   5,  10,  10,
      107,  45,  10, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107, 107, 107, 107, 107,
      107, 107, 107, 107, 107, 107
    };
  unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[static_cast<unsigned char>(str[5])];
      /*FALLTHROUGH*/
      case 5:
        hval += asso_values[static_cast<unsigned char>(str[4])];
      /*FALLTHROUGH*/
      case 4:
      case 3:
      case 2:
      case 1:
        hval += asso_values[static_cast<unsigned char>(str[0])];
        break;
    }
  return hval;
}

const struct word_type_10 *
Perfect_Hash_10::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 57,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 28,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 106
    };

  static const struct word_type_10 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"copy", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, copy);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"mode", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &mode_key);
                return 1;
            }},
      {"space", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, space);
                return 1;
            }},
      {"exists", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, exists);
                return 1;
            }},
      {"",{}},
      {"is_other", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_other);
                return 1;
            }},
      {"create_symlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, create_symlink);
                return 1;
            }},
      {"create_hardlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, create_hardlink);
                return 1;
            }},
      {"create_directory", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, create_directory);
                return 1;
            }},
      {"cap_set_file", 
            [](lua_State* L) -> int {
#if BOOST_OS_LINUX
                lua_pushcfunction(L, filesystem_cap_set_file);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_LINUX
                return 1;
            }},
      {"create_directories", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, create_directories);
                return 1;
            }},
      {"hardlink_count", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, hardlink_count);
                return 1;
            }},
      {"chmod", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, chmod);
                return 1;
            }},
      {"status", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, status);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"create_directory_symlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, create_directory_symlink);
                return 1;
            }},
      {"mknod", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, mknod);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"",{}},
      {"makedev", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, fs_makedev);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"relative", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, relative);
                return 1;
            }},
      {"dev_minor", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, fs_dev_minor);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"mkdir", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, mkdir);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"rename", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, rename);
                return 1;
            }},
      {"is_directory", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_directory);
                return 1;
            }},
      {"directory_iterator", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, directory_iterator::make);
                return 1;
            }},
      {"path", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &path_ctors_key);
                return 1;
            }},
      {"chown", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, chown);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"remove", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, remove);
                return 1;
            }},
      {"copy_symlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, copy_symlink);
                return 1;
            }},
      {"",{}},
      {"canonical", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, canonical);
                return 1;
            }},
      {"remove_all", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, remove_all);
                return 1;
            }},
      {"resize_file", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, resize_file);
                return 1;
            }},
      {"",{}},
      {"is_empty", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_empty);
                return 1;
            }},
      {"proximate", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, proximate);
                return 1;
            }},
      {"last_write_time", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, last_write_time);
                return 1;
            }},
      {"lchown", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, lchown);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"is_fifo", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_fifo);
                return 1;
            }},
      {"",{}},
      {"symlink_status", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, symlink_status);
                return 1;
            }},
      {"",{}},
      {"clock", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &clock_ctors_key);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"temp_directory_path", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, temp_directory_path);
                return 1;
            }},
      {"current_working_directory", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, current_working_directory);
                return 1;
            }},
      {"umask", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, filesystem_umask);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"read_symlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, read_symlink);
                return 1;
            }},
      {"",{}},
      {"open", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, emilua_open);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"is_symlink", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_symlink);
                return 1;
            }},
      {"chroot", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, chroot);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"cap_get_file", 
            [](lua_State* L) -> int {
#if BOOST_OS_LINUX
                lua_pushcfunction(L, filesystem_cap_get_file);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_LINUX
                return 1;
            }},
      {"",{}},
      {"is_socket", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_socket);
                return 1;
            }},
      {"is_regular_file", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_regular_file);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"absolute", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, absolute);
                return 1;
            }},
      {"copy_file", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, copy_file);
                return 1;
            }},
      {"is_block_device", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_block_device);
                return 1;
            }},
      {"weakly_canonical", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, weakly_canonical);
                return 1;
            }},
      {"",{}},
      {"recursive_directory_iterator", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, recursive_directory_iterator::make);
                return 1;
            }},
      {"file_size", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, file_size);
                return 1;
            }},
      {"equivalent", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, equivalent);
                return 1;
            }},
      {"lchmod", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, lchmod);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"is_character_device", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, is_character_device);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"dev_major", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, fs_dev_major);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}},
      {"mkfifo", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX || BOOST_OS_MACOS
                lua_pushcfunction(L, mkfifo);
#else
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX || BOOST_OS_MACOS
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail
#line 1 "../src/filesystem.cpp"
// Copyright (c) 2023 Vinícius dos Santos Oliveira
// SPDX-License-Identifier: MIT OR BSL-1.0

#line 26 "../src/filesystem.cpp"


namespace emilua {

#line 84 "../src/filesystem.cpp"


#line 805 "../src/filesystem.cpp"


static int path_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 5168 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_1::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 889 "../src/filesystem.cpp"
(L);
}

static int path_mt_tostring(lua_State* L)
{
    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));

    try {
        auto ret = path->u8string();
        lua_pushlstring(L, reinterpret_cast<char*>(ret.data()), ret.size());
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

static int path_mt_eq(lua_State* L)
{
    auto path1 = static_cast<fs::path*>(lua_touserdata(L, 1));
    auto path2 = static_cast<fs::path*>(lua_touserdata(L, 2));
    lua_pushboolean(L, *path1 == *path2);
    return 1;
}

static int path_mt_lt(lua_State* L)
{
    fs::path path1, path2;

    switch (lua_type(L, 1)) {
    case LUA_TSTRING:
        try {
            path1 = fs::path{tou8stringview(L, 1), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
        if (!path || !lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        path1 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!path || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    lua_pushboolean(L, path1 < path2);
    return 1;
}

static int path_mt_le(lua_State* L)
{
    fs::path path1, path2;

    switch (lua_type(L, 1)) {
    case LUA_TSTRING:
        try {
            path1 = fs::path{tou8stringview(L, 1), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
        if (!path || !lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        path1 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!path || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    lua_pushboolean(L, path1 <= path2);
    return 1;
}

static int path_mt_div(lua_State* L)
{
    lua_settop(L, 2);

    fs::path path1, path2;

    switch (lua_type(L, 1)) {
    case LUA_TSTRING:
        try {
            path1 = fs::path{tou8stringview(L, 1), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
        if (!path || !lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        path1 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!path || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto path = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (path) fs::path{};
    *path = path1 / path2;

    return 1;
}

static int path_mt_concat(lua_State* L)
{
    lua_settop(L, 2);

    fs::path path1, path2;

    switch (lua_type(L, 1)) {
    case LUA_TSTRING:
        try {
            path1 = fs::path{tou8stringview(L, 1), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
        if (!path || !lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        path1 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    case LUA_TSTRING:
        try {
            path2 = fs::path{tou8stringview(L, 2), fs::path::native_format};
            break;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    case LUA_TUSERDATA: {
        auto path = static_cast<fs::path*>(lua_touserdata(L, 2));
        if (!path || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        path2 = *path;
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto path = static_cast<fs::path*>(lua_newuserdata(L, sizeof(fs::path)));
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    setmetatable(L, -2);
    new (path) fs::path{path1};
    *path += path2;

    return 1;
}

#line 1275 "../src/filesystem.cpp"


#line 1421 "../src/filesystem.cpp"


static int file_clock_time_point_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 5514 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_2::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 1455 "../src/filesystem.cpp"
(L);
}

static int file_clock_time_point_mt_eq(lua_State* L)
{
    auto tp1 = static_cast<std::chrono::file_clock::time_point*>(
        lua_touserdata(L, 1));
    auto tp2 = static_cast<std::chrono::file_clock::time_point*>(
        lua_touserdata(L, 2));
    lua_pushboolean(L, *tp1 == *tp2);
    return 1;
}

static int file_clock_time_point_mt_lt(lua_State* L)
{
    auto tp1mem = lua_touserdata(L, 1);
    if (!tp1mem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp1 = aligned_userdata<std::chrono::file_clock::time_point>(tp1mem);

    auto tp2mem = lua_touserdata(L, 2);
    if (!tp2mem || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    auto tp2 = aligned_userdata<std::chrono::file_clock::time_point>(tp2mem);

    lua_pushboolean(L, *tp1 < *tp2);
    return 1;
}

static int file_clock_time_point_mt_le(lua_State* L)
{
    auto tp1mem = lua_touserdata(L, 1);
    if (!tp1mem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp1 = aligned_userdata<std::chrono::file_clock::time_point>(tp1mem);

    auto tp2mem = lua_touserdata(L, 2);
    if (!tp2mem || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    auto tp2 = aligned_userdata<std::chrono::file_clock::time_point>(tp2mem);

    lua_pushboolean(L, *tp1 <= *tp2);
    return 1;
}

static int file_clock_time_point_mt_add(lua_State* L)
{
    auto tpmem = lua_touserdata(L, 1);
    if (!tpmem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp = aligned_userdata<std::chrono::file_clock::time_point>(tpmem);

    lua_Number secs = luaL_checknumber(L, 2);
    if (std::isnan(secs) || std::isinf(secs)) {
        push(L, std::errc::argument_out_of_domain, "arg", 2);
        return lua_error(L);
    }

    lua_Seconds dur{secs};
    if (
        dur > std::chrono::file_clock::duration::max() ||
        dur < std::chrono::file_clock::duration::min()
    ) {
        push(L, std::errc::value_too_large);
        return lua_error(L);
    }

    auto ret = aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    setmetatable(L, -2);
    new (ret) std::chrono::file_clock::time_point{};

    try {
        *ret = *tp +
            std::chrono::round<std::chrono::file_clock::duration>(dur);
        return 1;
    } catch (const std::system_error& e) {
        push(L, e.code());
        return lua_error(L);
    } catch (const std::exception& e) {
        lua_pushstring(L, e.what());
        return lua_error(L);
    }
}

static int file_clock_time_point_mt_sub(lua_State* L)
{
    auto tpmem = lua_touserdata(L, 1);
    if (!tpmem || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    auto tp = aligned_userdata<std::chrono::file_clock::time_point>(tpmem);

    switch (lua_type(L, 2)) {
    case LUA_TNUMBER: {
        lua_Number secs = lua_tonumber(L, 2);
        if (std::isnan(secs) || std::isinf(secs)) {
            push(L, std::errc::argument_out_of_domain, "arg", 2);
            return lua_error(L);
        }

        lua_Seconds dur{secs};
        if (
            dur > std::chrono::file_clock::duration::max() ||
            dur < std::chrono::file_clock::duration::min()
        ) {
            push(L, std::errc::value_too_large);
            return lua_error(L);
        }

        auto ret =
            aligned_alloc_userdata<std::chrono::file_clock::time_point>(L);
        rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
        setmetatable(L, -2);
        new (ret) std::chrono::file_clock::time_point{};

        try {
            *ret = *tp -
                std::chrono::round<std::chrono::file_clock::duration>(dur);
            return 1;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    }
    case LUA_TUSERDATA: {
        auto tp2mem = lua_touserdata(L, 2);
        if (!tp2mem || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &file_clock_time_point_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        auto tp2 =
            aligned_userdata<std::chrono::file_clock::time_point>(tp2mem);

        try {
            lua_pushnumber(L, lua_Seconds{*tp - *tp2}.count());
            return 1;
        } catch (const std::system_error& e) {
            push(L, e.code());
            return lua_error(L);
        } catch (const std::exception& e) {
            lua_pushstring(L, e.what());
            return lua_error(L);
        }
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
}

#line 1677 "../src/filesystem.cpp"


static int space_info_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 5727 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_3::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 1691 "../src/filesystem.cpp"
(L);
}

static int space_info_mt_eq(lua_State* L)
{
    auto sp1 = static_cast<fs::space_info*>(lua_touserdata(L, 1));
    auto sp2 = static_cast<fs::space_info*>(lua_touserdata(L, 2));
    lua_pushboolean(L, *sp1 == *sp2);
    return 1;
}

#line 1761 "../src/filesystem.cpp"


static int file_status_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 5751 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_4::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 1774 "../src/filesystem.cpp"
(L);
}

static int file_status_mt_eq(lua_State* L)
{
    auto st1 = static_cast<fs::file_status*>(lua_touserdata(L, 1));
    auto st2 = static_cast<fs::file_status*>(lua_touserdata(L, 2));
#if BOOST_OS_MACOS
    lua_pushboolean(
        L,
        st1->type() == st2->type() &&
        st1->permissions() == st2->permissions());
#else // BOOST_OS_MACOS
    lua_pushboolean(L, *st1 == *st2);
#endif // BOOST_OS_MACOS
    return 1;
}

#line 1984 "../src/filesystem.cpp"


static int directory_entry_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 5782 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_5::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 2007 "../src/filesystem.cpp"
(L);
}

int directory_iterator::next(lua_State* L)
{
    auto self = static_cast<directory_iterator*>(
        lua_touserdata(L, lua_upvalueindex(1)));

    if (self->iterator == fs::directory_iterator{})
        return 0;

    if (self->increment) {
        std::error_code ec;
        self->iterator.increment(ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        if (self->iterator == fs::directory_iterator{})
            return 0;
    } else {
        self->increment = true;
    }

    auto ret = static_cast<fs::directory_entry*>(
        lua_newuserdata(L, sizeof(fs::directory_entry)));
    rawgetp(L, LUA_REGISTRYINDEX, &directory_entry_mt_key);
    setmetatable(L, -2);
    new (ret) fs::directory_entry{};
    *ret = *(self->iterator);
    return 1;
}

int directory_iterator::make(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::directory_options options = fs::directory_options::none;

    switch (lua_type(L, 2)) {
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        lua_getfield(L, 2, "skip_permission_denied");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TBOOLEAN:
            if (lua_toboolean(L, -1))
                options |= fs::directory_options::skip_permission_denied;
            break;
        default:
            push(L, std::errc::invalid_argument,
                 "arg", "skip_permission_denied");
            return lua_error(L);
        }
        break;
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    {
        std::error_code ec;
        auto iter = static_cast<directory_iterator*>(
            lua_newuserdata(L, sizeof(directory_iterator)));
        rawgetp(L, LUA_REGISTRYINDEX, &directory_iterator_mt_key);
        setmetatable(L, -2);
        new (iter) directory_iterator{*path, options, ec};

        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
    }
    lua_pushcclosure(L, next, 1);
    return 1;
}

#line 2169 "../src/filesystem.cpp"


int recursive_directory_iterator::mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 5889 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_6::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 2196 "../src/filesystem.cpp"
(L);
}

int recursive_directory_iterator::next(lua_State* L)
{
    auto self = static_cast<recursive_directory_iterator*>(
        lua_touserdata(L, lua_upvalueindex(1)));

    if (self->iterator == fs::recursive_directory_iterator{})
        return 0;

    if (self->increment) {
        std::error_code ec;
        self->iterator.increment(ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        if (self->iterator == fs::recursive_directory_iterator{})
            return 0;
    } else {
        self->increment = true;
    }

    auto ret = static_cast<fs::directory_entry*>(
        lua_newuserdata(L, sizeof(fs::directory_entry)));
    rawgetp(L, LUA_REGISTRYINDEX, &directory_entry_mt_key);
    setmetatable(L, -2);
    new (ret) fs::directory_entry{};
    *ret = *(self->iterator);

    lua_pushinteger(L, self->iterator.depth());

    return 2;
}

int recursive_directory_iterator::make(lua_State* L)
{
    lua_settop(L, 2);

    auto path = static_cast<fs::path*>(lua_touserdata(L, 1));
    if (!path || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &filesystem_path_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    fs::directory_options options = fs::directory_options::none;

    switch (lua_type(L, 2)) {
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        lua_getfield(L, 2, "skip_permission_denied");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TBOOLEAN:
            if (lua_toboolean(L, -1))
                options |= fs::directory_options::skip_permission_denied;
            break;
        default:
            push(L, std::errc::invalid_argument,
                 "arg", "skip_permission_denied");
            return lua_error(L);
        }

        lua_getfield(L, 2, "follow_directory_symlink");
        switch (lua_type(L, -1)) {
        case LUA_TNIL:
            break;
        case LUA_TBOOLEAN:
            if (lua_toboolean(L, -1))
                options |= fs::directory_options::follow_directory_symlink;
            break;
        default:
            push(L, std::errc::invalid_argument,
                 "arg", "follow_directory_symlink");
            return lua_error(L);
        }
        break;
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    {
        std::error_code ec;
        auto iter = static_cast<recursive_directory_iterator*>(
            lua_newuserdata(L, sizeof(recursive_directory_iterator)));
        rawgetp(L, LUA_REGISTRYINDEX, &recursive_directory_iterator_mt_key);
        setmetatable(L, -2);
        new (iter) recursive_directory_iterator{*path, options, ec};

        if (ec) {
            push(L, ec);
            lua_pushliteral(L, "path1");
            lua_pushvalue(L, 1);
            lua_rawset(L, -3);
            return lua_error(L);
        }
    }
    lua_pushvalue(L, -1);
    lua_pushcclosure(L, next, 1);
    lua_insert(L, -2);
    return 2;
}

#line 2373 "../src/filesystem.cpp"


#line 4593 "../src/filesystem.cpp"


static int path_ctors_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 6018 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_8::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 4623 "../src/filesystem.cpp"
(L);
}

static int clock_ctors_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 6031 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_9::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 4659 "../src/filesystem.cpp"
(L);
}

static int filesystem_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 6044 "libemilua.so.0.11.5.p/bc2_filesystem.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_10::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 5065 "../src/filesystem.cpp"
(L);
}

void init_filesystem(lua_State* L)
{
    lua_pushlightuserdata(L, &filesystem_path_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/9);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.path");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, path_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__tostring");
        lua_pushcfunction(L, path_mt_tostring);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__eq");
        lua_pushcfunction(L, path_mt_eq);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__lt");
        lua_pushcfunction(L, path_mt_lt);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__le");
        lua_pushcfunction(L, path_mt_le);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__div");
        lua_pushcfunction(L, path_mt_div);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__concat");
        lua_pushcfunction(L, path_mt_concat);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<fs::path>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &file_clock_time_point_mt_key);
    {
        static_assert(std::is_trivially_destructible_v<
            std::chrono::file_clock::time_point>);

        lua_createtable(L, /*narr=*/0, /*nrec=*/7);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.clock.time_point");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, file_clock_time_point_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__eq");
        lua_pushcfunction(L, file_clock_time_point_mt_eq);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__lt");
        lua_pushcfunction(L, file_clock_time_point_mt_lt);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__le");
        lua_pushcfunction(L, file_clock_time_point_mt_le);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__add");
        lua_pushcfunction(L, file_clock_time_point_mt_add);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__sub");
        lua_pushcfunction(L, file_clock_time_point_mt_sub);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &filesystem_path_iterator_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/1);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<fs::path::iterator>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &directory_iterator_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/1);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<directory_iterator>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &recursive_directory_iterator_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.recursive_directory_iterator");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, recursive_directory_iterator::mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<recursive_directory_iterator>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &space_info_mt_key);
    {
        static_assert(std::is_trivially_destructible_v<fs::space_info>);

        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.space_info");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, space_info_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__eq");
        lua_pushcfunction(L, space_info_mt_eq);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &file_status_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/4);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.file_status");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, file_status_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__eq");
        lua_pushcfunction(L, file_status_mt_eq);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<fs::file_status>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &directory_entry_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.directory_entry");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, directory_entry_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<fs::directory_entry>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &path_ctors_key);
    lua_newuserdata(L, 1);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/2);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.path");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, path_ctors_mt_index);
        lua_rawset(L, -3);
    }
    setmetatable(L, -2);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &clock_ctors_key);
    lua_newuserdata(L, 1);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/2);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem.clock");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, clock_ctors_mt_index);
        lua_rawset(L, -3);
    }
    setmetatable(L, -2);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &filesystem_key);
    lua_newuserdata(L, 1);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/2);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "filesystem");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, filesystem_mt_index);
        lua_rawset(L, -3);
    }
    setmetatable(L, -2);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &mode_key);
    int res = luaL_loadbuffer(
        L, reinterpret_cast<char*>(mode_bytecode), mode_bytecode_size, nullptr);
    assert(res == 0); boost::ignore_unused(res);
    lua_getglobal(L, "bit");
    lua_pushinteger(L, static_cast<int>(std::filesystem::perms::set_uid));
    lua_pushinteger(L, static_cast<int>(std::filesystem::perms::set_gid));
    lua_pushinteger(L, static_cast<int>(std::filesystem::perms::sticky_bit));
    lua_call(L, 4, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
}

} // namespace emilua
