#line 2 "libemilua.so.0.11.5.p/bc2_ip.cpp"
#include <cstring>

namespace emilua::gperf::detail { using std::size_t; using std::strcmp; }
#line 6 "libemilua.so.0.11.5.p/bc2_ip.cpp"


#line 9 "libemilua.so.0.11.5.p/bc2_ip.cpp"

#include <boost/scope_exit.hpp>

#include <charconv>

#include <emilua/file_descriptor.hpp>
#include <emilua/async_base.hpp>
#include <emilua/byte_span.hpp>
#include <emilua/ip.hpp>

#if EMILUA_CONFIG_USE_STANDALONE_ASIO
#include <asio/ip/host_name.hpp>
#include <asio/ip/multicast.hpp>
#include <asio/ip/address.hpp>
#include <asio/ip/unicast.hpp>
#include <asio/ip/v6_only.hpp>
#include <asio/ip/tcp.hpp>
#else // EMILUA_CONFIG_USE_STANDALONE_ASIO
#include <boost/asio/ip/host_name.hpp>
#include <boost/asio/ip/multicast.hpp>
#include <boost/asio/ip/address.hpp>
#include <boost/asio/ip/unicast.hpp>
#include <boost/asio/ip/v6_only.hpp>
#include <boost/asio/ip/tcp.hpp>
#endif // EMILUA_CONFIG_USE_STANDALONE_ASIO

#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
# if EMILUA_CONFIG_USE_STANDALONE_ASIO
#  include <asio/windows/object_handle.hpp>
# else // EMILUA_CONFIG_USE_STANDALONE_ASIO
#  include <boost/asio/windows/object_handle.hpp>
# endif // EMILUA_CONFIG_USE_STANDALONE_ASIO

# include <boost/nowide/convert.hpp>
# include <boost/scope_exit.hpp>
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

#if BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
# if EMILUA_CONFIG_USE_STANDALONE_ASIO
#  include <asio/windows/overlapped_ptr.hpp>
#  include <asio/random_access_file.hpp>
# else // EMILUA_CONFIG_USE_STANDALONE_ASIO
#  include <boost/asio/windows/overlapped_ptr.hpp>
#  include <boost/asio/random_access_file.hpp>
# endif // EMILUA_CONFIG_USE_STANDALONE_ASIO

# include <emilua/file.hpp>
#endif // BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO

#line 59 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static char tcp_socket_connect_key;
static char tcp_socket_read_some_key;
static char tcp_socket_write_some_key;
static char tcp_socket_receive_key;
static char tcp_socket_send_key;
static char tcp_socket_wait_key;
static char tcp_acceptor_accept_key;
static char tcp_acceptor_wait_key;
static char udp_socket_connect_key;
static char udp_socket_receive_key;
static char udp_socket_receive_from_key;
static char udp_socket_send_key;
static char udp_socket_send_to_key;

} // namespace emilua

#line 79 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


#if BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
static char tcp_socket_send_file_key;
#endif // BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO

} // namespace emilua

#line 89 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int address_to_v6(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2) || !a->is_v4()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (ret) asio::ip::address{
        asio::ip::make_address_v6(asio::ip::v4_mapped, a->to_v4())
    };

    return 1;
}

static int address_to_v4(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2) || !a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    try {
        new (ret) asio::ip::address{
            asio::ip::make_address_v4(asio::ip::v4_mapped, a->to_v6())
        };
    } catch (const asio::ip::bad_address_cast&) {
        static_assert(std::is_trivially_destructible_v<asio::ip::address>);
        push(L, std::errc::invalid_argument);
        return lua_error(L);
    }

    return 1;
}

inline int address_is_loopback(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    lua_pushboolean(L, a->is_loopback());
    return 1;
}

inline int address_is_multicast(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    lua_pushboolean(L, a->is_multicast());
    return 1;
}

inline int address_is_unspecified(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    lua_pushboolean(L, a->is_unspecified());
    return 1;
}

inline int address_is_v4(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    lua_pushboolean(L, a->is_v4());
    return 1;
}

inline int address_is_v6(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    lua_pushboolean(L, a->is_v6());
    return 1;
}

inline int address_is_link_local(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_link_local());
    return 1;
}

inline int address_is_multicast_global(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_multicast_global());
    return 1;
}

inline int address_is_multicast_link_local(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_multicast_link_local());
    return 1;
}

inline int address_is_multicast_node_local(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_multicast_node_local());
    return 1;
}

inline int address_is_multicast_org_local(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_multicast_org_local());
    return 1;
}

inline int address_is_multicast_site_local(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_multicast_site_local());
    return 1;
}

inline int address_is_site_local(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_site_local());
    return 1;
}

inline int address_is_v4_mapped(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushboolean(L, a->to_v6().is_v4_mapped());
    return 1;
}

inline int address_scope_id_get(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushnumber(L, a->to_v6().scope_id());
    return 1;
}

} // namespace emilua

#line 294 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_1  */
/* Computed positions: -k'1,4-5,14' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_1 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 26, duplicates = 0 */

class Perfect_Hash_1
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_1 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_1::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34,  5, 34, 15, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
       0,  0, 34,  0, 34,  0, 34, 34,  5,  0,
      10,  0,  0, 34, 34,  0,  5,  0,  0, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34, 34, 34, 34, 34,
      34, 34, 34, 34, 34, 34
    };
  unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[static_cast<unsigned char>(str[13])];
      /*FALLTHROUGH*/
      case 13:
      case 12:
      case 11:
      case 10:
      case 9:
      case 8:
      case 7:
      case 6:
      case 5:
        hval += asso_values[static_cast<unsigned char>(str[4])];
      /*FALLTHROUGH*/
      case 4:
        hval += asso_values[static_cast<unsigned char>(str[3])];
      /*FALLTHROUGH*/
      case 3:
      case 2:
      case 1:
        hval += asso_values[static_cast<unsigned char>(str[0])];
        break;
    }
  return hval;
}

const struct word_type_1 *
Perfect_Hash_1::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 16,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 23,
      MIN_HASH_VALUE = 8,
      MAX_HASH_VALUE = 33
    };

  static const struct word_type_1 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}},
      {"scope_id",  address_scope_id_get},
      {"",{}},
      {"is_v4",  address_is_v4},
      {"",{}},
      {"is_multicast",  address_is_multicast},
      {"is_site_local",  address_is_site_local},
      {"",{}},
      {"to_v4", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, address_to_v4);
                return 1;
            }},
      {"is_loopback",  address_is_loopback},
      {"is_v4_mapped",  address_is_v4_mapped},
      {"is_link_local",  address_is_link_local},
      {"is_multicast_global",  address_is_multicast_global},
      {"is_v6",  address_is_v6},
      {"",{}},
      {"is_multicast_org_local",  address_is_multicast_org_local},
      {"is_multicast_site_local",  address_is_multicast_site_local},
      {"is_unspecified",  address_is_unspecified},
      {"to_v6", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, address_to_v6);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"is_multicast_link_local",  address_is_multicast_link_local},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"is_multicast_node_local",  address_is_multicast_node_local}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 467 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


inline int address_scope_id_set(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TNUMBER);

    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a->is_v6()) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto as_v6 = a->to_v6();
    as_v6.scope_id(lua_tonumber(L, 3));
    *a = as_v6;
    return 0;
}

} // namespace emilua

#line 489 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_2  */
/* Computed positions: -k'' */

struct word_type_2 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 1, duplicates = 0 */

class Perfect_Hash_2
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_2 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_2::hash (const char *str, size_t len)
{
  return 0;
}

const struct word_type_2 *
Perfect_Hash_2::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 1,
      MIN_WORD_LENGTH = 8,
      MAX_WORD_LENGTH = 8,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 0
    };

  static const struct word_type_2 wordlist[] =
    {
      {"scope_id",  address_scope_id_set}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 549 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_3  */
/* Computed positions: -k'2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_3 { const char* name; asio::ip::tcp action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_3
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_3 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_3::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 1, 2, 0, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[1])];
}

const struct word_type_3 *
Perfect_Hash_3::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 2,
      MAX_WORD_LENGTH = 2,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_3 wordlist[] =
    {
      {"v6",  asio::ip::tcp::v6()},
      {"v4",  asio::ip::tcp::v4()}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 666 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_4  */
/* Computed positions: -k'2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_4 { const char* name; asio::ip::tcp action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_4
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_4 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_4::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 1, 2, 0, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[1])];
}

const struct word_type_4 *
Perfect_Hash_4::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 2,
      MAX_WORD_LENGTH = 2,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_4 wordlist[] =
    {
      {"v6",  asio::ip::tcp::v6()},
      {"v4",  asio::ip::tcp::v4()}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 783 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_5  */
/* Computed positions: -k'' */

struct word_type_5 { const char* name; int (*action)(lua_State*, tcp_socket*); };
/* maximum key range = 1, duplicates = 0 */

class Perfect_Hash_5
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_5 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_5::hash (const char *str, size_t len)
{
  return 0;
}

const struct word_type_5 *
Perfect_Hash_5::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 1,
      MIN_WORD_LENGTH = 14,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 0
    };

  static const struct word_type_5 wordlist[] =
    {
      {"bytes_readable", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::bytes_readable command;
                asio_error_code ec;
                socket->socket.io_control(command, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushnumber(L, command.get());
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 854 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_6  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_6 { const char* name; asio::ip::tcp::socket::shutdown_type action; };
/* maximum key range = 4, duplicates = 0 */

class Perfect_Hash_6
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_6 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_6::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 1, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 0, 0, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_6 *
Perfect_Hash_6::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 7,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 7
    };

  static const struct word_type_6 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"send",  asio::ip::tcp::socket::shutdown_send},
      {"both",  asio::ip::tcp::socket::shutdown_both},
      {"",{}},
      {"receive",  asio::ip::tcp::socket::shutdown_receive}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 973 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int tcp_socket_open(lua_State* L)
{
    lua_settop(L, 2);
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        sock->socket.open(asio::ip::tcp::endpoint{*addr, 0}.protocol(), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TSTRING: {
        auto key = tostringview(L, 2);
        auto protocol =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_3::in_word_set((key).data(), (key).size()));

        if (!protocol) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        sock->socket.open(*protocol, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int tcp_socket_bind(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TNUMBER);
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio::ip::tcp::endpoint ep(*addr, lua_tointeger(L, 3));
        asio_error_code ec;
        sock->socket.bind(ep, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TSTRING: {
        asio_error_code ec;
        auto addr = asio::ip::make_address(lua_tostring(L, 2), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        asio::ip::tcp::endpoint ep(addr, lua_tointeger(L, 3));
        sock->socket.bind(ep, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int tcp_socket_close(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    sock->socket.close(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

static int tcp_socket_cancel(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    sock->socket.cancel(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

#if BOOST_OS_UNIX
static int tcp_socket_assign(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto handle = static_cast<file_descriptor_handle*>(lua_touserdata(L, 3));
    if (!handle || !lua_getmetatable(L, 3)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    if (*handle == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::device_or_resource_busy);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        lua_pushnil(L);
        setmetatable(L, 3);

        asio_error_code ec;
        sock->socket.assign(
            asio::ip::tcp::endpoint{*addr, 0}.protocol(), *handle, ec);
        assert(!ec); boost::ignore_unused(ec);

        return 0;
    }
    case LUA_TSTRING: {
        auto key = tostringview(L, 2);
        auto protocol =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_4::in_word_set((key).data(), (key).size()));

        if (!protocol) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        lua_pushnil(L);
        setmetatable(L, 3);

        asio_error_code ec;
        sock->socket.assign(*protocol, *handle, ec);
        assert(!ec); boost::ignore_unused(ec);

        return 0;
    }
    }
}

static int tcp_socket_release(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    int rawfd = sock->socket.release(ec);
    BOOST_SCOPE_EXIT_ALL(&) {
        if (rawfd != INVALID_FILE_DESCRIPTOR) {
            int res = close(rawfd);
            boost::ignore_unused(res);
        }
    };

    if (ec) {
        push(L, ec);
        return lua_error(L);
    }

    auto fdhandle = static_cast<file_descriptor_handle*>(
        lua_newuserdata(L, sizeof(file_descriptor_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    setmetatable(L, -2);

    *fdhandle = rawfd;
    rawfd = INVALID_FILE_DESCRIPTOR;
    return 1;
}
#endif // BOOST_OS_UNIX

static int tcp_socket_io_control(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_5::in_word_set((key).data(), (key).size()), [](lua_State* L, tcp_socket*) -> int {
            push(L, std::errc::not_supported);
            return lua_error(L);
        })(L, socket);
}

static int tcp_socket_shutdown(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    auto what =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_6::in_word_set((key).data(), (key).size()));
    if (!what) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    asio_error_code ec;
    socket->socket.shutdown(*what, ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

static int tcp_socket_disconnect(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

#if BOOST_OS_WINDOWS
    push(L, std::errc::function_not_supported);
    return lua_error(L);
#else
    struct sockaddr_in sin;
    std::memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_UNSPEC;
    int res = connect(
        sock->socket.native_handle(), reinterpret_cast<struct sockaddr*>(&sin),
        sizeof(sin));
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    return 0;
#endif // BOOST_OS_WINDOWS
}

} // namespace emilua

#line 1338 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_7  */
/* Computed positions: -k'' */

struct word_type_7 { const char* name; asio::socket_base::message_flags action; };
/* maximum key range = 10, duplicates = 0 */

class Perfect_Hash_7
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_7 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_7::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_7 *
Perfect_Hash_7::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 13
    };

  static const struct word_type_7 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"peek",  asio::socket_base::message_peek},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"out_of_band",  asio::socket_base::message_out_of_band},
      {"do_not_route",  asio::socket_base::message_do_not_route},
      {"end_of_record",  asio::socket_base::message_end_of_record}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1403 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_8  */
/* Computed positions: -k'' */

struct word_type_8 { const char* name; asio::socket_base::message_flags action; };
/* maximum key range = 10, duplicates = 0 */

class Perfect_Hash_8
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_8 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_8::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_8 *
Perfect_Hash_8::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 13
    };

  static const struct word_type_8 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"peek",  asio::socket_base::message_peek},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"out_of_band",  asio::socket_base::message_out_of_band},
      {"do_not_route",  asio::socket_base::message_do_not_route},
      {"end_of_record",  asio::socket_base::message_end_of_record}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1468 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_9  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_9 { const char* name; asio::ip::tcp::socket::wait_type action; };
/* maximum key range = 3, duplicates = 0 */

class Perfect_Hash_9
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_9 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_9::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 1, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 0, 7, 7, 7, 7, 0,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_9 *
Perfect_Hash_9::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 5,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 6
    };

  static const struct word_type_9 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"read",  asio::ip::tcp::socket::wait_read},
      {"write",  asio::ip::tcp::socket::wait_write},
      {"error",  asio::ip::tcp::socket::wait_error}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1586 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_10  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_10 { const char* name; int (*action)(lua_State*, tcp_socket*); };
/* maximum key range = 19, duplicates = 0 */

class Perfect_Hash_10
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_10 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_10::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
       0, 24, 24, 24, 24, 24, 24,  0,  0, 24,
      24,  5, 24, 24,  0,  0,  5, 24,  0, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_10 *
Perfect_Hash_10::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 11,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 21,
      MIN_HASH_VALUE = 5,
      MAX_HASH_VALUE = 23
    };

  static const struct word_type_10 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"debug", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::debug o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"linger", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                luaL_checktype(L, 4, LUA_TNUMBER);
                asio::socket_base::linger o(
                    lua_toboolean(L, 3), lua_tointeger(L, 4));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"v6_only", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::ip::v6_only o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}},
      {"keep_alive", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::keep_alive o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"do_not_route", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::do_not_route o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"send_buffer_size", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::socket_base::send_buffer_size o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"tcp_no_delay", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::ip::tcp::no_delay o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"send_low_watermark", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::socket_base::send_low_watermark o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"receive_buffer_size", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::socket_base::receive_buffer_size o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"receive_low_watermark", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::socket_base::receive_low_watermark o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"out_of_band_inline", 
            [](lua_State* L, tcp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::out_of_band_inline o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 1840 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_11  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_11 { const char* name; int (*action)(lua_State*, tcp_socket* socket); };
/* maximum key range = 19, duplicates = 0 */

class Perfect_Hash_11
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_11 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_11::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
       0, 24, 24, 24, 24, 24, 24,  0,  0, 24,
      24,  5, 24, 24,  0,  0,  5, 24,  0, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
      24, 24, 24, 24, 24, 24
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_11 *
Perfect_Hash_11::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 11,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 21,
      MIN_HASH_VALUE = 5,
      MAX_HASH_VALUE = 23
    };

  static const struct word_type_11 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"debug", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::debug o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"linger", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::linger o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.enabled());
                lua_pushinteger(L, o.timeout());
                return 2;
            }},
      {"v6_only", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::ip::v6_only o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}},
      {"keep_alive", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::keep_alive o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}},
      {"do_not_route", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::do_not_route o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"send_buffer_size", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::send_buffer_size o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"tcp_no_delay", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::ip::tcp::no_delay o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"send_low_watermark", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::send_low_watermark o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"receive_buffer_size", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::receive_buffer_size o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"",{}},
      {"receive_low_watermark", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::receive_low_watermark o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"",{}},
      {"out_of_band_inline", 
            [](lua_State* L, tcp_socket* socket) -> int {
                asio::socket_base::out_of_band_inline o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2093 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int tcp_socket_set_option(lua_State* L)
{
    lua_settop(L, 4);
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_10::in_word_set((key).data(), (key).size()), [](lua_State* L, tcp_socket*) -> int {
            push(L, std::errc::not_supported);
            return lua_error(L);
        })(L, socket);
}

static int tcp_socket_get_option(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_11::in_word_set((key).data(), (key).size()), [](lua_State* L, tcp_socket* socket) -> int {
            push(L, std::errc::not_supported);
            return lua_error(L);
        })(L, socket);
}

inline int tcp_socket_is_open(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    lua_pushboolean(L, sock->socket.is_open());
    return 1;
}

inline int tcp_socket_local_address(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.local_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    auto addr = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (addr) asio::ip::address{ep.address()};
    return 1;
}

inline int tcp_socket_local_port(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.local_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    lua_pushinteger(L, ep.port());
    return 1;
}

inline int tcp_socket_remote_address(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.remote_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    auto addr = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (addr) asio::ip::address{ep.address()};
    return 1;
}

inline int tcp_socket_remote_port(lua_State* L)
{
    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.remote_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    lua_pushinteger(L, ep.port());
    return 1;
}

inline int tcp_socket_at_mark(lua_State* L)
{
    auto socket = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    bool ret = socket->socket.at_mark(ec);
    if (ec) {
        push(L, ec);
        return lua_error(L);
    }
    lua_pushboolean(L, ret);
    return 1;
}

} // namespace emilua

#line 2226 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_12  */
/* Computed positions: -k'1,3' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_12 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 32, duplicates = 0 */

class Perfect_Hash_12
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_12 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_12::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 10, 36,  5, 30,  0,
      25, 25, 36, 15, 36, 10, 36, 36,  5,  5,
       0,  0, 36, 36,  5,  0,  0,  0, 36,  0,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36
    };
  return len + asso_values[static_cast<unsigned char>(str[2])] + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_12 *
Perfect_Hash_12::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 24,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 35
    };

  static const struct word_type_12 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"send", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_send_key);
                return 1;
            }},
      {"close", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_close);
                return 1;
            }},
      {"cancel", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_cancel);
                return 1;
            }},
      {"connect", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_connect_key);
                return 1;
            }},
      {"shutdown", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_shutdown);
                return 1;
            }},
      {"send_file", 
            [](lua_State* L) -> int {
#if BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_send_file_key);
#else // BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
                return 1;
            }},
      {"set_option", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_set_option);
                return 1;
            }},
      {"assign", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, tcp_socket_assign);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"receive", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_receive_key);
                return 1;
            }},
      {"",{}},
      {"wait", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_wait_key);
                return 1;
            }},
      {"local_port",  tcp_socket_local_port},
      {"",{}},
      {"release", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, tcp_socket_release);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"local_address",  tcp_socket_local_address},
      {"read_some", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_read_some_key);
                return 1;
            }},
      {"write_some", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_socket_write_some_key);
                return 1;
            }},
      {"remote_port",  tcp_socket_remote_port},
      {"at_mark",  tcp_socket_at_mark},
      {"",{}},
      {"remote_address",  tcp_socket_remote_address},
      {"get_option", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_get_option);
                return 1;
            }},
      {"",{}},
      {"is_open",  tcp_socket_is_open},
      {"",{}},
      {"open", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_open);
                return 1;
            }},
      {"io_control", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_io_control);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"bind", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_bind);
                return 1;
            }},
      {"disconnect", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_socket_disconnect);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2455 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_13  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_13 { const char* name; asio::ip::tcp::acceptor::wait_type action; };
/* maximum key range = 3, duplicates = 0 */

class Perfect_Hash_13
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_13 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_13::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 1, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 0, 7, 7, 7, 7, 0,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
      7, 7, 7, 7, 7, 7
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_13 *
Perfect_Hash_13::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 5,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 6
    };

  static const struct word_type_13 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"read",  asio::ip::tcp::acceptor::wait_read},
      {"write",  asio::ip::tcp::acceptor::wait_write},
      {"error",  asio::ip::tcp::acceptor::wait_error}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2573 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_14  */
/* Computed positions: -k'2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_14 { const char* name; asio::ip::tcp action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_14
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_14 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_14::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 1, 2, 0, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[1])];
}

const struct word_type_14 *
Perfect_Hash_14::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 2,
      MAX_WORD_LENGTH = 2,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_14 wordlist[] =
    {
      {"v6",  asio::ip::tcp::v6()},
      {"v4",  asio::ip::tcp::v4()}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2690 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_15  */
/* Computed positions: -k'' */

struct word_type_15 { const char* name; int (*action)(lua_State*, asio::ip::tcp::acceptor*); };
/* maximum key range = 21, duplicates = 0 */

class Perfect_Hash_15
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_15 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_15::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_15 *
Perfect_Hash_15::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 25,
      MIN_HASH_VALUE = 5,
      MAX_HASH_VALUE = 25
    };

  static const struct word_type_15 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"debug", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::debug o(lua_toboolean(L, 3));
                asio_error_code ec;
                acceptor->set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"v6_only", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::ip::v6_only o(lua_toboolean(L, 3));
                asio_error_code ec;
                acceptor->set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"reuse_address", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::reuse_address o(lua_toboolean(L, 3));
                asio_error_code ec;
                acceptor->set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"enable_connection_aborted", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::enable_connection_aborted o(
                    lua_toboolean(L, 3));
                asio_error_code ec;
                acceptor->set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2803 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_16  */
/* Computed positions: -k'' */

struct word_type_16 { const char* name; int (*action)(lua_State*, asio::ip::tcp::acceptor*); };
/* maximum key range = 21, duplicates = 0 */

class Perfect_Hash_16
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_16 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_16::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_16 *
Perfect_Hash_16::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 25,
      MIN_HASH_VALUE = 5,
      MAX_HASH_VALUE = 25
    };

  static const struct word_type_16 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"debug", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                asio::socket_base::debug o;
                asio_error_code ec;
                acceptor->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}},
      {"v6_only", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                asio::ip::v6_only o;
                asio_error_code ec;
                acceptor->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"reuse_address", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                asio::socket_base::reuse_address o;
                asio_error_code ec;
                acceptor->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"enable_connection_aborted", 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                asio::socket_base::enable_connection_aborted o;
                asio_error_code ec;
                acceptor->get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 2915 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_17  */
/* Computed positions: -k'2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_17 { const char* name; asio::ip::tcp action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_17
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_17 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_17::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 1, 2, 0, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[1])];
}

const struct word_type_17 *
Perfect_Hash_17::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 2,
      MAX_WORD_LENGTH = 2,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_17 wordlist[] =
    {
      {"v6",  asio::ip::tcp::v6()},
      {"v4",  asio::ip::tcp::v4()}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 3032 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int tcp_acceptor_listen(lua_State* L)
{
    lua_settop(L, 2);
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TNIL: {
        asio_error_code ec;
        acceptor->listen(asio::socket_base::max_listen_connections, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TNUMBER: {
        asio_error_code ec;
        acceptor->listen(lua_tointeger(L, 2), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int tcp_acceptor_bind(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TNUMBER);
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        acceptor->bind(asio::ip::tcp::endpoint(*addr, lua_tointeger(L, 3)), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TSTRING: {
        asio_error_code ec;
        auto addr = asio::ip::make_address(lua_tostring(L, 2), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        acceptor->bind(asio::ip::tcp::endpoint(addr, lua_tointeger(L, 3)), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int tcp_acceptor_open(lua_State* L)
{
    lua_settop(L, 2);
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        acceptor->open(asio::ip::tcp::endpoint{*addr, 0}.protocol(), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TSTRING: {
        auto key = tostringview(L, 2);
        auto protocol =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_14::in_word_set((key).data(), (key).size()));

        if (!protocol) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        acceptor->open(*protocol, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int tcp_acceptor_close(lua_State* L)
{
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    acceptor->close(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

static int tcp_acceptor_set_option(lua_State* L)
{
    lua_settop(L, 3);
    luaL_checktype(L, 2, LUA_TSTRING);

    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_15::in_word_set((key).data(), (key).size()), 
            [](lua_State* L, asio::ip::tcp::acceptor*) -> int {
                push(L, std::errc::not_supported);
                return lua_error(L);
            })(L, acceptor);
}

static int tcp_acceptor_get_option(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_16::in_word_set((key).data(), (key).size()), 
            [](lua_State* L, asio::ip::tcp::acceptor* acceptor) -> int {
                push(L, std::errc::not_supported);
                return lua_error(L);
            })(L, acceptor);
}

static int tcp_acceptor_cancel(lua_State* L)
{
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    acceptor->cancel(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

#if BOOST_OS_UNIX
static int tcp_acceptor_assign(lua_State* L)
{
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto handle = static_cast<file_descriptor_handle*>(lua_touserdata(L, 3));
    if (!handle || !lua_getmetatable(L, 3)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    if (*handle == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::device_or_resource_busy);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        lua_pushnil(L);
        setmetatable(L, 3);

        asio_error_code ec;
        acceptor->assign(
            asio::ip::tcp::endpoint{*addr, 0}.protocol(), *handle, ec);
        assert(!ec); boost::ignore_unused(ec);

        return 0;
    }
    case LUA_TSTRING: {
        auto key = tostringview(L, 2);
        auto protocol =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_17::in_word_set((key).data(), (key).size()));

        if (!protocol) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        lua_pushnil(L);
        setmetatable(L, 3);

        asio_error_code ec;
        acceptor->assign(*protocol, *handle, ec);
        assert(!ec); boost::ignore_unused(ec);

        return 0;
    }
    }
}

static int tcp_acceptor_release(lua_State* L)
{
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    if (acceptor->native_handle() == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::bad_file_descriptor);
        return lua_error(L);
    }

    asio_error_code ec;
    int rawfd = acceptor->release(ec);
    BOOST_SCOPE_EXIT_ALL(&) {
        if (rawfd != INVALID_FILE_DESCRIPTOR) {
            int res = close(rawfd);
            boost::ignore_unused(res);
        }
    };

    if (ec) {
        push(L, ec);
        return lua_error(L);
    }

    auto fdhandle = static_cast<file_descriptor_handle*>(
        lua_newuserdata(L, sizeof(file_descriptor_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    setmetatable(L, -2);

    *fdhandle = rawfd;
    rawfd = INVALID_FILE_DESCRIPTOR;
    return 1;
}
#endif // BOOST_OS_UNIX

inline int tcp_acceptor_is_open(lua_State* L)
{
    auto a = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    lua_pushboolean(L, a->is_open());
    return 1;
}

inline int tcp_acceptor_local_address(lua_State* L)
{
    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = acceptor->local_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    auto addr = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (addr) asio::ip::address{ep.address()};
    return 1;
}

inline int tcp_acceptor_local_port(lua_State* L)
{
    auto a = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = a->local_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    lua_pushinteger(L, ep.port());
    return 1;
}

} // namespace emilua

#line 3440 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_18  */
/* Computed positions: -k'1-2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_18 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 22, duplicates = 0 */

class Perfect_Hash_18
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_18 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_18::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 10,  0,  5,
       5, 26,  5, 10, 26,  0,  0, 26,  0, 10,
      26,  5,  0,  5,  0,  0,  0, 26, 26,  5,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26, 26, 26, 26,
      26, 26, 26, 26, 26, 26, 26
    };
  return len + asso_values[static_cast<unsigned char>(str[1]+1)] + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_18 *
Perfect_Hash_18::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 14,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 25
    };

  static const struct word_type_18 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"bind", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_bind);
                return 1;
            }},
      {"",{}},
      {"listen", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_listen);
                return 1;
            }},
      {"is_open",  tcp_acceptor_is_open},
      {"",{}},
      {"wait", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_acceptor_wait_key);
                return 1;
            }},
      {"local_port",  tcp_acceptor_local_port},
      {"cancel", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_cancel);
                return 1;
            }},
      {"release", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, tcp_acceptor_release);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"local_address",  tcp_acceptor_local_address},
      {"open", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_open);
                return 1;
            }},
      {"set_option", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_set_option);
                return 1;
            }},
      {"assign", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, tcp_acceptor_assign);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"close", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_close);
                return 1;
            }},
      {"accept", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &tcp_acceptor_accept_key);
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}},
      {"get_option", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, tcp_acceptor_get_option);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 3625 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_19  */
/* Computed positions: -k'' */

struct word_type_19 { const char* name; asio::ip::resolver_base::flags action; };
/* maximum key range = 12, duplicates = 0 */

class Perfect_Hash_19
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_19 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_19::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_19 *
Perfect_Hash_19::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 18,
      MIN_HASH_VALUE = 7,
      MAX_HASH_VALUE = 18
    };

  static const struct word_type_19 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"passive",  asio::ip::resolver_base::passive},
      {"",{}},
      {"v4_mapped",  asio::ip::resolver_base::v4_mapped},
      {"",{}}, {"",{}},
      {"all_matching",  asio::ip::resolver_base::all_matching},
      {"",{}},
      {"canonical_name",  asio::ip::resolver_base::canonical_name},
      {"",{}}, {"",{}}, {"",{}},
      {"address_configured", 
                    asio::ip::resolver_base::address_configured}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 3696 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_20  */
/* Computed positions: -k'' */

struct word_type_20 { const char* name; asio::ip::resolver_base::flags action; };
/* maximum key range = 12, duplicates = 0 */

class Perfect_Hash_20
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_20 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_20::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_20 *
Perfect_Hash_20::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 18,
      MIN_HASH_VALUE = 7,
      MAX_HASH_VALUE = 18
    };

  static const struct word_type_20 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"passive",  asio::ip::resolver_base::passive},
      {"",{}},
      {"v4_mapped",  asio::ip::resolver_base::v4_mapped},
      {"",{}}, {"",{}},
      {"all_matching",  asio::ip::resolver_base::all_matching},
      {"",{}},
      {"canonical_name",  asio::ip::resolver_base::canonical_name},
      {"",{}}, {"",{}}, {"",{}},
      {"address_configured", 
                    asio::ip::resolver_base::address_configured}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 3767 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_21  */
/* Computed positions: -k'' */

struct word_type_21 { const char* name; asio::ip::resolver_base::flags action; };
/* maximum key range = 12, duplicates = 0 */

class Perfect_Hash_21
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_21 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_21::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_21 *
Perfect_Hash_21::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 18,
      MIN_HASH_VALUE = 7,
      MAX_HASH_VALUE = 18
    };

  static const struct word_type_21 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"passive",  asio::ip::resolver_base::passive},
      {"",{}},
      {"v4_mapped",  asio::ip::resolver_base::v4_mapped},
      {"",{}}, {"",{}},
      {"all_matching",  asio::ip::resolver_base::all_matching},
      {"",{}},
      {"canonical_name",  asio::ip::resolver_base::canonical_name},
      {"",{}}, {"",{}}, {"",{}},
      {"address_configured", 
                    asio::ip::resolver_base::address_configured}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 3838 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_22  */
/* Computed positions: -k'2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_22 { const char* name; asio::ip::udp action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_22
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_22 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_22::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 1, 2, 0, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[1])];
}

const struct word_type_22 *
Perfect_Hash_22::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 2,
      MAX_WORD_LENGTH = 2,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_22 wordlist[] =
    {
      {"v6",  asio::ip::udp::v6()},
      {"v4",  asio::ip::udp::v4()}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 3955 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_23  */
/* Computed positions: -k'1' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_23 { const char* name; asio::ip::udp::socket::shutdown_type action; };
/* maximum key range = 4, duplicates = 0 */

class Perfect_Hash_23
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_23 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_23::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 1, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 0, 0, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
      8, 8, 8, 8, 8, 8
    };
  return len + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_23 *
Perfect_Hash_23::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 3,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 7,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 7
    };

  static const struct word_type_23 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"send",  asio::ip::udp::socket::shutdown_send},
      {"both",  asio::ip::udp::socket::shutdown_both},
      {"",{}},
      {"receive",  asio::ip::udp::socket::shutdown_receive}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 4074 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int udp_socket_open(lua_State* L)
{
    lua_settop(L, 2);
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        sock->socket.open(asio::ip::udp::endpoint{*addr, 0}.protocol(), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TSTRING: {
        auto key = tostringview(L, 2);
        auto protocol =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_22::in_word_set((key).data(), (key).size()));

        if (!protocol) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio_error_code ec;
        sock->socket.open(*protocol, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int udp_socket_bind(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TNUMBER);
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        asio::ip::udp::endpoint ep(*addr, lua_tointeger(L, 3));
        asio_error_code ec;
        sock->socket.bind(ep, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    case LUA_TSTRING: {
        asio_error_code ec;
        auto addr = asio::ip::make_address(lua_tostring(L, 2), ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        asio::ip::udp::endpoint ep(addr, lua_tointeger(L, 3));
        sock->socket.bind(ep, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        return 0;
    }
    }
}

static int udp_socket_shutdown(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    auto what =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_23::in_word_set((key).data(), (key).size()));
    if (!what) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    asio_error_code ec;
    sock->socket.shutdown(*what, ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

static int udp_socket_disconnect(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

#if BOOST_OS_WINDOWS
    push(L, std::errc::function_not_supported);
    return lua_error(L);
#else
    struct sockaddr_in sin;
    std::memset(&sin, 0, sizeof(sin));
    sin.sin_family = AF_UNSPEC;
    int res = connect(
        sock->socket.native_handle(), reinterpret_cast<struct sockaddr*>(&sin),
        sizeof(sin));
    if (res == -1) {
        push(L, std::error_code{errno, std::system_category()});
        return lua_error(L);
    }

    return 0;
#endif // BOOST_OS_WINDOWS
}

} // namespace emilua

#line 4257 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_24  */
/* Computed positions: -k'2' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_24 { const char* name; asio::ip::udp action; };
/* maximum key range = 2, duplicates = 0 */

class Perfect_Hash_24
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_24 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_24::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 1, 2, 0, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
      2, 2, 2, 2, 2, 2
    };
  return asso_values[static_cast<unsigned char>(str[1])];
}

const struct word_type_24 *
Perfect_Hash_24::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 2,
      MIN_WORD_LENGTH = 2,
      MAX_WORD_LENGTH = 2,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 1
    };

  static const struct word_type_24 wordlist[] =
    {
      {"v6",  asio::ip::udp::v6()},
      {"v4",  asio::ip::udp::v4()}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 4374 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_25  */
/* Computed positions: -k'11' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_25 { const char* name; int (*action)(lua_State*, udp_socket*); };
/* maximum key range = 25, duplicates = 0 */

class Perfect_Hash_25
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_25 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_25::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30,  0,
      30,  0, 10, 30, 10,  0, 30, 30,  0, 30,
      30, 30,  5, 30,  0, 30,  0, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30, 30, 30, 30, 30,
      30, 30, 30, 30, 30, 30
    };
  unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[static_cast<unsigned char>(str[10])];
      /*FALLTHROUGH*/
      case 10:
      case 9:
      case 8:
      case 7:
      case 6:
      case 5:
        break;
    }
  return hval;
}

const struct word_type_25 *
Perfect_Hash_25::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 13,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 21,
      MIN_HASH_VALUE = 5,
      MAX_HASH_VALUE = 29
    };

  static const struct word_type_25 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"debug", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::debug o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"v6_only", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::ip::v6_only o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"broadcast", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::broadcast o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}},
      {"do_not_route", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::do_not_route o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"reuse_address", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::socket_base::reuse_address o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"multicast_loop", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TBOOLEAN);
                asio::ip::multicast::enable_loopback o(lua_toboolean(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"send_buffer_size", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::socket_base::send_buffer_size o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"unicast_hops", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::ip::unicast::hops o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}},
      {"multicast_interface", 
            [](lua_State* L, udp_socket* socket) -> int {
                auto addr = static_cast<asio::ip::address*>(
                    lua_touserdata(L, 3));
                if (!addr || !lua_getmetatable(L, 3)) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                if (!lua_rawequal(L, -1, -2)) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                if (!addr->is_v4()) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                asio::ip::multicast::outbound_interface o(addr->to_v4());
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"join_multicast_group", 
            [](lua_State* L, udp_socket* socket) -> int {
                auto addr = static_cast<asio::ip::address*>(
                    lua_touserdata(L, 3));
                if (!addr || !lua_getmetatable(L, 3)) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                if (!lua_rawequal(L, -1, -2)) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                asio::ip::multicast::join_group o(*addr);
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"leave_multicast_group", 
            [](lua_State* L, udp_socket* socket) -> int {
                auto addr = static_cast<asio::ip::address*>(
                    lua_touserdata(L, 3));
                if (!addr || !lua_getmetatable(L, 3)) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                if (!lua_rawequal(L, -1, -2)) {
                    push(L, std::errc::invalid_argument, "arg", 3);
                    return lua_error(L);
                }
                asio::ip::multicast::leave_group o(*addr);
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}},
      {"multicast_hops", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::ip::multicast::hops o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"receive_buffer_size", 
            [](lua_State* L, udp_socket* socket) -> int {
                luaL_checktype(L, 3, LUA_TNUMBER);
                asio::socket_base::receive_buffer_size o(lua_tointeger(L, 3));
                asio_error_code ec;
                socket->socket.set_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                return 0;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 4701 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_26  */
/* Computed positions: -k'11' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_26 { const char* name; int (*action)(lua_State*, udp_socket*); };
/* maximum key range = 20, duplicates = 0 */

class Perfect_Hash_26
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_26 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_26::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25,  0,  5, 25,  5, 25, 25, 25,  0, 25,
      25, 25,  5, 25,  0, 25,  0, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25, 25, 25, 25, 25,
      25, 25, 25, 25, 25, 25
    };
  unsigned int hval = len;

  switch (hval)
    {
      default:
        hval += asso_values[static_cast<unsigned char>(str[10])];
      /*FALLTHROUGH*/
      case 10:
      case 9:
      case 8:
      case 7:
      case 6:
      case 5:
        break;
    }
  return hval;
}

const struct word_type_26 *
Perfect_Hash_26::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 10,
      MIN_WORD_LENGTH = 5,
      MAX_WORD_LENGTH = 19,
      MIN_HASH_VALUE = 5,
      MAX_HASH_VALUE = 24
    };

  static const struct word_type_26 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"debug", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::debug o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}},
      {"v6_only", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::ip::v6_only o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}},
      {"broadcast", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::broadcast o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}},
      {"do_not_route", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::do_not_route o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"reuse_address", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::reuse_address o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"multicast_loop", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::ip::multicast::enable_loopback o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushboolean(L, o.value());
                return 1;
            }},
      {"",{}},
      {"send_buffer_size", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::send_buffer_size o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"unicast_hops", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::ip::unicast::hops o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"",{}},
      {"multicast_hops", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::ip::multicast::hops o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }},
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"receive_buffer_size", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::receive_buffer_size o;
                asio_error_code ec;
                socket->socket.get_option(o, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushinteger(L, o.value());
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 4957 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int udp_socket_close(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    sock->socket.close(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

static int udp_socket_cancel(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    sock->socket.cancel(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    return 0;
}

#if BOOST_OS_UNIX
static int udp_socket_assign(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto handle = static_cast<file_descriptor_handle*>(lua_touserdata(L, 3));
    if (!handle || !lua_getmetatable(L, 3)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    if (*handle == INVALID_FILE_DESCRIPTOR) {
        push(L, std::errc::device_or_resource_busy);
        return lua_error(L);
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TUSERDATA: {
        auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
        if (!addr || !lua_getmetatable(L, 2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        lua_pushnil(L);
        setmetatable(L, 3);

        asio_error_code ec;
        sock->socket.assign(
            asio::ip::udp::endpoint{*addr, 0}.protocol(), *handle, ec);
        assert(!ec); boost::ignore_unused(ec);

        return 0;
    }
    case LUA_TSTRING: {
        auto key = tostringview(L, 2);
        auto protocol =  ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_24::in_word_set((key).data(), (key).size()));

        if (!protocol) {
            push(L, std::errc::invalid_argument, "arg", 2);
            return lua_error(L);
        }

        lua_pushnil(L);
        setmetatable(L, 3);

        asio_error_code ec;
        sock->socket.assign(*protocol, *handle, ec);
        assert(!ec); boost::ignore_unused(ec);

        return 0;
    }
    }
}

static int udp_socket_release(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    asio_error_code ec;
    int rawfd = sock->socket.release(ec);
    BOOST_SCOPE_EXIT_ALL(&) {
        if (rawfd != INVALID_FILE_DESCRIPTOR) {
            int res = close(rawfd);
            boost::ignore_unused(res);
        }
    };

    if (ec) {
        push(L, ec);
        return lua_error(L);
    }

    auto fdhandle = static_cast<file_descriptor_handle*>(
        lua_newuserdata(L, sizeof(file_descriptor_handle))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &file_descriptor_mt_key);
    setmetatable(L, -2);

    *fdhandle = rawfd;
    rawfd = INVALID_FILE_DESCRIPTOR;
    return 1;
}
#endif // BOOST_OS_UNIX

static int udp_socket_set_option(lua_State* L)
{
    lua_settop(L, 3);
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_25::in_word_set((key).data(), (key).size()), [](lua_State* L, udp_socket*) -> int {
            push(L, std::errc::not_supported);
            return lua_error(L);
        })(L, socket);
}

static int udp_socket_get_option(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_26::in_word_set((key).data(), (key).size()), [](lua_State* L, udp_socket* socket) -> int {
            push(L, std::errc::not_supported);
            return lua_error(L);
        })(L, socket);
}

} // namespace emilua

#line 5168 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_27  */
/* Computed positions: -k'' */

struct word_type_27 { const char* name; asio::socket_base::message_flags action; };
/* maximum key range = 10, duplicates = 0 */

class Perfect_Hash_27
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_27 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_27::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_27 *
Perfect_Hash_27::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 13
    };

  static const struct word_type_27 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"peek",  asio::socket_base::message_peek},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"out_of_band",  asio::socket_base::message_out_of_band},
      {"do_not_route",  asio::socket_base::message_do_not_route},
      {"end_of_record",  asio::socket_base::message_end_of_record}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5233 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_28  */
/* Computed positions: -k'' */

struct word_type_28 { const char* name; asio::socket_base::message_flags action; };
/* maximum key range = 10, duplicates = 0 */

class Perfect_Hash_28
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_28 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_28::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_28 *
Perfect_Hash_28::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 13
    };

  static const struct word_type_28 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"peek",  asio::socket_base::message_peek},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"out_of_band",  asio::socket_base::message_out_of_band},
      {"do_not_route",  asio::socket_base::message_do_not_route},
      {"end_of_record",  asio::socket_base::message_end_of_record}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5298 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_29  */
/* Computed positions: -k'' */

struct word_type_29 { const char* name; asio::socket_base::message_flags action; };
/* maximum key range = 10, duplicates = 0 */

class Perfect_Hash_29
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_29 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_29::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_29 *
Perfect_Hash_29::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 13
    };

  static const struct word_type_29 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"peek",  asio::socket_base::message_peek},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"out_of_band",  asio::socket_base::message_out_of_band},
      {"do_not_route",  asio::socket_base::message_do_not_route},
      {"end_of_record",  asio::socket_base::message_end_of_record}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5363 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_30  */
/* Computed positions: -k'' */

struct word_type_30 { const char* name; asio::socket_base::message_flags action; };
/* maximum key range = 10, duplicates = 0 */

class Perfect_Hash_30
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_30 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_30::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_30 *
Perfect_Hash_30::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 4,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 13,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 13
    };

  static const struct word_type_30 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"peek",  asio::socket_base::message_peek},
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"out_of_band",  asio::socket_base::message_out_of_band},
      {"do_not_route",  asio::socket_base::message_do_not_route},
      {"end_of_record",  asio::socket_base::message_end_of_record}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5428 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_31  */
/* Computed positions: -k'' */

struct word_type_31 { const char* name; int (*action)(lua_State*, udp_socket*); };
/* maximum key range = 1, duplicates = 0 */

class Perfect_Hash_31
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_31 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_31::hash (const char *str, size_t len)
{
  return 0;
}

const struct word_type_31 *
Perfect_Hash_31::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 1,
      MIN_WORD_LENGTH = 14,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 0,
      MAX_HASH_VALUE = 0
    };

  static const struct word_type_31 wordlist[] =
    {
      {"bytes_readable", 
            [](lua_State* L, udp_socket* socket) -> int {
                asio::socket_base::bytes_readable command;
                asio_error_code ec;
                socket->socket.io_control(command, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                lua_pushnumber(L, command.get());
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5499 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua {


static int udp_socket_io_control(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto socket = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!socket || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    return  ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_31::in_word_set((key).data(), (key).size()), [](lua_State* L, udp_socket*) -> int {
            push(L, std::errc::not_supported);
            return lua_error(L);
        })(L, socket);
}

inline int udp_socket_is_open(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    lua_pushboolean(L, sock->socket.is_open());
    return 1;
}

inline int udp_socket_local_address(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.local_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    auto addr = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (addr) asio::ip::address{ep.address()};
    return 1;
}

inline int udp_socket_local_port(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.local_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    lua_pushinteger(L, ep.port());
    return 1;
}

inline int udp_socket_remote_address(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.remote_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    auto addr = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (addr) asio::ip::address{ep.address()};
    return 1;
}

inline int udp_socket_remote_port(lua_State* L)
{
    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    asio_error_code ec;
    auto ep = sock->socket.remote_endpoint(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    lua_pushinteger(L, ep.port());
    return 1;
}

} // namespace emilua

#line 5596 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_32  */
/* Computed positions: -k'1,3' */

#if !((' ' == 32) && ('!' == 33) && ('"' == 34) && ('#' == 35) \
      && ('%' == 37) && ('&' == 38) && ('\'' == 39) && ('(' == 40) \
      && (')' == 41) && ('*' == 42) && ('+' == 43) && (',' == 44) \
      && ('-' == 45) && ('.' == 46) && ('/' == 47) && ('0' == 48) \
      && ('1' == 49) && ('2' == 50) && ('3' == 51) && ('4' == 52) \
      && ('5' == 53) && ('6' == 54) && ('7' == 55) && ('8' == 56) \
      && ('9' == 57) && (':' == 58) && (';' == 59) && ('<' == 60) \
      && ('=' == 61) && ('>' == 62) && ('?' == 63) && ('A' == 65) \
      && ('B' == 66) && ('C' == 67) && ('D' == 68) && ('E' == 69) \
      && ('F' == 70) && ('G' == 71) && ('H' == 72) && ('I' == 73) \
      && ('J' == 74) && ('K' == 75) && ('L' == 76) && ('M' == 77) \
      && ('N' == 78) && ('O' == 79) && ('P' == 80) && ('Q' == 81) \
      && ('R' == 82) && ('S' == 83) && ('T' == 84) && ('U' == 85) \
      && ('V' == 86) && ('W' == 87) && ('X' == 88) && ('Y' == 89) \
      && ('Z' == 90) && ('[' == 91) && ('\\' == 92) && (']' == 93) \
      && ('^' == 94) && ('_' == 95) && ('a' == 97) && ('b' == 98) \
      && ('c' == 99) && ('d' == 100) && ('e' == 101) && ('f' == 102) \
      && ('g' == 103) && ('h' == 104) && ('i' == 105) && ('j' == 106) \
      && ('k' == 107) && ('l' == 108) && ('m' == 109) && ('n' == 110) \
      && ('o' == 111) && ('p' == 112) && ('q' == 113) && ('r' == 114) \
      && ('s' == 115) && ('t' == 116) && ('u' == 117) && ('v' == 118) \
      && ('w' == 119) && ('x' == 120) && ('y' == 121) && ('z' == 122) \
      && ('{' == 123) && ('|' == 124) && ('}' == 125) && ('~' == 126))
/* The character set is not based on ISO-646.  */
#error "gperf generated tables don't work with this execution character set. Please report a bug to <bug-gperf@gnu.org>."
#endif

struct word_type_32 { const char* name; int (*action)(lua_State*); };
/* maximum key range = 32, duplicates = 0 */

class Perfect_Hash_32
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_32 *in_word_set (const char *str, size_t len);
};

inline unsigned int
Perfect_Hash_32::hash (const char *str, size_t len)
{
  static const unsigned char asso_values[] =
    {
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 15, 36,  0, 10, 10,
      20,  5, 36, 15, 36, 10, 36, 36,  0,  5,
       0,  0, 36, 36,  5,  0,  0,  0, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36, 36, 36, 36, 36,
      36, 36, 36, 36, 36, 36
    };
  return len + asso_values[static_cast<unsigned char>(str[2])] + asso_values[static_cast<unsigned char>(str[0])];
}

const struct word_type_32 *
Perfect_Hash_32::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 21,
      MIN_WORD_LENGTH = 4,
      MAX_WORD_LENGTH = 14,
      MIN_HASH_VALUE = 4,
      MAX_HASH_VALUE = 35
    };

  static const struct word_type_32 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"send", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &udp_socket_send_key);
                return 1;
            }},
      {"",{}},
      {"assign", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, udp_socket_assign);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"send_to", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &udp_socket_send_to_key);
                return 1;
            }},
      {"shutdown", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_shutdown);
                return 1;
            }},
      {"open", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_open);
                return 1;
            }},
      {"set_option", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_set_option);
                return 1;
            }},
      {"",{}},
      {"release", 
            [](lua_State* L) -> int {
#if BOOST_OS_UNIX
                lua_pushcfunction(L, udp_socket_release);
#else // BOOST_OS_UNIX
                lua_pushcfunction(L, throw_enosys);
#endif // BOOST_OS_UNIX
                return 1;
            }},
      {"",{}},
      {"bind", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_bind);
                return 1;
            }},
      {"close", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_close);
                return 1;
            }},
      {"cancel", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_cancel);
                return 1;
            }},
      {"connect", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &udp_socket_connect_key);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"local_port",  udp_socket_local_port},
      {"remote_port",  udp_socket_remote_port},
      {"receive", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &udp_socket_receive_key);
                return 1;
            }},
      {"local_address",  udp_socket_local_address},
      {"remote_address",  udp_socket_remote_address},
      {"get_option", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_get_option);
                return 1;
            }},
      {"",{}},
      {"receive_from", 
            [](lua_State* L) -> int {
                rawgetp(L, LUA_REGISTRYINDEX, &udp_socket_receive_from_key);
                return 1;
            }},
      {"",{}}, {"",{}},
      {"disconnect", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_disconnect);
                return 1;
            }},
      {"",{}},
      {"is_open",  udp_socket_is_open},
      {"",{}}, {"",{}},
      {"io_control", 
            [](lua_State* L) -> int {
                lua_pushcfunction(L, udp_socket_io_control);
                return 1;
            }}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5812 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_33  */
/* Computed positions: -k'' */

struct word_type_33 { const char* name; asio::ip::resolver_base::flags action; };
/* maximum key range = 12, duplicates = 0 */

class Perfect_Hash_33
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_33 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_33::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_33 *
Perfect_Hash_33::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 18,
      MIN_HASH_VALUE = 7,
      MAX_HASH_VALUE = 18
    };

  static const struct word_type_33 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"passive",  asio::ip::resolver_base::passive},
      {"",{}},
      {"v4_mapped",  asio::ip::resolver_base::v4_mapped},
      {"",{}}, {"",{}},
      {"all_matching",  asio::ip::resolver_base::all_matching},
      {"",{}},
      {"canonical_name",  asio::ip::resolver_base::canonical_name},
      {"",{}}, {"",{}}, {"",{}},
      {"address_configured", 
                    asio::ip::resolver_base::address_configured}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5883 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_34  */
/* Computed positions: -k'' */

struct word_type_34 { const char* name; asio::ip::resolver_base::flags action; };
/* maximum key range = 12, duplicates = 0 */

class Perfect_Hash_34
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_34 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_34::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_34 *
Perfect_Hash_34::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 18,
      MIN_HASH_VALUE = 7,
      MAX_HASH_VALUE = 18
    };

  static const struct word_type_34 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"passive",  asio::ip::resolver_base::passive},
      {"",{}},
      {"v4_mapped",  asio::ip::resolver_base::v4_mapped},
      {"",{}}, {"",{}},
      {"all_matching",  asio::ip::resolver_base::all_matching},
      {"",{}},
      {"canonical_name",  asio::ip::resolver_base::canonical_name},
      {"",{}}, {"",{}}, {"",{}},
      {"address_configured", 
                    asio::ip::resolver_base::address_configured}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail

#line 5954 "libemilua.so.0.11.5.p/bc2_ip.cpp"
namespace emilua::gperf::detail {
namespace {
/* C++ code produced by gperf version 3.1 */
/* Command-line: /usr/bin/gperf --language=C++ --enum --readonly-tables --struct-type --initializer-suffix=',{}' --class-name=Perfect_Hash_35  */
/* Computed positions: -k'' */

struct word_type_35 { const char* name; asio::ip::resolver_base::flags action; };
/* maximum key range = 12, duplicates = 0 */

class Perfect_Hash_35
{
private:
  static inline unsigned int hash (const char *str, size_t len);
public:
  static const struct word_type_35 *in_word_set (const char *str, size_t len);
};

inline /*ARGSUSED*/
unsigned int
Perfect_Hash_35::hash (const char *str, size_t len)
{
  return len;
}

const struct word_type_35 *
Perfect_Hash_35::in_word_set (const char *str, size_t len)
{
  enum
    {
      TOTAL_KEYWORDS = 5,
      MIN_WORD_LENGTH = 7,
      MAX_WORD_LENGTH = 18,
      MIN_HASH_VALUE = 7,
      MAX_HASH_VALUE = 18
    };

  static const struct word_type_35 wordlist[] =
    {
      {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}}, {"",{}},
      {"",{}},
      {"passive",  asio::ip::resolver_base::passive},
      {"",{}},
      {"v4_mapped",  asio::ip::resolver_base::v4_mapped},
      {"",{}}, {"",{}},
      {"all_matching",  asio::ip::resolver_base::all_matching},
      {"",{}},
      {"canonical_name",  asio::ip::resolver_base::canonical_name},
      {"",{}}, {"",{}}, {"",{}},
      {"address_configured", 
                    asio::ip::resolver_base::address_configured}
    };

  if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH)
    {
      unsigned int key = hash (str, len);

      if (key <= MAX_HASH_VALUE)
        {
          const char *s = wordlist[key].name;

          if (*str == *s && !strcmp (str + 1, s + 1))
            return &wordlist[key];
        }
    }
  return 0;
}

} // namespace
} // namespace emilua::gperf::detail
#line 1 "../src/ip.cpp"
// Copyright (c) 2020, 2024 Vinícius dos Santos Oliveira
// SPDX-License-Identifier: MIT OR BSL-1.0

#line 52 "../src/ip.cpp"


namespace emilua {

extern unsigned char ip_connect_bytecode[];
extern std::size_t ip_connect_bytecode_size;
extern unsigned char ip_dial_bytecode[];
extern std::size_t ip_dial_bytecode_size;

char ip_key;
char ip_address_mt_key;
char ip_tcp_socket_mt_key;
char ip_tcp_acceptor_mt_key;
char ip_udp_socket_mt_key;

#line 82 "../src/ip.cpp"


#line 89 "../src/ip.cpp"


#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
struct get_address_info_context_t: public pending_operation
{
    get_address_info_context_t(asio::io_context& ctx)
        : pending_operation{/*shared_ownership=*/true}
        , hCompletion(ctx)
    {
        ZeroMemory(&overlapped, sizeof(OVERLAPPED));
    }

    void cancel() noexcept override
    {
        GetAddrInfoExCancel(&hCancel);
    }

    ~get_address_info_context_t()
    {
        // WARNING: If you call `asio::io_context::stop()` the handler will be
        // destroyed before `hCompletion->native_handle()` has been set (and
        // then we hit this very destructor). The underlying asynchronous
        // operation will not stop along with it. The underlying asynchronous
        // operation will be writing to this freed memory and a random memory
        // corruption will occur. If we're **lucky** this `assert()` will
        // trigger.
        //
        // Are you reading this code because the `assert()` triggered? Remove
        // the call to `asio::io_context::stop()` from your code to fix the
        // problem.
        //
        // A different approach would be to use the `lpCompletionRoutine`
        // parameter from `GetAddrInfoExW()` and take care of all Boost.Asio
        // integration ourselves (i.e. skip `asio::windows::object_handle`
        // altogether). However (1) this would be more work and (2) so far
        // Emilua has not been designed to rely on `asio::io_context::stop()` to
        // begin with. Therefore we can safely use the simpler approach
        // (i.e. remove `asio::io_context::stop()` from your new buggy code).
        //
        // Maybe you'll think that yet another approach would be to cancel the
        // operation right here and sync with it. This approach would have other
        // problems for which I won't write any prose.
        assert(results == nullptr);
    }

    asio::windows::object_handle hCompletion;
    OVERLAPPED overlapped;
    PADDRINFOEXW results = nullptr;
    HANDLE hCancel = nullptr;
};
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

struct resolver_service: public pending_operation
{
    resolver_service(asio::io_context& ioctx)
        : pending_operation{/*shared_ownership=*/false}
        , tcp_resolver{ioctx}
        , udp_resolver{ioctx}
    {}

    void cancel() noexcept override
    {}

    asio::ip::tcp::resolver tcp_resolver;
    asio::ip::udp::resolver udp_resolver;
};

static int ip_host_name(lua_State* L)
{
    asio_error_code ec;
    auto val = asio::ip::host_name(ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }
    push(L, val);
    return 1;
}

static int ip_tostring(lua_State* L)
{
    int nargs = lua_gettop(L);

    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto ret = a->to_string();

    switch (nargs) {
    case 1:
        break;
    case 2: {
        std::uint16_t port = luaL_checkinteger(L, 2);
        std::array<char, 5> buf;
        auto s_size = std::to_chars(
            buf.data(),
            buf.data() + buf.size(),
            port).ptr - buf.data();
        if (a->is_v4()) {
            ret.reserve(ret.size() + 1 + s_size);
            ret.push_back(':');
        } else {
            ret.reserve(ret.size() + 3 + s_size);
            ret.insert(ret.begin(), '[');
            ret.append("]:");
        }
        ret.append(buf.data(), s_size);
        break;
    }
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    push(L, ret);
    return 1;
}

static int ip_toendpoint(lua_State* L)
{
    luaL_checktype(L, 1, LUA_TSTRING);

    std::string_view host = tostringview(L, 1);
    std::uint16_t port;

    {
        auto idx = host.rfind(':');
        if (idx == std::string_view::npos) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        std::string_view p = host.substr(idx + 1);
        if (p.starts_with("0") && p.size() != 1) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        auto res = std::from_chars(p.data(), p.data() + p.size(), port);
        if (res.ec != std::errc{}) {
            push(L, res.ec, "arg", 1);
            return lua_error(L);
        } else if (res.ptr != p.data() + p.size()) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        host.remove_suffix(p.size() + 1);
    }

    bool is_ipv6 = false;
    if (host.starts_with("[")) {
        if (host.back() != ']') {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        host.remove_suffix(1);
        host.remove_prefix(1);
        is_ipv6 = true;
    }

    auto addr = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);

    asio_error_code ec;
    new (addr) asio::ip::address{asio::ip::make_address(host, ec)};
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }

    if ((is_ipv6 && addr->is_v4()) || (!is_ipv6 && addr->is_v6())) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    lua_pushinteger(L, port);

    return 2;
}

static int ip_toendpoint2(lua_State* L)
{
    luaL_checktype(L, 1, LUA_TSTRING);

    std::string_view host = tostringview(L, 1);
    std::string_view port;

    {
        auto idx = host.rfind(':');
        if (idx == std::string_view::npos) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        port = host.substr(idx + 1);
        if (port.starts_with("0") && port.size() != 1) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        host.remove_suffix(port.size() + 1);
    }

    bool is_ipv6 = false;
    if (host.starts_with("[")) {
        if (host.back() != ']') {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        host.remove_suffix(1);
        host.remove_prefix(1);
        is_ipv6 = true;
    }

    asio_error_code ec;
    asio::ip::address addr{asio::ip::make_address(host, ec)};
    if (!ec) {
        if ((is_ipv6 && addr.is_v4()) || (!is_ipv6 && addr.is_v6())) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
    }

    push(L, host);
    push(L, port);
    return 2;
}

static int address_new(lua_State* L)
{
    lua_settop(L, 1);
    auto a = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    switch (lua_type(L, 1)) {
    case LUA_TNIL:
        new (a) asio::ip::address{};
        break;
    case LUA_TSTRING: {
        asio_error_code ec;
        new (a) asio::ip::address{
            asio::ip::make_address(lua_tostring(L, 1), ec)
        };
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        break;
    }
    default:
        static_assert(std::is_trivially_destructible_v<asio::ip::address>);
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    return 1;
}

static int address_any_v4(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::address{asio::ip::address_v4::any()};
    return 1;
}

static int address_any_v6(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::address{asio::ip::address_v6::any()};
    return 1;
}

static int address_loopback_v4(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::address{asio::ip::address_v4::loopback()};
    return 1;
}

static int address_loopback_v6(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::address{asio::ip::address_v6::loopback()};
    return 1;
}

static int address_broadcast_v4(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(
        lua_newuserdata(L, sizeof(asio::ip::address))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::address{asio::ip::address_v4::broadcast()};
    return 1;
}

#line 615 "../src/ip.cpp"


static int address_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 6381 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_1::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 658 "../src/ip.cpp"
(L);
}

#line 678 "../src/ip.cpp"


static int address_mt_newindex(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 6397 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_2::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 690 "../src/ip.cpp"
(L);
}

static int address_mt_tostring(lua_State* L)
{
    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    auto ret = a->to_string();
    lua_pushlstring(L, ret.data(), ret.size());
    return 1;
}

static int address_mt_eq(lua_State* L)
{
    auto a1 = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    auto a2 = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
    lua_pushboolean(L, *a1 == *a2);
    return 1;
}

static int address_mt_lt(lua_State* L)
{
    auto a1 = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a1 || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto a2 = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
    if (!a2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    lua_pushboolean(L, *a1 < *a2);
    return 1;
}

static int address_mt_le(lua_State* L)
{
    auto a1 = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a1 || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto a2 = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
    if (!a2 || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    lua_pushboolean(L, *a1 <= *a2);
    return 1;
}

static int tcp_socket_new(lua_State* L)
{
    auto& vm_ctx = get_vm_context(L);
    auto a = static_cast<tcp_socket*>(lua_newuserdata(L, sizeof(tcp_socket)));
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    setmetatable(L, -2);
    new (a) tcp_socket{vm_ctx.strand().context()};
    return 1;
}

#line 1196 "../src/ip.cpp"


static int tcp_socket_connect(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TNUMBER);
    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
    if (!a || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::ip::tcp::endpoint ep{
        *a, static_cast<std::uint16_t>(lua_tointeger(L, 3))};

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_connect(ep, asio::bind_cancellation_slot(cancel_slot,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,s](const asio_error_code& ec) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(opt_args, hana::make_tuple(ec))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int tcp_socket_read_some(lua_State* L)
{
    lua_settop(L, 2);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_read_some(
        asio::buffer(bs->data.get(), bs->size),
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,s](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                boost::ignore_unused(buf);
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int tcp_socket_write_some(lua_State* L)
{
    lua_settop(L, 2);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_write_some(
        asio::buffer(bs->data.get(), bs->size),
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,s](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                boost::ignore_unused(buf);
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int tcp_socket_receive(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TTABLE);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::socket_base::message_flags flags = {};
    for (int i = 1 ;; ++i) {
        lua_rawgeti(L, 3, i);
        switch (lua_type(L, -1)) {
        default:
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        case LUA_TNIL:
            lua_pop(L, 1);
            goto end_for;
        case LUA_TSTRING:
            break;
        }

        auto s = tostringview(L);
        lua_pop(L, 1);
        auto f = 
#line 6717 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_7::in_word_set((s).data(), (s).size()), {})
#line 1433 "../src/ip.cpp"
;
        if (f == 0) {
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        }
        flags |= f;
    }
 end_for:

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_receive(
        asio::buffer(bs->data.get(), bs->size),
        flags,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,s](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                boost::ignore_unused(buf);
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int tcp_socket_send(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TTABLE);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::socket_base::message_flags flags = {};
    for (int i = 1 ;; ++i) {
        lua_rawgeti(L, 3, i);
        switch (lua_type(L, -1)) {
        default:
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        case LUA_TNIL:
            lua_pop(L, 1);
            goto end_for;
        case LUA_TSTRING:
            break;
        }

        auto s = tostringview(L);
        lua_pop(L, 1);
        auto f = 
#line 6808 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_8::in_word_set((s).data(), (s).size()), {})
#line 1530 "../src/ip.cpp"
;
        if (f == 0) {
            push(L, std::errc::invalid_argument, "arg", 3);
            return lua_error(L);
        }
        flags |= f;
    }
 end_for:

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_send(
        asio::buffer(bs->data.get(), bs->size),
        flags,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,s](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                boost::ignore_unused(buf);
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

#if BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
// https://www.boost.org/doc/libs/1_78_0/doc/html/boost_asio/example/cpp03/windows/transmit_file.cpp
static int tcp_socket_send_file(lua_State* L)
{
    lua_settop(L, 7);
    luaL_checktype(L, 3, LUA_TNUMBER);
    luaL_checktype(L, 4, LUA_TNUMBER);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto sock = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto file = static_cast<asio::random_access_file*>(
        lua_touserdata(L, 2));
    if (!file || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &file_random_access_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    TRANSMIT_FILE_BUFFERS transmitBuffers;
    std::shared_ptr<unsigned char[]> buf1, buf2;
    lua_Integer n_number_of_bytes_per_send;

    if (lua_type(L, 5) != LUA_TNIL) {
        auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 5));
        if (!bs || !lua_getmetatable(L, 5)) {
            push(L, std::errc::invalid_argument, "arg", 5);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 5);
            return lua_error(L);
        }
        buf1 = bs->data;
        transmitBuffers.Head = bs->data.get();
        transmitBuffers.HeadLength = bs->size;
    } else {
        transmitBuffers.Head = NULL;
        transmitBuffers.HeadLength = 0;
    }

    if (lua_type(L, 6) != LUA_TNIL) {
        auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 6));
        if (!bs || !lua_getmetatable(L, 6)) {
            push(L, std::errc::invalid_argument, "arg", 6);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 6);
            return lua_error(L);
        }
        buf2 = bs->data;
        transmitBuffers.Tail = bs->data.get();
        transmitBuffers.TailLength = bs->size;
    } else {
        transmitBuffers.Tail = NULL;
        transmitBuffers.TailLength = 0;
    }

    switch (lua_type(L, 7)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 7);
        return lua_error(L);
    case LUA_TNUMBER:
        n_number_of_bytes_per_send = lua_tointeger(L, 7);
        break;
    case LUA_TNIL:
        n_number_of_bytes_per_send = 0;
    }

    asio::windows::overlapped_ptr overlapped{
        vm_ctx->strand_using_defer(),
        [vm_ctx,current_fiber,buf1,buf2,sock](
            const asio_error_code& ec, std::size_t bytes_transferred
        ) {
            if (!vm_ctx->valid())
                return;

            --sock->nbusy;

            vm_ctx->fiber_resume(
                current_fiber,
                hana::make_set(
                    vm_context::options::auto_detect_interrupt,
                    hana::make_pair(
                        vm_context::options::arguments,
                        hana::make_tuple(ec, bytes_transferred))));
        }
    };

    DWORD64 offset = lua_tointeger(L, 3);
    overlapped.get()->Offset = static_cast<DWORD>(offset);
    overlapped.get()->OffsetHigh = static_cast<DWORD>(offset >> 32);

    lua_pushvalue(L, 1);
    lua_pushlightuserdata(L, overlapped.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto sock = static_cast<tcp_socket*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            CancelIoEx(
                reinterpret_cast<HANDLE>(
                    static_cast<SOCKET>(sock->socket.native_handle())),
                static_cast<LPOVERLAPPED>(
                    lua_touserdata(L, lua_upvalueindex(2))));
            return 0;
        },
        2);
    set_interrupter(L, *vm_ctx);

    ++sock->nbusy;
    BOOL ok = TransmitFile(sock->socket.native_handle(), file->native_handle(),
                           /*nNumberOfBytesToWrite=*/lua_tointeger(L, 4),
                           n_number_of_bytes_per_send,
                           overlapped.get(), &transmitBuffers,
                           /*dwReserved=*/0);
    DWORD last_error = GetLastError();

    // Check if the operation completed immediately.
    if (!ok && last_error != ERROR_IO_PENDING) {
        // The operation completed immediately, so a completion notification
        // needs to be posted. When complete() is called, ownership of the
        // OVERLAPPED-derived object passes to the io_context.
        asio_error_code ec(last_error, asio::error::get_system_category());
        overlapped.complete(ec, 0);
    } else {
        // The operation was successfully initiated, so ownership of the
        // OVERLAPPED-derived object has passed to the io_context.
        overlapped.release();
    }

    return lua_yield(L, 0);
}
#endif // BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO

static int tcp_socket_wait(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<tcp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    auto wait_type = 
#line 7027 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_9::in_word_set((key).data(), (key).size()))
#line 1750 "../src/ip.cpp"
;
    if (!wait_type) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_wait(
        *wait_type,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,s](const asio_error_code& ec) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args, hana::make_tuple(ec))));
            }
        ))
    );

    return lua_yield(L, 0);
}

#line 2507 "../src/ip.cpp"


static int tcp_socket_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 7070 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_12::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 2644 "../src/ip.cpp"
(L);
}

static int tcp_acceptor_new(lua_State* L)
{
    auto& vm_ctx = get_vm_context(L);
    auto a = static_cast<asio::ip::tcp::acceptor*>(
        lua_newuserdata(L, sizeof(asio::ip::tcp::acceptor))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::tcp::acceptor{vm_ctx.strand().context()};
    return 1;
}

static int tcp_acceptor_accept(lua_State* L)
{
    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    acceptor->async_accept(
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber](const asio_error_code& ec,
                                   asio::ip::tcp::socket peer) {
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(
                                ec,
                                [&ec,&peer](lua_State* fiber) {
                                    if (ec) {
                                        lua_pushnil(fiber);
                                    } else {
                                        auto s = static_cast<tcp_socket*>(
                                            lua_newuserdata(
                                                fiber, sizeof(tcp_socket)));
                                        rawgetp(fiber, LUA_REGISTRYINDEX,
                                                &ip_tcp_socket_mt_key);
                                        setmetatable(fiber, -2);
                                        new (s) tcp_socket{std::move(peer)};
                                    }
                                }))
                    ));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int tcp_acceptor_wait(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TSTRING);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto acceptor = static_cast<asio::ip::tcp::acceptor*>(lua_touserdata(L, 1));
    if (!acceptor || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto key = tostringview(L, 2);
    auto wait_type = 
#line 7165 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::make_optional(::emilua::gperf::detail::Perfect_Hash_13::in_word_set((key).data(), (key).size()))
#line 2737 "../src/ip.cpp"
;
    if (!wait_type) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    acceptor->async_wait(
        *wait_type,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber](const asio_error_code& ec) {
                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            opt_args, hana::make_tuple(ec))));
            }
        ))
    );

    return lua_yield(L, 0);
}

#line 3413 "../src/ip.cpp"


static int tcp_acceptor_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 7202 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_18::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 3501 "../src/ip.cpp"
(L);
}

static int tcp_get_address_info(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    asio::ip::resolver_base::flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        flags = asio::ip::resolver_base::address_configured |
            asio::ip::resolver_base::v4_mapped;
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 7245 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_19::in_word_set((s).data(), (s).size()), {})
#line 3550 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    std::string host;
    switch (lua_type(L, 1)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    case LUA_TSTRING:
        host = tostringview(L, 1);
        break;
    case LUA_TUSERDATA: {
        if (!lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        auto& a = *static_cast<asio::ip::address*>(lua_touserdata(L, 1));
        host = a.to_string();
        flags |= asio::ip::resolver_base::numeric_host;
    }
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TSTRING:
        break;
    case LUA_TNUMBER:
        flags |= asio::ip::resolver_base::numeric_service;
    }

#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    ADDRINFOEXW hints;
    hints.ai_flags = static_cast<int>(flags);
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_addrlen = 0;
    hints.ai_canonname = nullptr;
    hints.ai_addr = nullptr;
    hints.ai_blob = nullptr;
    hints.ai_bloblen = 0;
    hints.ai_provider = nullptr;
    hints.ai_next = nullptr;

    auto query_ctx = std::make_shared<get_address_info_context_t>(
        vm_ctx->strand().context());

    {
        HANDLE hCompletion = CreateEventA(/*lpEventAttributes=*/NULL,
                                          /*bManualReset=*/TRUE,
                                          /*bInitialState=*/FALSE,
                                          /*lpName=*/NULL);
        if (hCompletion == NULL) {
            asio_error_code ec(
                GetLastError(), asio::error::get_system_category());
            push(L, ec);
            return lua_error(L);
        }

        asio_error_code ec;
        query_ctx->hCompletion.assign(hCompletion, ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        query_ctx->overlapped.hEvent = hCompletion;
    }

    auto service = tostringview(L, 2);
    INT error = GetAddrInfoExW(nowide::widen(host).c_str(),
                               nowide::widen(service).c_str(), NS_DNS,
                               /*lpNspId=*/NULL, &hints, &query_ctx->results,
                               /*timeout=*/NULL, &query_ctx->overlapped,
                               /*lpCompletionRoutine=*/NULL,
                               &query_ctx->hCancel);
    if (error != WSA_IO_PENDING) {
        // the operation completed immediately
        asio_error_code ec(
            WSAGetLastError(), asio::error::get_system_category());
        push(L, ec);
        return lua_error(L);
    }

    lua_pushlightuserdata(L, query_ctx.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto query_ctx = static_cast<get_address_info_context_t*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            GetAddrInfoExCancel(&query_ctx->hCancel);
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    vm_ctx->pending_operations.push_back(*query_ctx);
    query_ctx->hCompletion.async_wait(
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,query_ctx](asio_error_code ec) {
                BOOST_SCOPE_EXIT_ALL(&) {
                    if (query_ctx->results) {
                        FreeAddrInfoExW(query_ctx->results);
#ifndef NDEBUG
                        query_ctx->results = nullptr;
#endif // !defined(NDEBUG)
                    }
                };

                if (vm_ctx->valid()) {
                    vm_ctx->pending_operations.erase(
                        vm_ctx->pending_operations.iterator_to(*query_ctx));
                }

                if (!ec) {
                    INT error = GetAddrInfoExOverlappedResult(
                        &query_ctx->overlapped);
                    if (error == WSA_E_CANCELLED) {
                        ec = asio::error::operation_aborted;
                    } else {
                        ec = asio_error_code(
                            error, asio::error::get_system_category());
                    }
                }

                auto push_results = [&ec,&query_ctx](lua_State* L) {
                    if (ec) {
                        lua_pushnil(L);
                        return;
                    }

                    lua_newtable(L);
                    lua_pushliteral(L, "address");
                    lua_pushliteral(L, "port");
                    lua_pushliteral(L, "canonical_name");

                    if (
                        query_ctx->results && query_ctx->results->ai_canonname
                    ) {
                        std::wstring_view w{query_ctx->results->ai_canonname};
                        push(L, nowide::narrow(w));
                    } else {
                        lua_pushnil(L);
                    }

                    int i = 1;
                    for (
                        auto it = query_ctx->results ; it ; it = it->ai_next
                    ) {
                        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

                        asio::ip::tcp::endpoint ep;
                        ep.resize(it->ai_addrlen);
                        std::memcpy(ep.data(), it->ai_addr, ep.size());

                        lua_pushvalue(L, -1 -4);
                        auto a = static_cast<asio::ip::address*>(
                            lua_newuserdata(L, sizeof(asio::ip::address))
                        );
                        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                        setmetatable(L, -2);
                        new (a) asio::ip::address{ep.address()};
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -3);
                        lua_pushinteger(L, ep.port());
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -2);
                        lua_pushvalue(L, -1 -1);
                        lua_rawset(L, -3);

                        lua_rawseti(L, -6, i++);
                    }
                    lua_pop(L, 4);
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::fast_auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))));
            }));
#else // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->tcp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    bool has_cname = (flags & asio::ip::resolver_base::canonical_name) ?
        true : false;

    service->tcp_resolver.async_resolve(
        host, tostringview(L, 2), flags,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,has_cname](
                const asio_error_code& ec,
                asio::ip::tcp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results,has_cname](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "address");
                        lua_pushliteral(fib, "port");
                        lua_pushliteral(fib, "canonical_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/3);

                            lua_pushvalue(fib, -1 -3);
                            auto a = static_cast<asio::ip::address*>(
                                lua_newuserdata(fib, sizeof(asio::ip::address))
                            );
                            rawgetp(fib, LUA_REGISTRYINDEX, &ip_address_mt_key);
                            setmetatable(fib, -2);
                            new (a) asio::ip::address{res.endpoint().address()};
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -2);
                            lua_pushinteger(fib, res.endpoint().port());
                            lua_rawset(fib, -3);

                            if (has_cname) {
                                lua_pushvalue(fib, -1 -1);
                                push(fib, res.host_name());
                                lua_rawset(fib, -3);
                            }

                            lua_rawseti(fib, -5, i++);
                        }
                        lua_pop(fib, 3);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

    return lua_yield(L, 0);
}

static int tcp_get_address_v4_info(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    asio::ip::resolver_base::flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 7574 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_20::in_word_set((s).data(), (s).size()), {})
#line 3888 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    std::string host;
    switch (lua_type(L, 1)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    case LUA_TSTRING:
        host = tostringview(L, 1);
        break;
    case LUA_TUSERDATA: {
        if (!lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        auto& a = *static_cast<asio::ip::address*>(lua_touserdata(L, 1));
        host = a.to_string();
        flags |= asio::ip::resolver_base::numeric_host;
    }
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TSTRING:
        break;
    case LUA_TNUMBER:
        flags |= asio::ip::resolver_base::numeric_service;
    }

#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    ADDRINFOEXW hints;
    hints.ai_flags = static_cast<int>(flags);
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_addrlen = 0;
    hints.ai_canonname = nullptr;
    hints.ai_addr = nullptr;
    hints.ai_blob = nullptr;
    hints.ai_bloblen = 0;
    hints.ai_provider = nullptr;
    hints.ai_next = nullptr;

    auto query_ctx = std::make_shared<get_address_info_context_t>(
        vm_ctx->strand().context());

    {
        HANDLE hCompletion = CreateEventA(/*lpEventAttributes=*/NULL,
                                          /*bManualReset=*/TRUE,
                                          /*bInitialState=*/FALSE,
                                          /*lpName=*/NULL);
        if (hCompletion == NULL) {
            asio_error_code ec(
                GetLastError(), asio::error::get_system_category());
            push(L, ec);
            return lua_error(L);
        }

        asio_error_code ec;
        query_ctx->hCompletion.assign(hCompletion, ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        query_ctx->overlapped.hEvent = hCompletion;
    }

    auto service = tostringview(L, 2);
    INT error = GetAddrInfoExW(nowide::widen(host).c_str(),
                               nowide::widen(service).c_str(), NS_DNS,
                               /*lpNspId=*/NULL, &hints, &query_ctx->results,
                               /*timeout=*/NULL, &query_ctx->overlapped,
                               /*lpCompletionRoutine=*/NULL,
                               &query_ctx->hCancel);
    if (error != WSA_IO_PENDING) {
        // the operation completed immediately
        asio_error_code ec(
            WSAGetLastError(), asio::error::get_system_category());
        push(L, ec);
        return lua_error(L);
    }

    lua_pushlightuserdata(L, query_ctx.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto query_ctx = static_cast<get_address_info_context_t*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            GetAddrInfoExCancel(&query_ctx->hCancel);
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    vm_ctx->pending_operations.push_back(*query_ctx);
    query_ctx->hCompletion.async_wait(
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,query_ctx](asio_error_code ec) {
                BOOST_SCOPE_EXIT_ALL(&) {
                    if (query_ctx->results) {
                        FreeAddrInfoExW(query_ctx->results);
#ifndef NDEBUG
                        query_ctx->results = nullptr;
#endif // !defined(NDEBUG)
                    }
                };

                if (vm_ctx->valid()) {
                    vm_ctx->pending_operations.erase(
                        vm_ctx->pending_operations.iterator_to(*query_ctx));
                }

                if (!ec) {
                    INT error = GetAddrInfoExOverlappedResult(
                        &query_ctx->overlapped);
                    if (error == WSA_E_CANCELLED) {
                        ec = asio::error::operation_aborted;
                    } else {
                        ec = asio_error_code(
                            error, asio::error::get_system_category());
                    }
                }

                auto push_results = [&ec,&query_ctx](lua_State* L) {
                    if (ec) {
                        lua_pushnil(L);
                        return;
                    }

                    lua_newtable(L);
                    lua_pushliteral(L, "address");
                    lua_pushliteral(L, "port");
                    lua_pushliteral(L, "canonical_name");

                    if (
                        query_ctx->results && query_ctx->results->ai_canonname
                    ) {
                        std::wstring_view w{query_ctx->results->ai_canonname};
                        push(L, nowide::narrow(w));
                    } else {
                        lua_pushnil(L);
                    }

                    int i = 1;
                    for (
                        auto it = query_ctx->results ; it ; it = it->ai_next
                    ) {
                        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

                        asio::ip::tcp::endpoint ep;
                        ep.resize(it->ai_addrlen);
                        std::memcpy(ep.data(), it->ai_addr, ep.size());

                        lua_pushvalue(L, -1 -4);
                        auto a = static_cast<asio::ip::address*>(
                            lua_newuserdata(L, sizeof(asio::ip::address))
                        );
                        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                        setmetatable(L, -2);
                        new (a) asio::ip::address{ep.address()};
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -3);
                        lua_pushinteger(L, ep.port());
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -2);
                        lua_pushvalue(L, -1 -1);
                        lua_rawset(L, -3);

                        lua_rawseti(L, -6, i++);
                    }
                    lua_pop(L, 4);
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::fast_auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))));
            }));
#else // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->tcp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    bool has_cname = (flags & asio::ip::resolver_base::canonical_name) ?
        true : false;

    service->tcp_resolver.async_resolve(
        asio::ip::tcp::v4(), host, tostringview(L, 2), flags,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,has_cname](
                const asio_error_code& ec,
                asio::ip::tcp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results,has_cname](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "address");
                        lua_pushliteral(fib, "port");
                        lua_pushliteral(fib, "canonical_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/3);

                            lua_pushvalue(fib, -1 -3);
                            auto a = static_cast<asio::ip::address*>(
                                lua_newuserdata(fib, sizeof(asio::ip::address))
                            );
                            rawgetp(fib, LUA_REGISTRYINDEX, &ip_address_mt_key);
                            setmetatable(fib, -2);
                            new (a) asio::ip::address{res.endpoint().address()};
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -2);
                            lua_pushinteger(fib, res.endpoint().port());
                            lua_rawset(fib, -3);

                            if (has_cname) {
                                lua_pushvalue(fib, -1 -1);
                                push(fib, res.host_name());
                                lua_rawset(fib, -3);
                            }

                            lua_rawseti(fib, -5, i++);
                        }
                        lua_pop(fib, 3);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

    return lua_yield(L, 0);
}

static int tcp_get_address_v6_info(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    asio::ip::resolver_base::flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        flags = asio::ip::resolver_base::v4_mapped;
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 7904 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_21::in_word_set((s).data(), (s).size()), {})
#line 4227 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    std::string host;
    switch (lua_type(L, 1)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    case LUA_TSTRING:
        host = tostringview(L, 1);
        break;
    case LUA_TUSERDATA: {
        if (!lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        auto& a = *static_cast<asio::ip::address*>(lua_touserdata(L, 1));
        host = a.to_string();
        flags |= asio::ip::resolver_base::numeric_host;
    }
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TSTRING:
        break;
    case LUA_TNUMBER:
        flags |= asio::ip::resolver_base::numeric_service;
    }


#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    ADDRINFOEXW hints;
    hints.ai_flags = static_cast<int>(flags);
    hints.ai_family = AF_INET6;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_addrlen = 0;
    hints.ai_canonname = nullptr;
    hints.ai_addr = nullptr;
    hints.ai_blob = nullptr;
    hints.ai_bloblen = 0;
    hints.ai_provider = nullptr;
    hints.ai_next = nullptr;

    auto query_ctx = std::make_shared<get_address_info_context_t>(
        vm_ctx->strand().context());

    {
        HANDLE hCompletion = CreateEventA(/*lpEventAttributes=*/NULL,
                                          /*bManualReset=*/TRUE,
                                          /*bInitialState=*/FALSE,
                                          /*lpName=*/NULL);
        if (hCompletion == NULL) {
            asio_error_code ec(
                GetLastError(), asio::error::get_system_category());
            push(L, ec);
            return lua_error(L);
        }

        asio_error_code ec;
        query_ctx->hCompletion.assign(hCompletion, ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        query_ctx->overlapped.hEvent = hCompletion;
    }

    auto service = tostringview(L, 2);
    INT error = GetAddrInfoExW(nowide::widen(host).c_str(),
                               nowide::widen(service).c_str(), NS_DNS,
                               /*lpNspId=*/NULL, &hints, &query_ctx->results,
                               /*timeout=*/NULL, &query_ctx->overlapped,
                               /*lpCompletionRoutine=*/NULL,
                               &query_ctx->hCancel);
    if (error != WSA_IO_PENDING) {
        // the operation completed immediately
        asio_error_code ec(
            WSAGetLastError(), asio::error::get_system_category());
        push(L, ec);
        return lua_error(L);
    }

    lua_pushlightuserdata(L, query_ctx.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto query_ctx = static_cast<get_address_info_context_t*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            GetAddrInfoExCancel(&query_ctx->hCancel);
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    vm_ctx->pending_operations.push_back(*query_ctx);
    query_ctx->hCompletion.async_wait(
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,query_ctx](asio_error_code ec) {
                BOOST_SCOPE_EXIT_ALL(&) {
                    if (query_ctx->results) {
                        FreeAddrInfoExW(query_ctx->results);
#ifndef NDEBUG
                        query_ctx->results = nullptr;
#endif // !defined(NDEBUG)
                    }
                };

                if (vm_ctx->valid()) {
                    vm_ctx->pending_operations.erase(
                        vm_ctx->pending_operations.iterator_to(*query_ctx));
                }

                if (!ec) {
                    INT error = GetAddrInfoExOverlappedResult(
                        &query_ctx->overlapped);
                    if (error == WSA_E_CANCELLED) {
                        ec = asio::error::operation_aborted;
                    } else {
                        ec = asio_error_code(
                            error, asio::error::get_system_category());
                    }
                }

                auto push_results = [&ec,&query_ctx](lua_State* L) {
                    if (ec) {
                        lua_pushnil(L);
                        return;
                    }

                    lua_newtable(L);
                    lua_pushliteral(L, "address");
                    lua_pushliteral(L, "port");
                    lua_pushliteral(L, "canonical_name");

                    if (
                        query_ctx->results && query_ctx->results->ai_canonname
                    ) {
                        std::wstring_view w{query_ctx->results->ai_canonname};
                        push(L, nowide::narrow(w));
                    } else {
                        lua_pushnil(L);
                    }

                    int i = 1;
                    for (
                        auto it = query_ctx->results ; it ; it = it->ai_next
                    ) {
                        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

                        asio::ip::tcp::endpoint ep;
                        ep.resize(it->ai_addrlen);
                        std::memcpy(ep.data(), it->ai_addr, ep.size());

                        lua_pushvalue(L, -1 -4);
                        auto a = static_cast<asio::ip::address*>(
                            lua_newuserdata(L, sizeof(asio::ip::address))
                        );
                        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                        setmetatable(L, -2);
                        new (a) asio::ip::address{ep.address()};
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -3);
                        lua_pushinteger(L, ep.port());
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -2);
                        lua_pushvalue(L, -1 -1);
                        lua_rawset(L, -3);

                        lua_rawseti(L, -6, i++);
                    }
                    lua_pop(L, 4);
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::fast_auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))));
            }));
#else // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->tcp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    bool has_cname = (flags & asio::ip::resolver_base::canonical_name) ?
        true : false;

    service->tcp_resolver.async_resolve(
        asio::ip::tcp::v6(), host, tostringview(L, 2), flags,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,has_cname](
                const asio_error_code& ec,
                asio::ip::tcp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results,has_cname](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "address");
                        lua_pushliteral(fib, "port");
                        lua_pushliteral(fib, "canonical_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/3);

                            lua_pushvalue(fib, -1 -3);
                            auto a = static_cast<asio::ip::address*>(
                                lua_newuserdata(fib, sizeof(asio::ip::address))
                            );
                            rawgetp(fib, LUA_REGISTRYINDEX, &ip_address_mt_key);
                            setmetatable(fib, -2);
                            new (a) asio::ip::address{res.endpoint().address()};
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -2);
                            lua_pushinteger(fib, res.endpoint().port());
                            lua_rawset(fib, -3);

                            if (has_cname) {
                                lua_pushvalue(fib, -1 -1);
                                push(fib, res.host_name());
                                lua_rawset(fib, -3);
                            }

                            lua_rawseti(fib, -5, i++);
                        }
                        lua_pop(fib, 3);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

    return lua_yield(L, 0);
}

static int tcp_get_name_info(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TNUMBER);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->tcp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    service->tcp_resolver.async_resolve(
        asio::ip::tcp::endpoint(*a, lua_tointeger(L, 2)),
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber](
                const asio_error_code& ec,
                asio::ip::tcp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "host_name");
                        lua_pushliteral(fib, "service_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/2);

                            lua_pushvalue(fib, -1 -2);
                            push(fib, res.host_name());
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -1);
                            push(fib, res.service_name());
                            lua_rawset(fib, -3);

                            lua_rawseti(fib, -4, i++);
                        }
                        lua_pop(fib, 2);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );

    return lua_yield(L, 0);
}

static int tcp_listen(lua_State* L)
{
    luaL_checktype(L, 1, LUA_TSTRING);

    auto& vm_ctx = get_vm_context(L);

    std::string_view host = tostringview(L, 1);
    std::uint16_t port;

    asio_error_code ec;

    {
        auto idx = host.rfind(':');
        if (idx == std::string_view::npos) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        std::string_view p = host.substr(idx + 1);
        if (p.starts_with("0") && p.size() != 1) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }

        if (p.size() == 0) {
            port = 0;
        } else {
            auto res = std::from_chars(p.data(), p.data() + p.size(), port);
            if (res.ec != std::errc{} || res.ptr != p.data() + p.size()) {
                asio::ip::tcp::resolver resolver{vm_ctx.strand().context()};
                asio::ip::tcp::resolver::flags flags =
                    asio::ip::resolver_base::numeric_host |
                    asio::ip::resolver_base::passive;
                auto results = resolver.resolve("", p, flags, ec);
                if (ec) {
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                if (results.size() == 0) {
                    ec = asio::error::addrinfo_errors::service_not_found;
                    push(L, static_cast<std::error_code>(ec));
                    return lua_error(L);
                }
                port = results.begin()->endpoint().port();
            }
        }

        host.remove_suffix(p.size() + 1);
    }

    bool is_ipv6 = false;
    if (host.starts_with("[")) {
        if (host.back() != ']') {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        host.remove_suffix(1);
        host.remove_prefix(1);
        is_ipv6 = true;
    }

    auto addr = asio::ip::make_address(host, ec);
    if (ec) {
        asio::ip::tcp::resolver resolver{vm_ctx.strand().context()};
        asio::ip::tcp::resolver::flags flags =
            asio::ip::resolver_base::numeric_service |
            asio::ip::resolver_base::passive;
        auto results = resolver.resolve(host, "0", flags, ec);
        if (ec) {
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        if (results.size() == 0) {
            ec = asio::error::netdb_errors::host_not_found;
            push(L, static_cast<std::error_code>(ec));
            return lua_error(L);
        }
        addr = results.begin()->endpoint().address();
        is_ipv6 = addr.is_v6();
    } else if ((is_ipv6 && addr.is_v4()) || (!is_ipv6 && addr.is_v6())) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto a = static_cast<asio::ip::tcp::acceptor*>(
        lua_newuserdata(L, sizeof(asio::ip::tcp::acceptor))
    );
    rawgetp(L, LUA_REGISTRYINDEX, &ip_tcp_acceptor_mt_key);
    setmetatable(L, -2);
    new (a) asio::ip::tcp::acceptor{vm_ctx.strand().context()};

    if (is_ipv6) {
        a->open(asio::ip::tcp::v6(), ec);
    } else {
        a->open(asio::ip::tcp::v4(), ec);
    }
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }

    a->set_option(asio::socket_base::reuse_address{true}, ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }

    a->bind({addr, port}, ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }

    a->listen(asio::socket_base::max_listen_connections, ec);
    if (ec) {
        push(L, static_cast<std::error_code>(ec));
        return lua_error(L);
    }

    return 1;
}

static int udp_socket_new(lua_State* L)
{
    auto& vm_ctx = get_vm_context(L);
    auto a = static_cast<udp_socket*>(lua_newuserdata(L, sizeof(udp_socket)));
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    setmetatable(L, -2);
    new (a) tcp_socket{vm_ctx.strand().context()};
    return 1;
}

#line 4946 "../src/ip.cpp"


static int udp_socket_connect(lua_State* L)
{
    luaL_checktype(L, 3, LUA_TNUMBER);
    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto s = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!s || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 2));
    if (!addr || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::ip::udp::endpoint ep(*addr, lua_tointeger(L, 3));

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++s->nbusy;
    s->socket.async_connect(ep, asio::bind_cancellation_slot(cancel_slot,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,s](const asio_error_code& ec) {
                if (!vm_ctx->valid())
                    return;

                --s->nbusy;

                auto opt_args = vm_context::options::arguments;
                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(opt_args, hana::make_tuple(ec))));
            }
        ))
    );

    return lua_yield(L, 0);
}

#line 5902 "../src/ip.cpp"


static int udp_socket_receive(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::socket_base::message_flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 8546 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_27::in_word_set((s).data(), (s).size()), {})
#line 5967 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++sock->nbusy;
    sock->socket.async_receive(
        asio::buffer(bs->data.get(), bs->size),
        flags,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,sock](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                boost::ignore_unused(buf);
                if (!vm_ctx->valid())
                    return;

                --sock->nbusy;

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int udp_socket_receive_from(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::socket_base::message_flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 8644 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_28::in_word_set((s).data(), (s).size()), {})
#line 6071 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    auto remote_sender = std::make_shared<asio::ip::udp::endpoint>();

    ++sock->nbusy;
    sock->socket.async_receive_from(
        asio::buffer(bs->data.get(), bs->size),
        *remote_sender,
        flags,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,remote_sender,buf=bs->data,sock](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                boost::ignore_unused(buf);
                if (!vm_ctx->valid())
                    return;

                --sock->nbusy;

                auto addr_pusher = [&remote_sender](lua_State* L) {
                    auto a = static_cast<asio::ip::address*>(
                        lua_newuserdata(L, sizeof(asio::ip::address))
                    );
                    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                    setmetatable(L, -2);
                    new (a) asio::ip::address{remote_sender->address()};
                    return 1;
                };
                auto port = remote_sender->port();

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(
                                ec, bytes_transferred, addr_pusher, port))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int udp_socket_send(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    asio::socket_base::message_flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 8757 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_29::in_word_set((s).data(), (s).size()), {})
#line 6190 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++sock->nbusy;
    sock->socket.async_send(
        asio::buffer(bs->data.get(), bs->size),
        flags,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,sock](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                boost::ignore_unused(buf);
                if (!vm_ctx->valid())
                    return;

                --sock->nbusy;

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

static int udp_socket_send_to(lua_State* L)
{
    lua_settop(L, 5);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto sock = static_cast<udp_socket*>(lua_touserdata(L, 1));
    if (!sock || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_udp_socket_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    auto bs = static_cast<byte_span_handle*>(lua_touserdata(L, 2));
    if (!bs || !lua_getmetatable(L, 2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &byte_span_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    }

    auto addr = static_cast<asio::ip::address*>(lua_touserdata(L, 3));
    if (!addr || !lua_getmetatable(L, 3)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    }

    std::uint16_t port;
    if (lua_type(L, 4) != LUA_TNUMBER) {
        push(L, std::errc::invalid_argument, "arg", 4);
        return lua_error(L);
    }
    port = lua_tointeger(L, 4);

    asio::socket_base::message_flags flags = {};
    switch (lua_type(L, 5)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 5);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 5, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 5);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 8873 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_30::in_word_set((s).data(), (s).size()), {})
#line 6312 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 5);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    auto cancel_slot = set_default_interrupter(L, *vm_ctx);

    ++sock->nbusy;
    sock->socket.async_send_to(
        asio::buffer(bs->data.get(), bs->size),
        asio::ip::udp::endpoint{*addr, port},
        flags,
        asio::bind_cancellation_slot(cancel_slot, asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,buf=bs->data,sock](
                const asio_error_code& ec, std::size_t bytes_transferred
            ) {
                boost::ignore_unused(buf);
                if (!vm_ctx->valid())
                    return;

                --sock->nbusy;

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, bytes_transferred))));
            }
        ))
    );

    return lua_yield(L, 0);
}

#line 6481 "../src/ip.cpp"


static int udp_socket_mt_index(lua_State* L)
{
    auto key = tostringview(L, 2);
    return 
#line 8925 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_32::in_word_set((key).data(), (key).size()), [](lua_State* L) -> int {
            push(L, errc::bad_index, "index", 2);
            return lua_error(L);
        })
#line 6601 "../src/ip.cpp"
(L);
}

static int udp_get_address_info(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    asio::ip::resolver_base::flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        flags = asio::ip::resolver_base::address_configured |
            asio::ip::resolver_base::v4_mapped;
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 8968 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_33::in_word_set((s).data(), (s).size()), {})
#line 6650 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    std::string host;
    switch (lua_type(L, 1)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    case LUA_TSTRING:
        host = tostringview(L, 1);
        break;
    case LUA_TUSERDATA: {
        if (!lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        auto& a = *static_cast<asio::ip::address*>(lua_touserdata(L, 1));
        host = a.to_string();
        flags |= asio::ip::resolver_base::numeric_host;
    }
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TSTRING:
        break;
    case LUA_TNUMBER:
        flags |= asio::ip::resolver_base::numeric_service;
    }

#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    ADDRINFOEXW hints;
    hints.ai_flags = static_cast<int>(flags);
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_protocol = IPPROTO_UDP;
    hints.ai_addrlen = 0;
    hints.ai_canonname = nullptr;
    hints.ai_addr = nullptr;
    hints.ai_blob = nullptr;
    hints.ai_bloblen = 0;
    hints.ai_provider = nullptr;
    hints.ai_next = nullptr;

    auto query_ctx = std::make_shared<get_address_info_context_t>(
        vm_ctx->strand().context());

    {
        HANDLE hCompletion = CreateEventA(/*lpEventAttributes=*/NULL,
                                          /*bManualReset=*/TRUE,
                                          /*bInitialState=*/FALSE,
                                          /*lpName=*/NULL);
        if (hCompletion == NULL) {
            asio_error_code ec(
                GetLastError(), asio::error::get_system_category());
            push(L, ec);
            return lua_error(L);
        }

        asio_error_code ec;
        query_ctx->hCompletion.assign(hCompletion, ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        query_ctx->overlapped.hEvent = hCompletion;
    }

    auto service = tostringview(L, 2);
    INT error = GetAddrInfoExW(nowide::widen(host).c_str(),
                               nowide::widen(service).c_str(), NS_DNS,
                               /*lpNspId=*/NULL, &hints, &query_ctx->results,
                               /*timeout=*/NULL, &query_ctx->overlapped,
                               /*lpCompletionRoutine=*/NULL,
                               &query_ctx->hCancel);
    if (error != WSA_IO_PENDING) {
        // the operation completed immediately
        asio_error_code ec(
            WSAGetLastError(), asio::error::get_system_category());
        push(L, ec);
        return lua_error(L);
    }

    lua_pushlightuserdata(L, query_ctx.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto query_ctx = static_cast<get_address_info_context_t*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            GetAddrInfoExCancel(&query_ctx->hCancel);
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    vm_ctx->pending_operations.push_back(*query_ctx);
    query_ctx->hCompletion.async_wait(
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,query_ctx](asio_error_code ec) {
                BOOST_SCOPE_EXIT_ALL(&) {
                    if (query_ctx->results) {
                        FreeAddrInfoExW(query_ctx->results);
#ifndef NDEBUG
                        query_ctx->results = nullptr;
#endif // !defined(NDEBUG)
                    }
                };

                if (vm_ctx->valid()) {
                    vm_ctx->pending_operations.erase(
                        vm_ctx->pending_operations.iterator_to(*query_ctx));
                }

                if (!ec) {
                    INT error = GetAddrInfoExOverlappedResult(
                        &query_ctx->overlapped);
                    if (error == WSA_E_CANCELLED) {
                        ec = asio::error::operation_aborted;
                    } else {
                        ec = asio_error_code(
                            error, asio::error::get_system_category());
                    }
                }

                auto push_results = [&ec,&query_ctx](lua_State* L) {
                    if (ec) {
                        lua_pushnil(L);
                        return;
                    }

                    lua_newtable(L);
                    lua_pushliteral(L, "address");
                    lua_pushliteral(L, "port");
                    lua_pushliteral(L, "canonical_name");

                    if (
                        query_ctx->results && query_ctx->results->ai_canonname
                    ) {
                        std::wstring_view w{query_ctx->results->ai_canonname};
                        push(L, nowide::narrow(w));
                    } else {
                        lua_pushnil(L);
                    }

                    int i = 1;
                    for (
                        auto it = query_ctx->results ; it ; it = it->ai_next
                    ) {
                        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

                        asio::ip::tcp::endpoint ep;
                        ep.resize(it->ai_addrlen);
                        std::memcpy(ep.data(), it->ai_addr, ep.size());

                        lua_pushvalue(L, -1 -4);
                        auto a = static_cast<asio::ip::address*>(
                            lua_newuserdata(L, sizeof(asio::ip::address))
                        );
                        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                        setmetatable(L, -2);
                        new (a) asio::ip::address{ep.address()};
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -3);
                        lua_pushinteger(L, ep.port());
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -2);
                        lua_pushvalue(L, -1 -1);
                        lua_rawset(L, -3);

                        lua_rawseti(L, -6, i++);
                    }
                    lua_pop(L, 4);
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::fast_auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))));
            }));
#else // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->udp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    bool has_cname = (flags & asio::ip::resolver_base::canonical_name) ?
        true : false;

    service->udp_resolver.async_resolve(
        host, tostringview(L, 2), flags,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,has_cname](
                const asio_error_code& ec,
                asio::ip::udp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results,has_cname](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "address");
                        lua_pushliteral(fib, "port");
                        lua_pushliteral(fib, "canonical_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/3);

                            lua_pushvalue(fib, -1 -3);
                            auto a = static_cast<asio::ip::address*>(
                                lua_newuserdata(fib, sizeof(asio::ip::address))
                            );
                            rawgetp(fib, LUA_REGISTRYINDEX, &ip_address_mt_key);
                            setmetatable(fib, -2);
                            new (a) asio::ip::address{res.endpoint().address()};
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -2);
                            lua_pushinteger(fib, res.endpoint().port());
                            lua_rawset(fib, -3);

                            if (has_cname) {
                                lua_pushvalue(fib, -1 -1);
                                push(fib, res.host_name());
                                lua_rawset(fib, -3);
                            }

                            lua_rawseti(fib, -5, i++);
                        }
                        lua_pop(fib, 3);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

    return lua_yield(L, 0);
}

static int udp_get_address_v4_info(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    asio::ip::resolver_base::flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 9297 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_34::in_word_set((s).data(), (s).size()), {})
#line 6988 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    std::string host;
    switch (lua_type(L, 1)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    case LUA_TSTRING:
        host = tostringview(L, 1);
        break;
    case LUA_TUSERDATA: {
        if (!lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        auto& a = *static_cast<asio::ip::address*>(lua_touserdata(L, 1));
        host = a.to_string();
        flags |= asio::ip::resolver_base::numeric_host;
    }
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TSTRING:
        break;
    case LUA_TNUMBER:
        flags |= asio::ip::resolver_base::numeric_service;
    }

#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    ADDRINFOEXW hints;
    hints.ai_flags = static_cast<int>(flags);
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_protocol = IPPROTO_UDP;
    hints.ai_addrlen = 0;
    hints.ai_canonname = nullptr;
    hints.ai_addr = nullptr;
    hints.ai_blob = nullptr;
    hints.ai_bloblen = 0;
    hints.ai_provider = nullptr;
    hints.ai_next = nullptr;

    auto query_ctx = std::make_shared<get_address_info_context_t>(
        vm_ctx->strand().context());

    {
        HANDLE hCompletion = CreateEventA(/*lpEventAttributes=*/NULL,
                                          /*bManualReset=*/TRUE,
                                          /*bInitialState=*/FALSE,
                                          /*lpName=*/NULL);
        if (hCompletion == NULL) {
            asio_error_code ec(
                GetLastError(), asio::error::get_system_category());
            push(L, ec);
            return lua_error(L);
        }

        asio_error_code ec;
        query_ctx->hCompletion.assign(hCompletion, ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        query_ctx->overlapped.hEvent = hCompletion;
    }

    auto service = tostringview(L, 2);
    INT error = GetAddrInfoExW(nowide::widen(host).c_str(),
                               nowide::widen(service).c_str(), NS_DNS,
                               /*lpNspId=*/NULL, &hints, &query_ctx->results,
                               /*timeout=*/NULL, &query_ctx->overlapped,
                               /*lpCompletionRoutine=*/NULL,
                               &query_ctx->hCancel);
    if (error != WSA_IO_PENDING) {
        // the operation completed immediately
        asio_error_code ec(
            WSAGetLastError(), asio::error::get_system_category());
        push(L, ec);
        return lua_error(L);
    }

    lua_pushlightuserdata(L, query_ctx.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto query_ctx = static_cast<get_address_info_context_t*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            GetAddrInfoExCancel(&query_ctx->hCancel);
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    vm_ctx->pending_operations.push_back(*query_ctx);
    query_ctx->hCompletion.async_wait(
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,query_ctx](asio_error_code ec) {
                BOOST_SCOPE_EXIT_ALL(&) {
                    if (query_ctx->results) {
                        FreeAddrInfoExW(query_ctx->results);
#ifndef NDEBUG
                        query_ctx->results = nullptr;
#endif // !defined(NDEBUG)
                    }
                };

                if (vm_ctx->valid()) {
                    vm_ctx->pending_operations.erase(
                        vm_ctx->pending_operations.iterator_to(*query_ctx));
                }

                if (!ec) {
                    INT error = GetAddrInfoExOverlappedResult(
                        &query_ctx->overlapped);
                    if (error == WSA_E_CANCELLED) {
                        ec = asio::error::operation_aborted;
                    } else {
                        ec = asio_error_code(
                            error, asio::error::get_system_category());
                    }
                }

                auto push_results = [&ec,&query_ctx](lua_State* L) {
                    if (ec) {
                        lua_pushnil(L);
                        return;
                    }

                    lua_newtable(L);
                    lua_pushliteral(L, "address");
                    lua_pushliteral(L, "port");
                    lua_pushliteral(L, "canonical_name");

                    if (
                        query_ctx->results && query_ctx->results->ai_canonname
                    ) {
                        std::wstring_view w{query_ctx->results->ai_canonname};
                        push(L, nowide::narrow(w));
                    } else {
                        lua_pushnil(L);
                    }

                    int i = 1;
                    for (
                        auto it = query_ctx->results ; it ; it = it->ai_next
                    ) {
                        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

                        asio::ip::tcp::endpoint ep;
                        ep.resize(it->ai_addrlen);
                        std::memcpy(ep.data(), it->ai_addr, ep.size());

                        lua_pushvalue(L, -1 -4);
                        auto a = static_cast<asio::ip::address*>(
                            lua_newuserdata(L, sizeof(asio::ip::address))
                        );
                        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                        setmetatable(L, -2);
                        new (a) asio::ip::address{ep.address()};
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -3);
                        lua_pushinteger(L, ep.port());
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -2);
                        lua_pushvalue(L, -1 -1);
                        lua_rawset(L, -3);

                        lua_rawseti(L, -6, i++);
                    }
                    lua_pop(L, 4);
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::fast_auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))));
            }));
#else // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->udp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    bool has_cname = (flags & asio::ip::resolver_base::canonical_name) ?
        true : false;

    service->udp_resolver.async_resolve(
        asio::ip::udp::v4(), host, tostringview(L, 2), flags,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,has_cname](
                const asio_error_code& ec,
                asio::ip::udp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results,has_cname](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "address");
                        lua_pushliteral(fib, "port");
                        lua_pushliteral(fib, "canonical_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/3);

                            lua_pushvalue(fib, -1 -3);
                            auto a = static_cast<asio::ip::address*>(
                                lua_newuserdata(fib, sizeof(asio::ip::address))
                            );
                            rawgetp(fib, LUA_REGISTRYINDEX, &ip_address_mt_key);
                            setmetatable(fib, -2);
                            new (a) asio::ip::address{res.endpoint().address()};
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -2);
                            lua_pushinteger(fib, res.endpoint().port());
                            lua_rawset(fib, -3);

                            if (has_cname) {
                                lua_pushvalue(fib, -1 -1);
                                push(fib, res.host_name());
                                lua_rawset(fib, -3);
                            }

                            lua_rawseti(fib, -5, i++);
                        }
                        lua_pop(fib, 3);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

    return lua_yield(L, 0);
}

static int udp_get_address_v6_info(lua_State* L)
{
    lua_settop(L, 3);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    asio::ip::resolver_base::flags flags = {};
    switch (lua_type(L, 3)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 3);
        return lua_error(L);
    case LUA_TNIL:
        flags = asio::ip::resolver_base::v4_mapped;
        break;
    case LUA_TTABLE:
        for (int i = 1 ;; ++i) {
            lua_rawgeti(L, 3, i);
            switch (lua_type(L, -1)) {
            default:
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            case LUA_TNIL:
                lua_pop(L, 1);
                goto end_for;
            case LUA_TSTRING:
                break;
            }

            auto s = tostringview(L);
            lua_pop(L, 1);
            auto f = 
#line 9627 "libemilua.so.0.11.5.p/bc2_ip.cpp"
 ::emilua::gperf::detail::value_or(::emilua::gperf::detail::Perfect_Hash_35::in_word_set((s).data(), (s).size()), {})
#line 7327 "../src/ip.cpp"
;
            if (f == 0) {
                push(L, std::errc::invalid_argument, "arg", 3);
                return lua_error(L);
            }
            flags |= f;
        }
    }
 end_for:

    std::string host;
    switch (lua_type(L, 1)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    case LUA_TSTRING:
        host = tostringview(L, 1);
        break;
    case LUA_TUSERDATA: {
        if (!lua_getmetatable(L, 1)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
        if (!lua_rawequal(L, -1, -2)) {
            push(L, std::errc::invalid_argument, "arg", 1);
            return lua_error(L);
        }
        auto& a = *static_cast<asio::ip::address*>(lua_touserdata(L, 1));
        host = a.to_string();
        flags |= asio::ip::resolver_base::numeric_host;
    }
    }

    switch (lua_type(L, 2)) {
    default:
        push(L, std::errc::invalid_argument, "arg", 2);
        return lua_error(L);
    case LUA_TSTRING:
        break;
    case LUA_TNUMBER:
        flags |= asio::ip::resolver_base::numeric_service;
    }

#if defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    ADDRINFOEXW hints;
    hints.ai_flags = static_cast<int>(flags);
    hints.ai_family = AF_INET6;
    hints.ai_socktype = SOCK_DGRAM;
    hints.ai_protocol = IPPROTO_UDP;
    hints.ai_addrlen = 0;
    hints.ai_canonname = nullptr;
    hints.ai_addr = nullptr;
    hints.ai_blob = nullptr;
    hints.ai_bloblen = 0;
    hints.ai_provider = nullptr;
    hints.ai_next = nullptr;

    auto query_ctx = std::make_shared<get_address_info_context_t>(
        vm_ctx->strand().context());

    {
        HANDLE hCompletion = CreateEventA(/*lpEventAttributes=*/NULL,
                                          /*bManualReset=*/TRUE,
                                          /*bInitialState=*/FALSE,
                                          /*lpName=*/NULL);
        if (hCompletion == NULL) {
            asio_error_code ec(
                GetLastError(), asio::error::get_system_category());
            push(L, ec);
            return lua_error(L);
        }

        asio_error_code ec;
        query_ctx->hCompletion.assign(hCompletion, ec);
        if (ec) {
            push(L, ec);
            return lua_error(L);
        }

        query_ctx->overlapped.hEvent = hCompletion;
    }

    auto service = tostringview(L, 2);
    INT error = GetAddrInfoExW(nowide::widen(host).c_str(),
                               nowide::widen(service).c_str(), NS_DNS,
                               /*lpNspId=*/NULL, &hints, &query_ctx->results,
                               /*timeout=*/NULL, &query_ctx->overlapped,
                               /*lpCompletionRoutine=*/NULL,
                               &query_ctx->hCancel);
    if (error != WSA_IO_PENDING) {
        // the operation completed immediately
        asio_error_code ec(
            WSAGetLastError(), asio::error::get_system_category());
        push(L, ec);
        return lua_error(L);
    }

    lua_pushlightuserdata(L, query_ctx.get());
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto query_ctx = static_cast<get_address_info_context_t*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            GetAddrInfoExCancel(&query_ctx->hCancel);
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    vm_ctx->pending_operations.push_back(*query_ctx);
    query_ctx->hCompletion.async_wait(
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,query_ctx](asio_error_code ec) {
                BOOST_SCOPE_EXIT_ALL(&) {
                    if (query_ctx->results) {
                        FreeAddrInfoExW(query_ctx->results);
#ifndef NDEBUG
                        query_ctx->results = nullptr;
#endif // !defined(NDEBUG)
                    }
                };

                if (vm_ctx->valid()) {
                    vm_ctx->pending_operations.erase(
                        vm_ctx->pending_operations.iterator_to(*query_ctx));
                }

                if (!ec) {
                    INT error = GetAddrInfoExOverlappedResult(
                        &query_ctx->overlapped);
                    if (error == WSA_E_CANCELLED) {
                        ec = asio::error::operation_aborted;
                    } else {
                        ec = asio_error_code(
                            error, asio::error::get_system_category());
                    }
                }

                auto push_results = [&ec,&query_ctx](lua_State* L) {
                    if (ec) {
                        lua_pushnil(L);
                        return;
                    }

                    lua_newtable(L);
                    lua_pushliteral(L, "address");
                    lua_pushliteral(L, "port");
                    lua_pushliteral(L, "canonical_name");

                    if (
                        query_ctx->results && query_ctx->results->ai_canonname
                    ) {
                        std::wstring_view w{query_ctx->results->ai_canonname};
                        push(L, nowide::narrow(w));
                    } else {
                        lua_pushnil(L);
                    }

                    int i = 1;
                    for (
                        auto it = query_ctx->results ; it ; it = it->ai_next
                    ) {
                        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

                        asio::ip::tcp::endpoint ep;
                        ep.resize(it->ai_addrlen);
                        std::memcpy(ep.data(), it->ai_addr, ep.size());

                        lua_pushvalue(L, -1 -4);
                        auto a = static_cast<asio::ip::address*>(
                            lua_newuserdata(L, sizeof(asio::ip::address))
                        );
                        rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
                        setmetatable(L, -2);
                        new (a) asio::ip::address{ep.address()};
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -3);
                        lua_pushinteger(L, ep.port());
                        lua_rawset(L, -3);

                        lua_pushvalue(L, -1 -2);
                        lua_pushvalue(L, -1 -1);
                        lua_rawset(L, -3);

                        lua_rawseti(L, -6, i++);
                    }
                    lua_pop(L, 4);
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::fast_auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))));
            }));
#else // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)
    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->udp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    bool has_cname = (flags & asio::ip::resolver_base::canonical_name) ?
        true : false;

    service->udp_resolver.async_resolve(
        asio::ip::udp::v6(), host, tostringview(L, 2), flags,
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber,has_cname](
                const asio_error_code& ec,
                asio::ip::udp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results,has_cname](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "address");
                        lua_pushliteral(fib, "port");
                        lua_pushliteral(fib, "canonical_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/3);

                            lua_pushvalue(fib, -1 -3);
                            auto a = static_cast<asio::ip::address*>(
                                lua_newuserdata(fib, sizeof(asio::ip::address))
                            );
                            rawgetp(fib, LUA_REGISTRYINDEX, &ip_address_mt_key);
                            setmetatable(fib, -2);
                            new (a) asio::ip::address{res.endpoint().address()};
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -2);
                            lua_pushinteger(fib, res.endpoint().port());
                            lua_rawset(fib, -3);

                            if (has_cname) {
                                lua_pushvalue(fib, -1 -1);
                                push(fib, res.host_name());
                                lua_rawset(fib, -3);
                            }

                            lua_rawseti(fib, -5, i++);
                        }
                        lua_pop(fib, 3);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );
#endif // defined(_WIN32_WINNT) && (_WIN32_WINNT >= 0x0602)

    return lua_yield(L, 0);
}

static int udp_get_name_info(lua_State* L)
{
    luaL_checktype(L, 2, LUA_TNUMBER);

    auto vm_ctx = get_vm_context(L).shared_from_this();
    auto current_fiber = vm_ctx->current_fiber();
    EMILUA_CHECK_SUSPEND_ALLOWED(*vm_ctx, L);

    auto a = static_cast<asio::ip::address*>(lua_touserdata(L, 1));
    if (!a || !lua_getmetatable(L, 1)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }
    rawgetp(L, LUA_REGISTRYINDEX, &ip_address_mt_key);
    if (!lua_rawequal(L, -1, -2)) {
        push(L, std::errc::invalid_argument, "arg", 1);
        return lua_error(L);
    }

    resolver_service* service = nullptr;

    for (auto& op: vm_ctx->pending_operations) {
        service = dynamic_cast<resolver_service*>(&op);
        if (service)
            break;
    }

    if (!service) {
        service = new resolver_service{vm_ctx->strand().context()};
        vm_ctx->pending_operations.push_back(*service);
    }

    lua_pushlightuserdata(L, service);
    lua_pushcclosure(
        L,
        [](lua_State* L) -> int {
            auto service = static_cast<resolver_service*>(
                lua_touserdata(L, lua_upvalueindex(1)));
            try {
                service->udp_resolver.cancel();
            } catch (const asio_system_error&) {}
            return 0;
        },
        1);
    set_interrupter(L, *vm_ctx);

    service->udp_resolver.async_resolve(
        asio::ip::udp::endpoint(*a, lua_tointeger(L, 2)),
        asio::bind_executor(
            vm_ctx->strand_using_defer(),
            [vm_ctx,current_fiber](
                const asio_error_code& ec,
                asio::ip::udp::resolver::results_type results
            ) {
                auto push_results = [&ec,&results](lua_State* fib) {
                    if (ec) {
                        lua_pushnil(fib);
                    } else {
                        lua_createtable(fib, /*narr=*/results.size(),
                                        /*nrec=*/0);
                        lua_pushliteral(fib, "host_name");
                        lua_pushliteral(fib, "service_name");

                        int i = 1;
                        for (const auto& res: results) {
                            lua_createtable(fib, /*narr=*/0, /*nrec=*/2);

                            lua_pushvalue(fib, -1 -2);
                            push(fib, res.host_name());
                            lua_rawset(fib, -3);

                            lua_pushvalue(fib, -1 -1);
                            push(fib, res.service_name());
                            lua_rawset(fib, -3);

                            lua_rawseti(fib, -4, i++);
                        }
                        lua_pop(fib, 2);
                    }
                };

                vm_ctx->fiber_resume(
                    current_fiber,
                    hana::make_set(
                        vm_context::options::auto_detect_interrupt,
                        hana::make_pair(
                            vm_context::options::arguments,
                            hana::make_tuple(ec, push_results))
                    ));
            }
        )
    );

    return lua_yield(L, 0);
}

void init_ip(lua_State* L)
{
    lua_pushlightuserdata(L, &ip_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/7);

        lua_pushliteral(L, "tostring");
        lua_pushcfunction(L, ip_tostring);
        lua_rawset(L, -3);

        lua_pushliteral(L, "toendpoint");
        lua_pushcfunction(L, ip_toendpoint);
        lua_rawset(L, -3);

        lua_pushliteral(L, "host_name");
        lua_pushcfunction(L, ip_host_name);
        lua_rawset(L, -3);

        lua_pushliteral(L, "connect");
        int res = luaL_loadbuffer(
            L, reinterpret_cast<char*>(ip_connect_bytecode),
            ip_connect_bytecode_size, nullptr);
        assert(res == 0); boost::ignore_unused(res);
        rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
        rawgetp(L, LUA_REGISTRYINDEX, &raw_type_key);
        rawgetp(L, LUA_REGISTRYINDEX, &raw_next_key);
        rawgetp(L, LUA_REGISTRYINDEX, &raw_pcall_key);
        push(L, make_error_code(asio::error::not_found));
        push(L, std::errc::invalid_argument);
        lua_call(L, 6, 1);
        lua_rawset(L, -3);

        lua_pushliteral(L, "address");
        {
            lua_createtable(L, /*narr=*/0, /*nrec=*/6);

            lua_pushliteral(L, "new");
            lua_pushcfunction(L, address_new);
            lua_rawset(L, -3);

            lua_pushliteral(L, "any_v4");
            lua_pushcfunction(L, address_any_v4);
            lua_rawset(L, -3);

            lua_pushliteral(L, "any_v6");
            lua_pushcfunction(L, address_any_v6);
            lua_rawset(L, -3);

            lua_pushliteral(L, "loopback_v4");
            lua_pushcfunction(L, address_loopback_v4);
            lua_rawset(L, -3);

            lua_pushliteral(L, "loopback_v6");
            lua_pushcfunction(L, address_loopback_v6);
            lua_rawset(L, -3);

            lua_pushliteral(L, "broadcast_v4");
            lua_pushcfunction(L, address_broadcast_v4);
            lua_rawset(L, -3);
        }
        lua_rawset(L, -3);

        lua_pushliteral(L, "tcp");
        {
            lua_createtable(L, /*narr=*/0, /*nrec=*/8);

            lua_pushliteral(L, "socket");
            {
                lua_createtable(L, /*narr=*/0, /*nrec=*/1);

                lua_pushliteral(L, "new");
                lua_pushcfunction(L, tcp_socket_new);
                lua_rawset(L, -3);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "acceptor");
            {
                lua_createtable(L, /*narr=*/0, /*nrec=*/1);

                lua_pushliteral(L, "new");
                lua_pushcfunction(L, tcp_acceptor_new);
                lua_rawset(L, -3);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_address_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, tcp_get_address_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_address_v4_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, tcp_get_address_v4_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_address_v6_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, tcp_get_address_v6_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_name_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, tcp_get_name_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "listen");
            lua_pushcfunction(L, tcp_listen);
            lua_rawset(L, -3);

            lua_pushliteral(L, "dial");
            int res = luaL_loadbuffer(
                L, reinterpret_cast<char*>(ip_dial_bytecode),
                ip_dial_bytecode_size, nullptr);
            assert(res == 0); boost::ignore_unused(res);
            {
                lua_pushliteral(L, "get_address_info");
                lua_rawget(L, -4);
            }
            lua_pushcfunction(L, tcp_socket_new);
            lua_pushcfunction(L, ip_toendpoint2);
            rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
            rawgetp(L, LUA_REGISTRYINDEX, &raw_next_key);
            rawgetp(L, LUA_REGISTRYINDEX, &raw_pcall_key);
            push(L, make_error_code(asio::error::not_found));
            lua_call(L, 7, 1);
            lua_rawset(L, -3);
        }
        lua_rawset(L, -3);

        lua_pushliteral(L, "udp");
        {
            lua_createtable(L, /*narr=*/0, /*nrec=*/6);

            lua_pushliteral(L, "socket");
            {
                lua_createtable(L, /*narr=*/0, /*nrec=*/1);

                lua_pushliteral(L, "new");
                lua_pushcfunction(L, udp_socket_new);
                lua_rawset(L, -3);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_address_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, udp_get_address_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_address_v4_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, udp_get_address_v4_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_address_v6_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, udp_get_address_v6_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "get_name_info");
            {
                rawgetp(L, LUA_REGISTRYINDEX,
                        &var_args__retval1_to_error__fwd_retval2__key);
                rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
                lua_pushcfunction(L, udp_get_name_info);
                lua_call(L, 2, 1);
            }
            lua_rawset(L, -3);

            lua_pushliteral(L, "dial");
            int res = luaL_loadbuffer(
                L, reinterpret_cast<char*>(ip_dial_bytecode),
                ip_dial_bytecode_size, nullptr);
            assert(res == 0); boost::ignore_unused(res);
            {
                lua_pushliteral(L, "get_address_info");
                lua_rawget(L, -4);
            }
            lua_pushcfunction(L, udp_socket_new);
            lua_pushcfunction(L, ip_toendpoint2);
            rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
            rawgetp(L, LUA_REGISTRYINDEX, &raw_next_key);
            rawgetp(L, LUA_REGISTRYINDEX, &raw_pcall_key);
            push(L, make_error_code(asio::error::not_found));
            lua_call(L, 7, 1);
            lua_rawset(L, -3);
        }
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &ip_address_mt_key);
    {
        static_assert(std::is_trivially_destructible_v<asio::ip::address>);

        lua_createtable(L, /*narr=*/0, /*nrec=*/7);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "ip.address");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, address_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__newindex");
        lua_pushcfunction(L, address_mt_newindex);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__tostring");
        lua_pushcfunction(L, address_mt_tostring);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__eq");
        lua_pushcfunction(L, address_mt_eq);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__lt");
        lua_pushcfunction(L, address_mt_lt);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__le");
        lua_pushcfunction(L, address_mt_le);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &ip_tcp_socket_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "ip.tcp.socket");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, tcp_socket_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<tcp_socket>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &ip_tcp_acceptor_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "ip.tcp.acceptor");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, tcp_acceptor_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<asio::ip::tcp::acceptor>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &ip_udp_socket_mt_key);
    {
        lua_createtable(L, /*narr=*/0, /*nrec=*/3);

        lua_pushliteral(L, "__metatable");
        lua_pushliteral(L, "ip.udp.socket");
        lua_rawset(L, -3);

        lua_pushliteral(L, "__index");
        lua_pushcfunction(L, udp_socket_mt_index);
        lua_rawset(L, -3);

        lua_pushliteral(L, "__gc");
        lua_pushcfunction(L, finalizer<udp_socket>);
        lua_rawset(L, -3);
    }
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_socket_connect_key);
    rawgetp(L, LUA_REGISTRYINDEX, &var_args__retval1_to_error__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_connect);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_socket_read_some_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_read_some);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_socket_write_some_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_write_some);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_socket_receive_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_receive);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_socket_send_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_send);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

#if BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO
    lua_pushlightuserdata(L, &tcp_socket_send_file_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_send_file);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
#endif // BOOST_OS_WINDOWS && EMILUA_CONFIG_ENABLE_FILE_IO

    lua_pushlightuserdata(L, &tcp_socket_wait_key);
    rawgetp(L, LUA_REGISTRYINDEX, &var_args__retval1_to_error__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_socket_wait);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_acceptor_accept_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_acceptor_accept);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &tcp_acceptor_wait_key);
    rawgetp(L, LUA_REGISTRYINDEX, &var_args__retval1_to_error__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, tcp_acceptor_wait);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &udp_socket_connect_key);
    rawgetp(L, LUA_REGISTRYINDEX, &var_args__retval1_to_error__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, udp_socket_connect);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &udp_socket_receive_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, udp_socket_receive);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &udp_socket_receive_from_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval234__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, udp_socket_receive_from);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &udp_socket_send_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, udp_socket_send);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);

    lua_pushlightuserdata(L, &udp_socket_send_to_key);
    rawgetp(L, LUA_REGISTRYINDEX,
            &var_args__retval1_to_error__fwd_retval2__key);
    rawgetp(L, LUA_REGISTRYINDEX, &raw_error_key);
    lua_pushcfunction(L, udp_socket_send_to);
    lua_call(L, 2, 1);
    lua_rawset(L, LUA_REGISTRYINDEX);
}

} // namespace emilua
